"""
Created on Thu Aug 10 16:25:13 2023

@author: smvanessen

"""

# Clear all variables
from IPython import get_ipython
get_ipython().magic('reset -sf')

# Standard toolboxes
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import glob
import os
import re

# Add some other functions
from functions.subsample_routines import subsample_binned
from functions.distribution_scripts import MPM_from_DNR
from functions.theoretical_distributions import make_cdf_Forristall as Forr
from functions.distribution_scripts import calc_distribution
from functions.T_enc import calc_T_enc

# Convergence criterion
from functions.convergenceAS import stopping_crit_AS

# Close all figures
plt.close('all')


#############
# START INPUT
#############
    
# Acquisition function type to evaluate.
allacqnames = ['US','MV','USMV','MD','USPD','MVPD','IVR']
# allacqnames = ['USMV']

# Select used initial sampling option. Note: you can only select options for 
# which you first generated data with the Adaptive Screening method. For a 
# description of options A, B and C, also see Adaptive Screening. 
# Options: 'A', 'B' or 'C'
sampopt = 'B'  

# Line styles and colors in the plots
stylesgpr = ['-.']
colsgpr = ['green']  # matching distribution plots
stylesmfgpr = ['--']
colsmfgpr = ['blue'] # matching distribution plots
stylesacq = ['-','--','-.',(0, (3, 5, 1, 5)),(0, (3, 1, 1, 1, 1, 1)),':','--']
colsacq = ['darkseagreen','indianred','lightseagreen','darkmagenta','gold','royalblue','orange']

# Kernel name single-fidelity HF-GPR (preferred Mat32 for all)
HFkern = 'Mat32'
MFkern1 = 'Mat32' 
MFkern2 = 'Mat32'

# Annotation switch for the plot of the MPM as function of iterations until 
# convergence. Paper figure 12 includes annotation.
# 1 = with annotation, 0 = without annotation
annon = 1

# # SQI values for the different indicator quality cases (sqi), number of HF 
# # samples for convergence with GPR (nev_gpr) and number of HF samples for 
# # convergence with MF-GPR (nev_mfgpr). 
# # Cases: sa = Sagging, ho = Hogging
# sqi_sa = 14.9
# sqi_ho = 22.8

# # Lower and upper limit for event duration CFD [s]
# tlow = 10
# thigh = 50

# Plotsize
fsize_h = 7
fsize_v = 4.5

# Cases to run
# plotcases = ['GOOD','HOG','WORSE','SAG'] 
plotcases = ['GOOD','HOG','WORSE']
# plotcases = ['HOG']

#############
# END INPUT
#############

# Initialise
MPMtrue = np.nan*np.zeros((len(plotcases),len(allacqnames)))
MPMpredGPR = np.nan*np.zeros((len(plotcases),len(allacqnames)))
MPMpredMFGPR = np.nan*np.zeros((len(plotcases),len(allacqnames)))
M1_gpr = np.nan*np.zeros((len(plotcases),len(allacqnames)))
M1_mfgpr = np.nan*np.zeros((len(plotcases),len(allacqnames)))
M2_gpr = np.nan*np.zeros((len(plotcases),len(allacqnames)))
M2_mfgpr = np.nan*np.zeros((len(plotcases),len(allacqnames)))
M3_gpr = np.nan*np.zeros((len(plotcases),len(allacqnames)))
M3_mfgpr = np.nan*np.zeros((len(plotcases),len(allacqnames)))


# Text in name of plots
case_string = '_'.join(plotcases)
print(case_string)


# Loop over cases
for bb,plotcase in enumerate(plotcases):
    
    # # Considered cases (for plot name)
    # conscase = conscase + '_' + plotcase
    
    # Plot settings
    if plotcase == 'HOG' or plotcase == 'SAG':
        
        # Folder where the input CSV files are (generated by the Adaptive Screening procedure) (in)
        # Folder where we write the output figures (out)
        in_files = 'out_files_vbm_PRADS'
        out_figs = 'out_figs_paper_PRADS'
        
        # VBM plot limits for convergence plots [Nm]
        min_gpr_unc = 0.5e9 # combined plot with uncertainty
        max_gpr_unc = 1.3e9 
        min_mfgpr_unc = 0.5e9 
        max_mfgpr_unc = 1.3e9 
        min_gpr_nounc = 0.85e9 # plot without uncertainty
        max_gpr_nounc = 1.15e9 
        min_mfgpr_nounc = 0.85e9 
        max_mfgpr_nounc = 1.15e9 
        min_gpr_uncsep = 0.85e9 # separate plot with uncertainty
        max_gpr_uncsep = 1.15e9 
        min_mfgpr_uncsep = 0.85e9 
        max_mfgpr_uncsep = 1.15e9 
        
        # Spectral properties basis wave system (for ref. Pexp)
        dur = 1800 # target duration [s]
        Hs = 13.2 # sign wave height [m]
        Tp = 10.0 # peak wave period [s]
        Vs = 10*0.51444444 # current speed or vessel speed [m/s]
        mu = np.pi # rad
    
        # Total available data duration [s]
        Tdata = 30*3600
        
        # Convergence limits stopping criterion
        limMAE = 5e6 # [Nm]
        limCOV = 0.003
        
        # # Number of variations of each case to run (VBM: hogging / sagging)
        # numvariations = 1
        
        
    elif plotcase == 'GOOD' or plotcase == 'WORSE':
        
        in_files = 'out_files_waves_PRADS'
        out_figs = 'out_figs_paper_PRADS'
        
        # VBM plot limits for convergence plots [Nm]
        min_gpr_unc = 9 # combined plot with uncertainty
        max_gpr_unc = 11.5
        min_mfgpr_unc = 9 
        max_mfgpr_unc = 11.5 
        min_gpr_nounc = 9 # plot without uncertainty
        max_gpr_nounc = 11.5 
        min_mfgpr_nounc = 9 
        max_mfgpr_nounc = 11.5 
        min_gpr_uncsep = 9 # separate plot with uncertainty
        max_gpr_uncsep = 11.5 
        min_mfgpr_uncsep = 9 
        max_mfgpr_uncsep = 11.5
        
        # Spectral properties basis wave system (for ref. Forristall and Pexp)
        dur = 3600 # target duration [s]
        Hs = 10.0 # sign wave height [m]
        Tp = 11.0 # peak wave period [s]
        Vs = 0*0.51444444 # current speed or vessel speed [m/s]
        mu = np.pi # rad
    
        # Total available data duration [s]
        Tdata = 50*3600
        
        # Convergence limits stopping criterion
        limMAE = 0.05 # [m]
        limCOV = 0.003
        
        # # Number of variations of each case to run (WAVES: GoodInd / WorseInd)
        # numvariations = 2
        
        
    # Noise texts (figure titles and names)
    if plotcase == 'HOG':
        indcase = 'Hogging'
        # indalias = 'V_{hog}\prime'
        indalias = 'V\prime'
        indunit = 'Nm'
        # hfalias = 'V_{hog}\prime\prime'
        hfalias = 'V\prime\prime'
        hfunit = 'Nm'
        
        # Number of HF samples to include in plot
        sampmaxplot1 = 150.001 # Convergence
        sampstep1 = 25
        sampmaxplot2 = 200.001 # All considered samples
        sampstep2 = 25
        
        # Selected acquisition functions for this case (they can only be removed at the end, e.g. IVR and/or USPD!)
        acqnames = allacqnames.copy()
        
        # Speed [m/s], heading [rad] & wave period [s]
        Tp = 10.0               # [s]
        Vs = 10.00*0.51444444   # [m/s]
        mu = np.pi              # [rad]
        
        # Target exposure duration [s]
        dur = 1800
        
    elif plotcase == 'SAG':
        indcase = 'Sagging'
        indalias = 'V_{sag}\prime'
        indunit = 'Nm'
        hfalias = 'V_{sag}\prime\prime'
        hfunit = 'Nm'
        
        # Number of HF samples to include in plot
        sampmaxplot1 = 150.001 # Convergence
        sampstep1 = 25
        sampmaxplot2 = 200.001 # All considered samples
        sampstep2 = 25
        
        # Selected acquisition functions for this case (they can only be removed at the end, e.g. IVR and/or USPD!)
        acqnames = allacqnames[:-1]
        
        # Speed [m/s], heading [rad] & wave period [s]
        Tp = 10.0               # [s]
        Vs = 10.00*0.51444444   # [m/s]
        mu = np.pi              # [rad]
        
        # Target exposure duration [s]
        dur = 1800

        
    elif plotcase == 'GOOD':
        indcase = 'GoodInd'
        indalias = 'C_{good}\prime'
        indunit = 'm'
        hfalias = 'C\prime\prime'
        hfunit = 'm'
            
        # Number of HF samples to include in plot
        sampmaxplot1 = 150.001 # Convergence
        sampstep1 = 25
        sampmaxplot2 = 200.001 # All considered samples
        sampstep2 = 25
        
        # Selected acquisition functions for this case (they can only be removed at the end, e.g. IVR and/or USPD!)
        acqnames = allacqnames.copy()
        
        # Speed [m/s], heading [rad] & wave period [s]
        Tp = 11.0  # [s]
        Vs = 0.0   # [kn]
        mu = 0     # [deg]
        
        # Target exposure duration [s]
        dur = 3600

        
    elif plotcase == 'WORSE':
        indcase = 'WorseInd'
        indalias = 'C_{worse}\prime'
        indunit = 'm'
        hfalias = 'C\prime\prime'
        hfunit = 'm'
        
        # Number of HF samples to include in plot
        sampmaxplot1 = 150.001 # Convergence
        sampstep1 = 25
        sampmaxplot2 = 200.001 # All considered samples
        sampstep2 = 25
                
        # Selected acquisition functions for this case (they can only be removed at the end, e.g. IVR and/or USPD!)
        acqnames = allacqnames.copy()
        
        # Speed [m/s], heading [rad] & wave period [s]
        Tp = 11.0  # [s]
        Vs = 0.0   # [kn]
        mu = 0     # [deg]
        
        # Target exposure duration [s]
        dur = 3600

        
    # Check wether output directory exist, if not make it
    isExist1 = os.path.exists(out_figs)
    if not isExist1:
       os.makedirs(out_figs)
       
    # Initial sample options (see Adaptive Screening)
    if sampopt == 'A':
        sampopt_txt = 'A_LFsubsamp'; N_hf = 2
        txtadd = f'A_LFsubsamp_exp{np.int(dur/60)}min'
    elif sampopt == 'B':
        sampopt_txt = 'B_LFfit'; N_hf = 2
        txtadd = f'B_LFfit_exp{np.int(dur/60)}min'
    elif sampopt == 'C':
        sampopt_txt = 'C_LFisHFfit'; N_hf = 2
        txtadd = f'C_LFisHFfit_exp{np.int(dur/60)}min'
    
    # Derive target probability of exceedance, using the average number of wave encounters in target duration
    Tpe = calc_T_enc(Tp,Vs,mu)
    ne_exp = dur/Tpe
    pexp = 1/ne_exp
    
    # Initialise
    conv_mpm_gpr = np.nan*np.zeros((len(plotcases),len(allacqnames)))
    conv_mpmu95_gpr = np.nan*np.zeros((len(plotcases),len(allacqnames)))
    conv_mpm_mfgpr = np.nan*np.zeros((len(plotcases),len(allacqnames)))
    conv_mpmu95_mfgpr = np.nan*np.zeros((len(plotcases),len(allacqnames)))
    
    # Initialise plots as function of SQI (for selected acquisition function)
    sqi_all = []
    mpm_gpr_all = [];        mpm_mfgpr_all = []
    mpmu95_gpr_all = [];     mpmu95_mfgpr_all = []
    time_gpr_low = [];       time_mfgpr_low =[]
    time_gpr_hig = [];       time_mfgpr_hig = []
    
    # Prepare location of annotation
    locan_gpr_10_unc = np.zeros(len(allacqnames))
    locan_mfgpr_10_unc = np.zeros(len(allacqnames))
    locan_gpr_unc = np.zeros(len(allacqnames))
    locan_mfgpr_unc = np.zeros(len(allacqnames))
    locan_gpr_10_nounc = np.zeros(len(allacqnames))
    locan_mfgpr_10_nounc = np.zeros(len(allacqnames))
    locan_gpr_nounc = np.zeros(len(allacqnames))
    locan_mfgpr_nounc = np.zeros(len(allacqnames))
    
    # Synthesise ylabels
    ylabs = r'$\widehat{%s}$ [%s]' % (hfalias,hfunit)    
        
    # # Append SQI values and number of HF events
    # sqi_all = np.append(sqi_all,sqi)    
    
    # Read MCS LF and HF true and training time traces from CSV file   
    # ALREADY CORRECTED FOR BASIS NUMBER OF WAVES, INSTEAD OF NUMBER OF LF OR HF PEAKS, WITHIN ADAPTIVE SCREENING SCRIPT
    DatLFtrue = np.loadtxt(f'{in_files}/TrueLFdata_{indcase}_{txtadd}.csv', delimiter=',')
    v_true_l = DatLFtrue[:,0]
    p_true_l = DatLFtrue[:,1]
    DatHFtrue = np.loadtxt(f'{in_files}/TrueHFdata_{indcase}_{txtadd}.csv', delimiter=',') # 50h MCS realisation used in Adaptive Screening
    v_true_h = DatHFtrue[:,0]
    p_true_h = DatHFtrue[:,1]
    DatLFtrain = np.loadtxt(f'{in_files}/TrainLFdata_{indcase}_{txtadd}.csv', delimiter=',')
    v_train_l = DatLFtrain[:,0]
    p_train_l = DatLFtrain[:,1]

    # Subsample the MCS true distribution data for plotting later
    binsLF = np.arange(0,np.nanmax(v_true_l),np.nanmax(v_true_l)/100)
    binsHF = np.arange(0,np.nanmax(v_true_h),np.nanmax(v_true_h)/100)
    v_true_l_sampled,p_true_l_sampled = subsample_binned(v_true_l,p_true_l,binsLF)
    v_true_h_sampled,p_true_h_sampled = subsample_binned(v_true_h,p_true_h,binsHF)
      
    # Max values (for plotting) 
    maxVal = np.ceil(1.1*np.nanmax(v_true_h)) # HF
    maxVal_lf = np.ceil(1.1*np.nanmax(v_true_l)) # LF
    
    # Min PoE value (for plotting)
    minP = 0.9*np.nanmin(p_true_h) # HF
    
    # Calculate / derive true MPM from full available dataset
    MPM_true_MCS = MPM_from_DNR(v_true_h,p_true_h,ne_exp) 
    
    # For reference, also derive the LF true MPM
    MPM_true_MCS_lf = MPM_from_DNR(v_true_l,p_true_l,ne_exp) 
        
    # Initialise
    sizehor = 6.5
    sizever = 3
    plt.figure(20000+bb*100,figsize=(sizehor,sizever)) # GPR all data
    plt.figure(30000+bb*100,figsize=(sizehor,sizever)) # MF-GPR all data
    plt.figure(40000+bb*100,figsize=(sizehor,sizever)) # GPR until convergence
    plt.figure(50000+bb*100,figsize=(sizehor,sizever)) # MF-GPR until convergence
    plt.figure(140000+bb*100,figsize=(sizehor,sizever)) # GPR until convergence, without 10-samples indication
    plt.figure(150000+bb*100,figsize=(sizehor,sizever)) # MF-GPR until convergence, without 10-samples indication
    # plt.figure(60000+bb*100,figsize=(sizehor,sizever)) # GPR until convergence, black/white without uncertainty
    # plt.figure(70000+bb*100,figsize=(sizehor,sizever)) # MF-GPR until convergence, black/white without uncertainty
    # plt.figure(80000+bb*100,figsize=(sizehor,sizever)) # GPR until convergence, colored without uncertainty
    # plt.figure(90000+bb*100,figsize=(sizehor,sizever)) # MF-GPR until convergence, colored without uncertainty
    x_ticks1 = np.arange(0,sampmaxplot1,sampstep1)
    x_ticks2 = np.arange(0,sampmaxplot2,sampstep2)
    
    # Version with subplots for GPR and MF-GPR
    plt.figure(160000+bb*100,figsize=(8.5,5))
    
    # Loop over acquisition functions (only functions that are available for both GPR and MF-GPR)
    for mm,acqname in enumerate(acqnames):
        
        acqname = acqnames[mm]
        print(f'Case {indcase}, acq. function {acqname} ({mm}/{len(acqnames)-1})')

        #### Find last available iteration (in this case probably #200, as we later interpolate in this database using a convergence criterion)
    
        # Load data from HF-GPR with all iteration results (MPM etc.)
        file_conv_hf = f'{in_files}/Conv_HF-GPR_{indcase}_Set-{HFkern}_{txtadd}_{acqname}_Nhf{N_hf}.csv'
        dat_conv_hf_all = pd.read_csv(file_conv_hf, delimiter=',')
        
        # Extract last iteration number and max. number of samples
        nosamp_hf = dat_conv_hf_all['NoSamp'].to_numpy()
        
        # Max. number of samples, and max. number of iterations
        maxsamp_hf = int(np.nanmax(nosamp_hf))
        maxit_hf = int(maxsamp_hf-N_hf)
        
        print(f'{indcase} with HF-GPR and {acqname}: available are {maxit_hf} iterations ({maxsamp_hf} samples)')
        
        # Load data from MF-GPR with all iteration results (MPM etc.)
        if acqname != 'IVR':
            file_conv_mf = f'{in_files}/Conv_MF-GPR_{indcase}_Set-{MFkern1}-{MFkern2}_{txtadd}_{acqname}_Nhf{N_hf}.csv'
            dat_conv_mf_all = pd.read_csv(file_conv_mf, delimiter=',')
            
            # Extract last iteration number and max. number of samples
            nosamp_mf = dat_conv_mf_all['NoSamp'].to_numpy()
            
            # Max. number of samples, and max. number of iterations
            maxsamp_mf = int(np.nanmax(nosamp_mf))
            maxit_mf = int(maxsamp_mf-N_hf)
            
            print(f'{indcase} with MF-GPR and {acqname}: available are {maxit_mf} iterations ({maxsamp_mf} samples)')
        
        
        #### Load data for this last iteration and plot convergence
    
        # Actual MPM values GPR
        mpms_GPR = dat_conv_hf_all['MPM_HFGPR'].to_numpy()
        mpmsU95_GPR = dat_conv_hf_all['U95_MPM_HFGPR'].to_numpy()
        
        # Plot all iterations for GPR
        plt.figure(20000+bb*100)
        if mm == 0:
            plt.plot([0,sampmaxplot2],[MPM_true_MCS,MPM_true_MCS],linestyle=':',linewidth=1,color='black'); plt.text(sampmaxplot2-30,MPM_true_MCS+0.08,'True MCS',color='black')
        plt.plot(nosamp_hf,mpms_GPR,linestyle=stylesacq[mm],color=colsacq[mm],label=f'{acqname}')  
        plt.fill_between(nosamp_hf.flatten(), (mpms_GPR - 0.5*mpmsU95_GPR).flatten(), 
                          (mpms_GPR + 0.5*mpmsU95_GPR).flatten(), facecolor=colsacq[mm], alpha=0.3)#,label=f'{acqname} U95%')
        
        if acqname != 'IVR':
            # Actual MPM values MF-GPR
            mpms_MFGPR = dat_conv_mf_all['MPM_MFGPR'].to_numpy() # Use when AS is run again after fixing typo
            mpmsU95_MFGPR = dat_conv_mf_all['U95_MPM_MFGPR'].to_numpy() # Use when AS is run again after fixing typo
                      
            # Plot all iterations for MF-GPR
            plt.figure(30000+bb*100)
            if mm == 0:
                plt.plot([0,sampmaxplot2],[MPM_true_MCS,MPM_true_MCS],linestyle=':',linewidth=1,color='black'); plt.text(sampmaxplot2-30,MPM_true_MCS+0.08,'True MCS',color='black')
            plt.plot(nosamp_mf,mpms_MFGPR,linestyle=stylesacq[mm],color=colsacq[mm],label=f'{acqname}')  
            plt.fill_between(nosamp_mf.flatten(), (mpms_MFGPR - 0.5*mpmsU95_MFGPR).flatten(), 
                              (mpms_MFGPR + 0.5*mpmsU95_MFGPR).flatten(), facecolor=colsacq[mm], alpha=0.3)#,label=f'{acqname} U95%')
        
        
        #### Apply stopping criterion to determine at which iteration we are converged
        
        # File name start of CSV with predicted data per iteration (exclude "itXX.csv", will be added later)
        fn_preds_GPR = f'{in_files}/HFGPR_HFpred_{indcase}_Set-{HFkern}-nf_{txtadd}_HFacq-{acqname}_Nhf{N_hf}_HF'
        casename_GPR = f'{indcase} - GPR - {acqname}'
        min_it1 = 5
        ploton1 = False
        convit1,convsamp1 = stopping_crit_AS(nosamp_hf,mpms_GPR,mpmsU95_GPR,fn_preds_GPR,casename_GPR,pexp,min_it1,limMAE,limCOV,ploton1) # final
                
        if acqname != 'IVR':
            fn_preds_MFGPR = f'{in_files}/MFGPR_HFpred_{indcase}_Set-{MFkern1}-{MFkern2}-nf_{txtadd}_MFacq-{acqname}_Nhf{N_hf}_MF'
            casename_MFGPR = f'{indcase} - MF-GPR - {acqname}'
            min_it2 = 5
            ploton2 = False
            convit2,convsamp2 = stopping_crit_AS(nosamp_mf,mpms_MFGPR,mpmsU95_MFGPR,fn_preds_MFGPR,casename_MFGPR,pexp,min_it2,limMAE,limCOV,ploton2) # final
        
        # # Derive range in required duration [h]
        # HFtime_gpr = [convsamp1*tlow/3600,convsamp1*thigh/3600]
        # time_gpr_low = np.append(time_gpr_low,HFtime_gpr[0])
        # time_gpr_hig = np.append(time_gpr_hig,HFtime_gpr[1])
        
        # if acqname != 'IVR':
        #     HFtime_mfgpr = [convsamp2*tlow/3600,convsamp2*thigh/3600]    
        #     time_mfgpr_low = np.append(time_mfgpr_low,HFtime_mfgpr[0])
        #     time_mfgpr_hig = np.append(time_mfgpr_hig,HFtime_mfgpr[1])        

        # Prepare for annotation
        if annon == 1:
            i_samp10 = 10-N_hf
            
            if len(mpms_GPR) >= 10:
                gpr10 = np.round((mpms_GPR[i_samp10]/MPM_true_MCS-1)*100,1);  signgpr10 = ''
                if gpr10>0: 
                    signgpr10 = '+' 
            else:
                gpr10 = np.nan; signgpr10 = ''        
            
            if np.isnan(convsamp1):
                gprconv = np.nan
            else:
                gprconv = np.round((mpms_GPR[convit1]/MPM_true_MCS-1)*100,1); signgprconv = ''
            if gprconv>0: 
                signgprconv = '+' 
            
            if acqname != 'IVR':
                
                if len(mpms_MFGPR) >= 10:
                    mfgpr10 = np.round((mpms_MFGPR[i_samp10]/MPM_true_MCS-1)*100,1); signmfgpr10 = ''
                    if mfgpr10>0: 
                        signmfgpr10 = '+' 
                else: 
                    mfgpr10 = np.nan; signmfgpr10 = ''
                
                if np.isnan(convsamp2):
                    mfgprconv = np.nan
                else:
                    mfgprconv = np.round((mpms_MFGPR[convit2]/MPM_true_MCS-1)*100,1); signmfgprconv = ''
                if mfgprconv>0: 
                    signmfgprconv = '+'  
        
        # GPR Prepare location of annotation for combined plots with uncertainty
        locan_gpr_10_unc[mm] = min_gpr_unc + (max_gpr_unc - min_gpr_unc)/20*(mm+1.5)
        locan_gpr_unc[mm] = min_gpr_unc + (max_gpr_unc - min_gpr_unc)/12*(mm+0.2)
                  
        # GPR Prepare location of annotation for combined plots without uncertainty
        locan_gpr_10_nounc[mm] = min_gpr_nounc + (max_gpr_nounc - min_gpr_nounc)/20*(mm+1.5)
        locan_gpr_nounc[mm] = min_gpr_nounc + (max_gpr_nounc - min_gpr_nounc)/10*(mm+0.5)
           
        # GPR Prepare location of annotation for combined plots with uncertainty (same for each mm and same for 10 samples & converged, but keep here for readability)
        locan_gpr_uncsep = min_gpr_uncsep + (max_gpr_uncsep - min_gpr_uncsep)/10*4
        
        if acqname != 'IVR':
            # MF-GPR Prepare location of annotation for combined plots with uncertainty
            locan_mfgpr_10_unc[mm] = min_mfgpr_unc + (max_mfgpr_unc - min_mfgpr_unc)/20*(mm+1.5)
            locan_mfgpr_unc[mm] = min_mfgpr_unc + (max_mfgpr_unc - min_mfgpr_unc)/10*(mm+0.5)
           
            # MF-GPR Prepare location of annotation for combined plots without uncertainty
            locan_mfgpr_10_nounc[mm] = min_mfgpr_nounc + (max_mfgpr_nounc - min_mfgpr_nounc)/20*(mm+1.5)
            locan_mfgpr_nounc[mm] = min_mfgpr_nounc + (max_mfgpr_nounc - min_mfgpr_nounc)/10*(mm+0.5)
        
            # MF-GPR Prepare location of annotation for combined plots with uncertainty (same for each mm and same for 10 samples & converged, but keep here for readability)
            locan_mfgpr_uncsep = min_mfgpr_uncsep + (max_mfgpr_uncsep - min_mfgpr_uncsep)/10*4



        #### Now plot the same, but until convergence
        
        # GPR
        plt.figure(140000+bb*100)
        if mm == 0:
            plt.plot([0,sampmaxplot1],[MPM_true_MCS,MPM_true_MCS],linestyle=':',linewidth=1,color='black'); plt.text(0.85*sampmaxplot1,MPM_true_MCS+0.08,'True MCS',color='black')
        if ~np.isnan(gprconv):
            plt.plot(nosamp_hf[:convsamp1-1],mpms_GPR[:convsamp1-1],linestyle=stylesacq[mm],color=colsacq[mm],linewidth = 0.75,label=f'{acqname}')  
            plt.fill_between(nosamp_hf[:convsamp1-1].flatten(), (mpms_GPR[:convsamp1-1] - 0.5*mpmsU95_GPR[:convsamp1-1]).flatten(), 
                              (mpms_GPR[:convsamp1-1] + 0.5*mpmsU95_GPR[:convsamp1-1]).flatten(), facecolor=colsacq[mm], alpha=0.3)#,label=f'{acqname} U95%')
        else: 
            plt.plot(nosamp_hf,mpms_GPR,linestyle=stylesacq[mm],color=colsacq[mm],linewidth = 0.75,label=f'{acqname} (not converged)')  
            plt.fill_between(nosamp_hf.flatten(), (mpms_GPR - 0.5*mpmsU95_GPR).flatten(), 
                              (mpms_GPR + 0.5*mpmsU95_GPR).flatten(), facecolor=colsacq[mm], alpha=0.3)#,label=f'{acqname} U95%')
        if annon == 1:
            if ~np.isnan(gprconv):
                plt.plot([nosamp_hf[convsamp1-1]-1,nosamp_hf[convsamp1-1]-1],[mpms_GPR[convsamp1-1],locan_gpr_unc[mm]],color=colsacq[mm],linewidth=0.2)
                plt.text(nosamp_hf[convsamp1-1],locan_gpr_unc[mm],f'converged\n{acqname}: {signgprconv}{gprconv}%',color=colsacq[mm],fontsize=8)
  
                   
        # MF-GPR
        if acqname != 'IVR':
            plt.figure(150000+bb*100)
            if mm == 0:
                plt.plot([0,sampmaxplot1],[MPM_true_MCS,MPM_true_MCS],linestyle=':',linewidth=1,color='black'); plt.text(0.85*sampmaxplot1,MPM_true_MCS+0.08,'True MCS',color='black')
            if ~np.isnan(mfgprconv):
                plt.plot(nosamp_mf[:convsamp2-1],mpms_MFGPR[:convsamp2-1],linestyle=stylesacq[mm],color=colsacq[mm],linewidth = 0.75,label=f'{acqname}')    
                plt.fill_between(nosamp_mf[:convsamp2-1].flatten(), (mpms_MFGPR[:convsamp2-1] - 0.5*mpmsU95_MFGPR[:convsamp2-1]).flatten(), 
                                  (mpms_MFGPR[:convsamp2-1] + 0.5*mpmsU95_MFGPR[:convsamp2-1]).flatten(), facecolor=colsacq[mm], alpha=0.3)#,label=f'{acqname} U95%')
            else:
                plt.plot(nosamp_mf,mpms_MFGPR,linestyle=stylesacq[mm],color=colsacq[mm],linewidth = 0.75,label=f'{acqname} (not converged)')  
                plt.fill_between(nosamp_mf.flatten(), (mpms_MFGPR - 0.5*mpmsU95_MFGPR).flatten(), 
                                  (mpms_MFGPR + 0.5*mpmsU95_MFGPR).flatten(), facecolor=colsacq[mm], alpha=0.3)#,label=f'{acqname} U95%')
            if annon == 1:
                if ~np.isnan(mfgprconv):
                    plt.plot([nosamp_mf[convit2],nosamp_mf[convit2]],[mpms_MFGPR[convit2],locan_mfgpr_unc[mm]],color=colsacq[mm],linewidth=0.2)
                    plt.text(nosamp_mf[convit2],locan_mfgpr_unc[mm],f'converged\n{acqname}: {signmfgprconv}{mfgprconv}%',color=colsacq[mm],fontsize=8)
        
        # Both together in subplots
        plt.figure(160000+bb*100)
        plt.subplot(211)
        if mm == 0:
            plt.plot([0,sampmaxplot1],[MPM_true_MCS,MPM_true_MCS],linestyle=':',linewidth=1,color='black'); plt.text(0.85*sampmaxplot1,MPM_true_MCS+0.08,'True MCS',color='black')
        if ~np.isnan(gprconv):
            plt.plot(nosamp_hf[:convsamp1-1],mpms_GPR[:convsamp1-1],linestyle=stylesacq[mm],color=colsacq[mm],linewidth = 0.75,label=f'{acqname}')  
            plt.fill_between(nosamp_hf[:convsamp1-1].flatten(), (mpms_GPR[:convsamp1-1] - 0.5*mpmsU95_GPR[:convsamp1-1]).flatten(), 
                              (mpms_GPR[:convsamp1-1] + 0.5*mpmsU95_GPR[:convsamp1-1]).flatten(), facecolor=colsacq[mm], alpha=0.2)#,label=f'{acqname} U95%')
        else: 
            plt.plot(nosamp_hf,mpms_GPR,linestyle=stylesacq[mm],color=colsacq[mm],linewidth = 0.75,label=f'{acqname} (not converged)')  
            plt.fill_between(nosamp_hf.flatten(), (mpms_GPR - 0.5*mpmsU95_GPR).flatten(), 
                              (mpms_GPR + 0.5*mpmsU95_GPR).flatten(), facecolor=colsacq[mm], alpha=0.2)#,label=f'{acqname} U95%')
        if annon == 1:
            if ~np.isnan(gprconv):
                plt.plot([nosamp_hf[convsamp1-1]-1,nosamp_hf[convsamp1-1]-1],[mpms_GPR[convsamp1-1],locan_gpr_unc[mm]],color=colsacq[mm],linewidth=0.2)
                # plt.text(nosamp_hf[convsamp1-1],locan_gpr_unc[mm],f'converged\n{acqname}: {signgprconv}{gprconv}%',color=colsacq[mm],fontsize=8)
                plt.text(nosamp_hf[convsamp1-1],locan_gpr_unc[mm],f'{acqname}: {signgprconv}{gprconv}%',color=colsacq[mm],fontsize=8)
        if acqname != 'IVR':
            plt.subplot(212)
            if mm == 0:
                plt.plot([0,sampmaxplot1],[MPM_true_MCS,MPM_true_MCS],linestyle=':',linewidth=1,color='black'); plt.text(0.85*sampmaxplot1,MPM_true_MCS+0.08,'True MCS',color='black')
            if ~np.isnan(mfgprconv):
                plt.plot(nosamp_mf[:convsamp2-1],mpms_MFGPR[:convsamp2-1],linestyle=stylesacq[mm],color=colsacq[mm],linewidth = 0.75,label=f'{acqname}')    
                plt.fill_between(nosamp_mf[:convsamp2-1].flatten(), (mpms_MFGPR[:convsamp2-1] - 0.5*mpmsU95_MFGPR[:convsamp2-1]).flatten(), 
                                  (mpms_MFGPR[:convsamp2-1] + 0.5*mpmsU95_MFGPR[:convsamp2-1]).flatten(), facecolor=colsacq[mm], alpha=0.2)#,label=f'{acqname} U95%')
            else:
                plt.plot(nosamp_mf,mpms_MFGPR,linestyle=stylesacq[mm],color=colsacq[mm],linewidth = 0.75,label=f'{acqname} (not converged)')  
                plt.fill_between(nosamp_mf.flatten(), (mpms_MFGPR - 0.5*mpmsU95_MFGPR).flatten(), 
                                  (mpms_MFGPR + 0.5*mpmsU95_MFGPR).flatten(), facecolor=colsacq[mm], alpha=0.2)#,label=f'{acqname} U95%')
            if annon == 1:
                if ~np.isnan(mfgprconv):
                    plt.plot([nosamp_mf[convit2],nosamp_mf[convit2]],[mpms_MFGPR[convit2],locan_mfgpr_unc[mm]],color=colsacq[mm],linewidth=0.2)
                    # plt.text(nosamp_mf[convit2],locan_mfgpr_unc[mm],f'converged\n{acqname}: {signmfgprconv}{mfgprconv}%',color=colsacq[mm],fontsize=8)
                    plt.text(nosamp_mf[convit2],locan_mfgpr_unc[mm],f'{acqname}: {signmfgprconv}{mfgprconv}%',color=colsacq[mm],fontsize=8)

        #### Now plot the same, but until convergence and including 10-samples indication
        
        # GPR
        plt.figure(40000+bb*100)
        if mm == 0:
            plt.plot([0,sampmaxplot1],[MPM_true_MCS,MPM_true_MCS],linestyle=':',linewidth=1,color='black'); plt.text(0.85*sampmaxplot1,MPM_true_MCS+0.08,'True MCS',color='black')
        if ~np.isnan(gprconv):
            plt.plot(nosamp_hf[:convsamp1-1],mpms_GPR[:convsamp1-1],linestyle=stylesacq[mm],color=colsacq[mm],linewidth = 0.75,label=f'{acqname}')  
            plt.fill_between(nosamp_hf[:convsamp1-1].flatten(), (mpms_GPR[:convsamp1-1] - 0.5*mpmsU95_GPR[:convsamp1-1]).flatten(), 
                              (mpms_GPR[:convsamp1-1] + 0.5*mpmsU95_GPR[:convsamp1-1]).flatten(), facecolor=colsacq[mm], alpha=0.3)#,label=f'{acqname} U95%')
        else: 
            plt.plot(nosamp_hf,mpms_GPR,linestyle=stylesacq[mm],color=colsacq[mm],linewidth = 0.75,label=f'{acqname} (not converged)')  
            plt.fill_between(nosamp_hf.flatten(), (mpms_GPR - 0.5*mpmsU95_GPR).flatten(), 
                              (mpms_GPR + 0.5*mpmsU95_GPR).flatten(), facecolor=colsacq[mm], alpha=0.3)#,label=f'{acqname} U95%')
        if annon == 1:
            plt.plot([nosamp_hf[i_samp10],nosamp_hf[i_samp10]],[MPM_true_MCS,locan_gpr_unc[mm]],color='silver',linewidth=0.2)
            if mm == len(acqnames)-1:
                plt.text(nosamp_hf[i_samp10]+1,min_gpr_unc+(max_gpr_unc-min_gpr_unc)/20*(mm+2.5),f'10 samples',color='silver',fontsize=8)
            plt.text(nosamp_hf[i_samp10]+1,locan_gpr_10_unc[mm],f'{acqname}: {signgpr10}{gpr10}%',color=colsacq[mm],fontsize=8)
            if ~np.isnan(gprconv):
                plt.plot([nosamp_hf[convsamp1-1]-1,nosamp_hf[convsamp1-1]-1],[mpms_GPR[convsamp1-1],locan_gpr_unc[mm]],color=colsacq[mm],linewidth=0.2)
                plt.text(nosamp_hf[convsamp1-1],locan_gpr_unc[mm],f'converged\n{acqname}: {signgprconv}{gprconv}%',color=colsacq[mm],fontsize=8)
  
                   
        # MF-GPR
        if acqname != 'IVR':
            plt.figure(50000+bb*100)
            if mm == 0:
                plt.plot([0,sampmaxplot1],[MPM_true_MCS,MPM_true_MCS],linestyle=':',linewidth=1,color='black'); plt.text(0.85*sampmaxplot1,MPM_true_MCS+0.08,'True MCS',color='black')
            if ~np.isnan(mfgprconv):
                plt.plot(nosamp_mf[:convsamp2-1],mpms_MFGPR[:convsamp2-1],linestyle=stylesacq[mm],color=colsacq[mm],linewidth = 0.75,label=f'{acqname}')    
                plt.fill_between(nosamp_mf[:convsamp2-1].flatten(), (mpms_MFGPR[:convsamp2-1] - 0.5*mpmsU95_MFGPR[:convsamp2-1]).flatten(), 
                                  (mpms_MFGPR[:convsamp2-1] + 0.5*mpmsU95_MFGPR[:convsamp2-1]).flatten(), facecolor=colsacq[mm], alpha=0.3)#,label=f'{acqname} U95%')
            else:
                plt.plot(nosamp_mf,mpms_MFGPR,linestyle=stylesacq[mm],color=colsacq[mm],linewidth = 0.75,label=f'{acqname} (not converged)')  
                plt.fill_between(nosamp_mf.flatten(), (mpms_MFGPR - 0.5*mpmsU95_MFGPR).flatten(), 
                                  (mpms_MFGPR + 0.5*mpmsU95_MFGPR).flatten(), facecolor=colsacq[mm], alpha=0.3)#,label=f'{acqname} U95%')
            if annon == 1:
                plt.plot([nosamp_mf[i_samp10],nosamp_mf[i_samp10]],[MPM_true_MCS,locan_mfgpr_unc[mm]],color=colsacq[mm],linewidth=0.2)
                if mm == len(acqnames)-2:
                    plt.text(nosamp_mf[i_samp10]+1,min_mfgpr_unc+(max_mfgpr_unc-min_mfgpr_unc)/20*(mm+2.5),f'10 samples',color='silver',fontsize=8)
                plt.text(nosamp_mf[i_samp10]+1,locan_mfgpr_10_unc[mm],f'{acqname}: {signmfgpr10}{mfgpr10}%',color=colsacq[mm],fontsize=8)
                if ~np.isnan(mfgprconv):
                    plt.plot([nosamp_mf[convit2],nosamp_mf[convit2]],[mpms_MFGPR[convit2],locan_mfgpr_unc[mm]],color=colsacq[mm],linewidth=0.2)
                    plt.text(nosamp_mf[convit2],locan_mfgpr_unc[mm],f'converged\n{acqname}: {signmfgprconv}{mfgprconv}%',color=colsacq[mm],fontsize=8)
        
            
        
        #### Now plot the same until convergence, but without uncertainty band
        
        # # GPR
        # plt.figure(80000+bb*100)
        # if mm == 0:
        #     plt.plot([0,sampmaxplot1],[MPM_true_MCS,MPM_true_MCS],linestyle=':',linewidth=1,color='black'); plt.text(0.85*sampmaxplot1,MPM_true_MCS+0.08,'True MCS',color='black')
        # if ~np.isnan(gprconv):
        #     plt.plot(nosamp_hf[:convsamp1-1],mpms_GPR[:convsamp1-1],linestyle=stylesacq[mm],color=colsacq[mm],linewidth = 0.75,label=f'{acqname}')  
        #     # plt.fill_between(nosamp_hf[:convsamp1-1].flatten(), (mpms_GPR[:convsamp1-1] - 0.5*mpmsU95_GPR[:convsamp1-1]).flatten(), 
        #     #                   (mpms_GPR[:convsamp1-1] + 0.5*mpmsU95_GPR[:convsamp1-1]).flatten(), facecolor=colsacq[mm], alpha=0.3)#,label=f'{acqname} U95%')
        # else: 
        #     plt.plot(nosamp_hf,mpms_GPR,linestyle=stylesacq[mm],color=colsacq[mm],linewidth = 0.75,label=f'{acqname} (not converged)')  
        #     # plt.fill_between(nosamp_hf.flatten(), (mpms_GPR - 0.5*mpmsU95_GPR).flatten(), 
        #     #                   (mpms_GPR + 0.5*mpmsU95_GPR).flatten(), facecolor=colsacq[mm], alpha=0.3)#,label=f'{acqname} U95%')
        # if annon == 1:
        #     plt.plot([nosamp_hf[i_samp10],nosamp_hf[i_samp10]],[MPM_true_MCS,locan_gpr_nounc[mm]],color='silver',linewidth=0.2)
        #     if mm == len(acqnames)-1:
        #         plt.text(nosamp_hf[i_samp10]+1,min_gpr_nounc+(max_gpr_nounc-min_gpr_nounc)/20*(mm+2.5),f'10 samples',color='silver',fontsize=8)
        #     plt.text(nosamp_hf[i_samp10]+1,locan_gpr_10_nounc[mm],f'{acqname}: {signgpr10}{gpr10}%',color=colsacq[mm],fontsize=8)
        #     if ~np.isnan(gprconv):
        #         plt.plot([nosamp_hf[convsamp1-1]-1,nosamp_hf[convsamp1-1]-1],[mpms_GPR[convsamp1-1],locan_gpr_nounc[mm]],color=colsacq[mm],linewidth=0.2)
        #         plt.text(nosamp_hf[convsamp1-1],locan_gpr_nounc[mm],f'converged\n{acqname}: {signgprconv}{gprconv}%',color=colsacq[mm],fontsize=8)
                   
        # # MF-GPR
        # if acqname != 'IVR':
        #     plt.figure(90000+bb*100)
        #     if mm == 0:
        #         plt.plot([0,sampmaxplot1],[MPM_true_MCS,MPM_true_MCS],linestyle=':',linewidth=1,color='black'); plt.text(0.85*sampmaxplot1,MPM_true_MCS+0.08,'True MCS',color='black')
        #     if ~np.isnan(mfgprconv):
        #         plt.plot(nosamp_mf[:convsamp2-1],mpms_MFGPR[:convsamp2-1],linestyle=stylesacq[mm],color=colsacq[mm],linewidth = 0.75,label=f'{acqname}')    
        #         # plt.fill_between(nosamp_mf[:convsamp2-1].flatten(), (mpms_MFGPR[:convsamp2-1] - 0.5*mpmsU95_MFGPR[:convsamp2-1]).flatten(), 
        #         #                   (mpms_MFGPR[:convsamp2-1] + 0.5*mpmsU95_MFGPR[:convsamp2-1]).flatten(), facecolor=colsacq[mm], alpha=0.3)#,label=f'{acqname} U95%')
        #     else:
        #         plt.plot(nosamp_mf,mpms_MFGPR,linestyle=stylesacq[mm],color=colsacq[mm],linewidth = 0.75,label=f'{acqname} (not converged)')  
        #         # plt.fill_between(nosamp_mf.flatten(), (mpms_MFGPR - 0.5*mpmsU95_MFGPR).flatten(), 
        #         #                   (mpms_MFGPR + 0.5*mpmsU95_MFGPR).flatten(), facecolor=colsacq[mm], alpha=0.3)#,label=f'{acqname} U95%')
        #     if annon == 1:
        #         plt.plot([nosamp_mf[i_samp10],nosamp_mf[i_samp10]],[MPM_true_MCS,locan_mfgpr_nounc[mm]],color=colsacq[mm],linewidth=0.2)
        #         if mm == len(acqnames)-2:
        #             plt.text(nosamp_mf[i_samp10]+1,min_mfgpr_nounc+(max_mfgpr_nounc-min_mfgpr_nounc)/20*(mm+2.5),f'10 samples',color='silver',fontsize=8)
        #         plt.text(nosamp_mf[i_samp10]+1,locan_mfgpr_10_nounc[mm],f'{acqname}: {signmfgpr10}{mfgpr10}%',color=colsacq[mm],fontsize=8)
        #         if ~np.isnan(mfgprconv):
        #             plt.plot([nosamp_mf[convit2],nosamp_mf[convit2]],[mpms_MFGPR[convit2],locan_mfgpr_nounc[mm]],color=colsacq[mm],linewidth=0.2)
        #             plt.text(nosamp_mf[convit2],locan_mfgpr_nounc[mm],f'converged\n{acqname}: {signmfgprconv}{mfgprconv}%',color=colsacq[mm],fontsize=8)
        
        
        
        #### Similar plots but visible in black/white (and without uncertainty band)
        
        # # GPR
        # plt.figure(60000+bb*100)
        # if mm == 0:
        #     plt.plot([0,sampmaxplot1],[MPM_true_MCS,MPM_true_MCS],linestyle=':',linewidth=1,color='black'); plt.text(0.85*sampmaxplot1,MPM_true_MCS+0.08,'True MCS',color='black')
        # if ~np.isnan(gprconv):
        #     plt.plot(nosamp_hf[:convsamp1-1],mpms_GPR[:convsamp1-1],linestyle=stylesacq[mm],color='k',linewidth = 0.75,label=f'{acqname}')  
        #     # plt.plot(nosamp_hf[:convsamp1-1],mpms_GPR[:convsamp1-1] - 0.5*mpmsU95_GPR[:convsamp1-1],linestyle=stylesacq[mm],color='k',linewidth = 0.5)  
        #     # plt.plot(nosamp_hf[:convsamp1-1],mpms_GPR[:convsamp1-1] + 0.5*mpmsU95_GPR[:convsamp1-1],linestyle=stylesacq[mm],color='k',linewidth = 0.5)  
        #     # plt.fill_between(nosamp_hf[:convsamp1-1].flatten(), (mpms_GPR[:convsamp1-1] - 0.5*mpmsU95_GPR[:convsamp1-1]).flatten(), 
        #     #                   (mpms_GPR[:convsamp1-1] + 0.5*mpmsU95_GPR[:convsamp1-1]).flatten(), facecolor=colsacq[mm], alpha=0.3)#,label=f'{acqname} U95%')
        # else: 
        #     plt.plot(nosamp_hf,mpms_GPR,linestyle=stylesacq[mm],color='k',linewidth = 0.75,label=f'{acqname} (not converged)')  
        #     # plt.plot(nosamp_hf,mpms_GPR - 0.5*mpmsU95_GPR,linestyle=stylesacq[mm],color='k',linewidth = 0.5)  
        #     # plt.plot(nosamp_hf,mpms_GPR + 0.5*mpmsU95_GPR,linestyle=stylesacq[mm],color='k',linewidth = 0.5)  
        #     # plt.fill_between(nosamp_hf.flatten(), (mpms_GPR - 0.5*mpmsU95_GPR).flatten(), 
        #     #                   (mpms_GPR + 0.5*mpmsU95_GPR).flatten(), facecolor=colsacq[mm], alpha=0.3)#,label=f'{acqname} U95%')
        # # plt.xticks(np.arange(0,sampmaxplot,sampstep,dtype=int))
        # if annon == 1:
        #     plt.plot([nosamp_hf[i_samp10],nosamp_hf[i_samp10]],[MPM_true_MCS,locan_gpr_nounc[mm]],color='silver',linewidth=0.2)
        #     if mm == len(acqnames)-1:
        #         plt.text(nosamp_hf[i_samp10]+1,locan_gpr_10_nounc[mm],f'10 samples\n{acqname}: {signgpr10}{gpr10}%',color='silver',fontsize=8)
        #     else:
        #         plt.text(nosamp_hf[i_samp10]+1,locan_gpr_10_nounc[mm],f'{acqname}: {signgpr10}{gpr10}%',color='silver',fontsize=8)
        #     if ~np.isnan(gprconv):
        #         plt.plot([nosamp_hf[convsamp1-1]-1,nosamp_hf[convsamp1-1]-1],[mpms_GPR[convsamp1-1],locan_gpr_nounc[mm]],color='silver',linewidth=0.2)
        #         plt.text(nosamp_hf[convsamp1-1],locan_gpr_nounc[mm],f'converged\n{acqname}: {signgprconv}{gprconv}%',color='silver',fontsize=8)
        
        # # MF-GPR
        # if acqname != 'IVR':
        #     plt.figure(70000+bb*100)
        #     if mm == 0:
        #         plt.plot([0,sampmaxplot1],[MPM_true_MCS,MPM_true_MCS],linestyle=':',linewidth=1,color='black'); plt.text(0.85*sampmaxplot1,MPM_true_MCS+0.08,'True MCS',color='black')
        #     if ~np.isnan(mfgprconv):
        #         plt.plot(nosamp_mf[:convsamp2-1],mpms_MFGPR[:convsamp2-1],linestyle=stylesacq[mm],color='k',linewidth = 0.75,label=f'{acqname}')    
        #         # plt.plot(nosamp_mf[:convsamp2-1],mpms_MFGPR[:convsamp2-1] - 0.5*mpmsU95_MFGPR[:convsamp2-1],linestyle=stylesacq[mm],color='k',linewidth = 0.5)   
        #         # plt.plot(nosamp_mf[:convsamp2-1],mpms_MFGPR[:convsamp2-1] + 0.5*mpmsU95_MFGPR[:convsamp2-1],linestyle=stylesacq[mm],color='k',linewidth = 0.5)
        #         # plt.fill_between(nosamp_mf[:convsamp2-1].flatten(), (mpms_MFGPR[:convsamp2-1] - 0.5*mpmsU95_MFGPR[:convsamp2-1]).flatten(), 
        #         #                   (mpms_MFGPR[:convsamp2-1] + 0.5*mpmsU95_MFGPR[:convsamp2-1]).flatten(), facecolor=colsacq[mm], alpha=0.3)#,label=f'{acqname} U95%')
        #     else:
        #         plt.plot(nosamp_mf,mpms_MFGPR,linestyle=stylesacq[mm],color='k',linewidth = 0.75,label=f'{acqname} (not converged)')  
        #         # plt.plot(nosamp_mf,mpms_MFGPR - 0.5*mpmsU95_MFGPR,linestyle=stylesacq[mm],color='k',linewidth = 0.5)  
        #         # plt.plot(nosamp_mf,mpms_MFGPR + 0.5*mpmsU95_MFGPR,linestyle=stylesacq[mm],color='k',linewidth = 0.5)  
        #         # plt.fill_between(nosamp_mf.flatten(), (mpms_MFGPR - 0.5*mpmsU95_MFGPR).flatten(), 
        #         #                   (mpms_MFGPR + 0.5*mpmsU95_MFGPR).flatten(), facecolor=colsacq[mm], alpha=0.3)#,label=f'{acqname} U95%')
        #     # plt.xticks(np.arange(0,sampmaxplot,sampstep,dtype=int))
        #     if annon == 1:
        #         plt.plot([nosamp_mf[i_samp10],nosamp_mf[i_samp10]],[MPM_true_MCS,locan_mfgpr_nounc[mm]],color='silver',linewidth=0.2)
        #         if mm == len(acqnames)-2:
        #             plt.text(nosamp_mf[i_samp10]+1,locan_mfgpr_10_nounc[mm],f'10 samples\n{acqname}: {signmfgpr10}{mfgpr10}%',color='silver',fontsize=8)
        #         else:
        #             plt.text(nosamp_mf[i_samp10]+1,locan_mfgpr_10_nounc[mm],f'{acqname}: {signmfgpr10}{mfgpr10}%',color='silver',fontsize=8)
        #         if ~np.isnan(mfgprconv):
        #             plt.plot([nosamp_mf[convit2],nosamp_mf[convit2]],[mpms_MFGPR[convit2],locan_mfgpr_nounc[mm]],color='silver',linewidth=0.2)
        #             plt.text(nosamp_mf[convit2],locan_mfgpr_nounc[mm],f'converged\n{acqname}: {signmfgprconv}{mfgprconv}%',color='silver',fontsize=8)
       
        
        
        #### Also plot the result including uncertainty, for each acquisition function separately in black / white
        
        # # GPR
        # plt.figure(figsize=(7,3))
        # plt.plot([0,sampmaxplot1],[MPM_true_MCS,MPM_true_MCS],linestyle=':',linewidth=1,color='grey'); plt.text(0.85*sampmaxplot1,MPM_true_MCS+0.08,'True MCS',color='grey')
        # if ~np.isnan(gprconv):
        #     plt.plot(nosamp_hf[:convsamp1-1],mpms_GPR[:convsamp1-1],linestyle=stylesacq[mm],color='k',linewidth = 0.75,label=f'{acqname}')  
        #     plt.fill_between(nosamp_hf[:convsamp1-1].flatten(), (mpms_GPR[:convsamp1-1] - 0.5*mpmsU95_GPR[:convsamp1-1]).flatten(), 
        #                       (mpms_GPR[:convsamp1-1] + 0.5*mpmsU95_GPR[:convsamp1-1]).flatten(), facecolor='lightgrey', alpha=0.3)#,label=f'{acqname} U95%')
        # else: 
        #     plt.plot(nosamp_hf,mpms_GPR,linestyle=stylesacq[mm],color='k',linewidth = 0.75,label=f'{acqname} (not converged)')  
        #     plt.fill_between(nosamp_hf.flatten(), (mpms_GPR - 0.5*mpmsU95_GPR).flatten(), 
        #                       (mpms_GPR + 0.5*mpmsU95_GPR).flatten(), facecolor='lightgrey', alpha=0.3)#,label=f'{acqname} U95%')
        # if annon == 1:
        #     plt.plot([nosamp_hf[i_samp10],nosamp_hf[i_samp10]],[MPM_true_MCS,locan_gpr_uncsep],color='silver',linewidth=0.2)
        #     plt.text(nosamp_hf[i_samp10]+1,locan_gpr_uncsep,f'10 samples\n{acqname}: {signgpr10}{gpr10}%',color='silver',fontsize=8)
        #     if ~np.isnan(gprconv):
        #         plt.plot([nosamp_hf[convsamp1-1]-1,nosamp_hf[convsamp1-1]-1],[mpms_GPR[convsamp1-1],locan_gpr_uncsep],color='silver',linewidth=0.2)
        #         plt.text(nosamp_hf[convsamp1-1],locan_gpr_uncsep,f'converged\n{acqname}: {signgprconv}{gprconv}%',color='silver',fontsize=8)
        # plt.ylabel(ylabs)
        # plt.xlabel('Number of samples $j$ [-]')
        # plt.xticks(ticks=x_ticks2)
        # plt.ylim([min_gpr_uncsep,max_gpr_uncsep])
        # plt.xlim([N_hf-1,sampmaxplot1])
        # # plt.title(f'Adaptive Screening with GPR, {indcase} and acquisition func. {acqname}')
        # plt.title(f'Adaptive Screening with GPR, {indcase}')
        # plt.legend(ncol=len(acqnames))
        # if annon == 1:
        #     plt.savefig(f'{out_figs}/{plotcase}_Convergence_GPR-{sampopt_txt}_{acqname}_bw_ann.pdf',bbox_inches='tight')
        # else:
        #     plt.savefig(f'{out_figs}/{plotcase}_Convergence_GPR-{sampopt_txt}_{acqname}_bw.pdf',bbox_inches='tight')

        # # MF-GPR
        # if acqname != 'IVR':
        #     plt.figure(figsize=(7,3))
        #     plt.plot([0,sampmaxplot1],[MPM_true_MCS,MPM_true_MCS],linestyle=':',linewidth=1,color='grey'); plt.text(0.85*sampmaxplot1,MPM_true_MCS+0.08,'True MCS',color='grey')
        #     if ~np.isnan(mfgprconv):
        #         plt.plot(nosamp_mf[:convsamp2-1],mpms_MFGPR[:convsamp2-1],linestyle=stylesacq[mm],color='k',linewidth = 0.75,label=f'{acqname}')    
        #         plt.fill_between(nosamp_mf[:convsamp2-1].flatten(), (mpms_MFGPR[:convsamp2-1] - 0.5*mpmsU95_MFGPR[:convsamp2-1]).flatten(), 
        #                           (mpms_MFGPR[:convsamp2-1] + 0.5*mpmsU95_MFGPR[:convsamp2-1]).flatten(), facecolor='lightgrey', alpha=0.3)#,label=f'{acqname} U95%')
        #     else:
        #         plt.plot(nosamp_mf,mpms_MFGPR,linestyle=stylesacq[mm],color='k',linewidth = 0.75,label=f'{acqname} (not converged)')  
        #         plt.fill_between(nosamp_mf.flatten(), (mpms_MFGPR - 0.5*mpmsU95_MFGPR).flatten(), 
        #                           (mpms_MFGPR + 0.5*mpmsU95_MFGPR).flatten(), facecolor='lightgrey', alpha=0.3)#,label=f'{acqname} U95%')
        #     if annon == 1:
        #         plt.plot([nosamp_mf[i_samp10],nosamp_mf[i_samp10]],[MPM_true_MCS,locan_mfgpr_uncsep],color='silver',linewidth=0.2)
        #         plt.text(nosamp_mf[i_samp10]+1,locan_mfgpr_uncsep,f'10 samples\n{acqname}: {signmfgpr10}{mfgpr10}%',color='silver',fontsize=8)
        #         if ~np.isnan(mfgprconv):
        #             plt.plot([nosamp_mf[convit2],nosamp_mf[convit2]],[mpms_MFGPR[convit2],locan_mfgpr_uncsep],color='silver',linewidth=0.2)
        #             plt.text(nosamp_mf[convit2],locan_mfgpr_uncsep,f'converged\n{acqname}: {signmfgprconv}{mfgprconv}%',color='silver',fontsize=8)
            # plt.ylabel(ylabs)
            # plt.xlabel('Number of samples $j$ [-]')
            # plt.xticks(ticks=x_ticks2)
            # plt.ylim([min_gpr_uncsep,max_gpr_uncsep])
            # plt.xlim([N_hf-1,sampmaxplot1])
            # # plt.title(f'Adaptive Screening with MF-GPR, {indcase} and acquisition func. {acqname}')
            # plt.title(f'Adaptive Screening with MF-GPR, {indcase}')
            # plt.legend(ncol=len(acqnames))
        #     if annon == 1:
        #         plt.savefig(f'{out_figs}/{plotcase}_Convergence_MFGPR-{sampopt_txt}_{acqname}_bw_ann.pdf',bbox_inches='tight')
        #     else:
        #         plt.savefig(f'{out_figs}/{plotcase}_Convergence_MFGPR-{sampopt_txt}_{acqname}_bw.pdf',bbox_inches='tight')
                
            
        
        #### Also plot the result including uncertainty, for each acquisition function separately in color
        
        # GPR
        plt.figure(figsize=(7,3))
        plt.plot([0,sampmaxplot1],[MPM_true_MCS,MPM_true_MCS],linestyle=':',linewidth=1,color='grey'); plt.text(0.85*sampmaxplot1,MPM_true_MCS+0.08,'True MCS',color='grey')
        if ~np.isnan(gprconv):
            plt.plot(nosamp_hf[:convsamp1-1],mpms_GPR[:convsamp1-1],linestyle=stylesacq[mm],color=colsacq[mm],linewidth = 0.75,label=f'{acqname}')  
            plt.fill_between(nosamp_hf[:convsamp1-1].flatten(), (mpms_GPR[:convsamp1-1] - 0.5*mpmsU95_GPR[:convsamp1-1]).flatten(), 
                              (mpms_GPR[:convsamp1-1] + 0.5*mpmsU95_GPR[:convsamp1-1]).flatten(), facecolor=colsacq[mm], alpha=0.3)#,label=f'{acqname} U95%')
        else: 
            plt.plot(nosamp_hf,mpms_GPR,linestyle=stylesacq[mm],color=colsacq[mm],linewidth = 0.75,label=f'{acqname} (not converged)')  
            plt.fill_between(nosamp_hf.flatten(), (mpms_GPR - 0.5*mpmsU95_GPR).flatten(), 
                              (mpms_GPR + 0.5*mpmsU95_GPR).flatten(), facecolor=colsacq[mm], alpha=0.3)#,label=f'{acqname} U95%')
        if annon == 1:
            plt.plot([nosamp_hf[i_samp10],nosamp_hf[i_samp10]],[MPM_true_MCS,locan_gpr_uncsep],color=colsacq[mm],linewidth=0.2)
            plt.text(nosamp_hf[i_samp10]+1,locan_gpr_uncsep,f'10 samples\n{acqname}: {signgpr10}{gpr10}%',color=colsacq[mm],fontsize=8)
            if ~np.isnan(gprconv):
                plt.plot([nosamp_hf[convsamp1-1]-1,nosamp_hf[convsamp1-1]-1],[mpms_GPR[convsamp1-1],locan_gpr_uncsep],color=colsacq[mm],linewidth=0.2)
                plt.text(nosamp_hf[convsamp1-1],locan_gpr_uncsep,f'converged\n{acqname}: {signgprconv}{gprconv}%',color=colsacq[mm],fontsize=8)
        plt.ylabel(ylabs)
        plt.xlabel('Number of samples $j$ [-]')
        plt.xticks(ticks=x_ticks2)
        plt.ylim([min_gpr_uncsep,max_gpr_uncsep])
        plt.xlim([N_hf-1,sampmaxplot1])
        # plt.title(f'Adaptive Screening with GPR, {indcase} and acquisition func. {acqname}')
        plt.title(f'Adaptive Screening with GPR, {indcase}')
        plt.legend(ncol=len(acqnames))
        if annon == 1:
            plt.savefig(f'{out_figs}/{plotcase}_Convergence_GPR-{sampopt_txt}_{acqname}_col_ann.pdf',bbox_inches='tight')
        else:
            plt.savefig(f'{out_figs}/{plotcase}_Convergence_GPR-{sampopt_txt}_{acqname}_col.pdf',bbox_inches='tight')
        plt.close()
        
        # MF-GPR
        if acqname != 'IVR':
            plt.figure(figsize=(7,3))
            plt.plot([0,sampmaxplot1],[MPM_true_MCS,MPM_true_MCS],linestyle=':',linewidth=1,color=colsacq[mm]); plt.text(0.85*sampmaxplot1,MPM_true_MCS+0.08,'True MCS',color='grey')
            if ~np.isnan(mfgprconv):
                plt.plot(nosamp_mf[:convsamp2-1],mpms_MFGPR[:convsamp2-1],linestyle=stylesacq[mm],color=colsacq[mm],linewidth = 0.75,label=f'{acqname}')    
                plt.fill_between(nosamp_mf[:convsamp2-1].flatten(), (mpms_MFGPR[:convsamp2-1] - 0.5*mpmsU95_MFGPR[:convsamp2-1]).flatten(), 
                                  (mpms_MFGPR[:convsamp2-1] + 0.5*mpmsU95_MFGPR[:convsamp2-1]).flatten(), facecolor=colsacq[mm], alpha=0.3)#,label=f'{acqname} U95%')
            else:
                plt.plot(nosamp_mf,mpms_MFGPR,linestyle=stylesacq[mm],color=colsacq[mm],linewidth = 0.75,label=f'{acqname} (not converged)')  
                plt.fill_between(nosamp_mf.flatten(), (mpms_MFGPR - 0.5*mpmsU95_MFGPR).flatten(), 
                                  (mpms_MFGPR + 0.5*mpmsU95_MFGPR).flatten(), facecolor=colsacq[mm], alpha=0.3)#,label=f'{acqname} U95%')
            if annon == 1:
                plt.plot([nosamp_mf[i_samp10],nosamp_mf[i_samp10]],[MPM_true_MCS,locan_mfgpr_uncsep],color=colsacq[mm],linewidth=0.2)
                plt.text(nosamp_mf[i_samp10]+1,locan_mfgpr_uncsep,f'10 samples\n{acqname}: {signmfgpr10}{mfgpr10}%',color=colsacq[mm],fontsize=8)
                if ~np.isnan(mfgprconv):
                    plt.plot([nosamp_mf[convit2],nosamp_mf[convit2]],[mpms_MFGPR[convit2],locan_mfgpr_uncsep],color=colsacq[mm],linewidth=0.2)
                    plt.text(nosamp_mf[convit2],locan_mfgpr_uncsep,f'converged\n{acqname}: {signmfgprconv}{mfgprconv}%',color=colsacq[mm],fontsize=8)
            plt.ylabel(ylabs)
            plt.xlabel('Number of samples $j$ [-]')
            plt.xticks(ticks=x_ticks2)
            plt.ylim([min_gpr_uncsep,max_gpr_uncsep])
            plt.xlim([N_hf-1,sampmaxplot1])
            # plt.title(f'Adaptive Screening with MF-GPR, {indcase} and acquisition func. {acqname}')
            plt.title(f'Adaptive Screening with MF-GPR, {indcase}')
            plt.legend(ncol=len(acqnames))
            if annon == 1:
                plt.savefig(f'{out_figs}/{plotcase}_Convergence_MFGPR-{sampopt_txt}_{acqname}_col_ann.pdf',bbox_inches='tight')
            else:
                plt.savefig(f'{out_figs}/{plotcase}_Convergence_MFGPR-{sampopt_txt}_{acqname}_col.pdf',bbox_inches='tight')
            plt.close()
        
        
        #### CALCULATE AND PRINT MPM VALUES AND U95% FOR CONVERGED RESULTS
        
        print(f'Case: {indcase}, acquisition function: {acqname}, {dur/3600}h:')
        print(f'MPM from true HF MCS: {MPM_true_MCS} m:')
        print(f'MPM from true LF MCS: {MPM_true_MCS_lf} m')
        
        # GPR
        if ~np.isnan(gprconv):
            conv_mpm_gpr[bb,mm] = mpms_GPR[convit1]
            conv_mpmu95_gpr[bb,mm] = mpmsU95_GPR[convit1]
        else:
            conv_mpm_gpr[bb,mm] = np.nan
            conv_mpmu95_gpr[bb,mm] = np.nan
        print(f'MPM GPR {acqname}: {conv_mpm_gpr[bb,mm]} m (w.r.t. MCS {(conv_mpm_gpr[bb,mm]/MPM_true_MCS-1)*100}%), U95 {conv_mpmu95_gpr[bb,mm]} m')
        
        # Assemble number of samples for convergence
        M1_gpr[bb,mm] = convsamp1
        
        # MF-GPR
        if acqname != 'IVR':
            if ~np.isnan(mfgprconv):
                conv_mpm_mfgpr[bb,mm] = mpms_MFGPR[convit2]
                conv_mpmu95_mfgpr[bb,mm] = mpmsU95_MFGPR[convit2]
            else:
                conv_mpm_mfgpr[bb,mm] = np.nan
                conv_mpmu95_mfgpr[bb,mm] = np.nan
            print(f'MPM MF-GPR {acqname}: {conv_mpm_mfgpr[bb,mm]} m (w.r.t. MCS {(conv_mpm_mfgpr[bb,mm]/MPM_true_MCS-1)*100}%), U95 {conv_mpmu95_mfgpr[bb,mm]} m')
            # Assemble number of samples for convergence
            M1_mfgpr[bb,mm] = convsamp2
        
        ############### READ AND PLOT CONVERGED GPR RESULTS FOR PRESENT CASE & ACQ FUNCT
        if ~np.isnan(gprconv):
            # GPR (2x)
            DatHFtrainIt1 = np.genfromtxt(f'{in_files}/HFGPR_TrainData_{indcase}_Set-{HFkern}-nf_{txtadd}_HFacq-{acqname}_Nhf{N_hf}_HFit{convit1}.csv', skip_header=1, delimiter=',')
            v_train_h1 = DatHFtrainIt1[:,0]
            p_train_h1 = DatHFtrainIt1[:,1]
            
            DatHFpredIt1 = np.genfromtxt(f'{in_files}/HFGPR_HFpred_{indcase}_Set-{HFkern}-nf_{txtadd}_HFacq-{acqname}_Nhf{N_hf}_HFit{convit1}.csv', skip_header=1, delimiter=',')
            Prange_predHF = DatHFpredIt1[:,0]
            mean_high_gp_model = DatHFpredIt1[:,1]
            var_high_gp_model = DatHFpredIt1[:,2]
            U95_high_gp_model = 1.96*np.sqrt(var_high_gp_model)
            
            ###### Plot HF
            # Plot the posterior mean and variance (U95) from the reference HF gaussian process
            plt.figure(1000+100*mm+bb,figsize=(fsize_h,fsize_v))
            plt.yscale('log')
            plt.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None',label='HF "true" MCS')
            plt.plot(v_train_h1,p_train_h1, color='r', marker='d',linestyle='None', label=f'{convsamp1} HF samples')
            plt.plot(mean_high_gp_model,Prange_predHF,'-.',color='g',label='GPR prediction')
            plt.fill_betweenx(Prange_predHF.flatten(), (mean_high_gp_model - 0.5*U95_high_gp_model).flatten(), 
                              (mean_high_gp_model + 0.5*U95_high_gp_model).flatten(), facecolor='g', alpha=0.1,label='GPR prediction U95%')
            plt.plot([0,maxVal],[pexp,pexp],':',color='black')
            plt.text(0.9*maxVal,pexp*1.4,'$P_{exp}$')
            plt.ylabel('P(C\'\'$\geq$c) [-]')
            plt.xlabel('c [m]')
            plt.ylim([minP,1])
            plt.xlim([0,maxVal])
            plt.legend()
            plt.savefig(f'{out_figs}/{plotcase}_Prediction_HFGPR_Set-{HFkern}_{sampopt_txt}_{acqname}_{convsamp1}HFsamps.pdf',bbox_inches='tight')
            plt.close()
            
        ############### READ AND PLOT CONVERGED MF-GPR RESULTS FOR PRESENT CASE & ACQ FUNCT
        if acqname != 'IVR':
            if ~np.isnan(mfgprconv):
                
                # MF-GPR (4x)
                DatLFtrainIt2 = np.genfromtxt(f'{in_files}/MFGPR_LFTrainData_{indcase}_Set-{MFkern1}-{MFkern2}-nf_{txtadd}_MFacq-{acqname}_Nhf{N_hf}_MFit{convit2}.csv', skip_header=1, delimiter=',')
                v_train_l2 = DatLFtrainIt2[:,0]
                p_train_l2 = DatLFtrainIt2[:,1]
                
                DatHFtrainIt2 = np.genfromtxt(f'{in_files}/MFGPR_HFTrainData_{indcase}_Set-{MFkern1}-{MFkern2}-nf_{txtadd}_MFacq-{acqname}_Nhf{N_hf}_MFit{convit2}.csv', skip_header=1, delimiter=',')
                v_train_h2 = DatHFtrainIt2[:,0]
                p_train_h2 = DatHFtrainIt2[:,1]
                        
                DatLFpredIt2 = np.genfromtxt(f'{in_files}/MFGPR_LFpred_{indcase}_Set-{MFkern1}-{MFkern2}-nf_{txtadd}_MFacq-{acqname}_Nhf{N_hf}_MFit{convit2}.csv', skip_header=1, delimiter=',')
                Prange_predLF = DatLFpredIt2[:,0]
                mean_low_mfgp_model = DatLFpredIt2[:,1]
                var_low_mfgp_model = DatLFpredIt2[:,2]
                U95_low_mfgp_model = 1.96*np.sqrt(var_low_mfgp_model)
                
                DatHFpredIt2 = np.genfromtxt(f'{in_files}/MFGPR_HFpred_{indcase}_Set-{MFkern1}-{MFkern2}-nf_{txtadd}_MFacq-{acqname}_Nhf{N_hf}_MFit{convit2}.csv', skip_header=1, delimiter=',')
                Prange_predHF = DatHFpredIt2[:,0]
                mean_high_mfgp_model = DatHFpredIt2[:,1]
                var_high_mfgp_model = DatHFpredIt2[:,2]
                U95_high_mfgp_model = 1.96*np.sqrt(var_high_mfgp_model)
                
                # # Debugging
                # plt.figure()
                # plt.yscale('log')
                # plt.plot(v_true_l_sampled,p_true_l_sampled,marker='.',color=(0.7,0.7,0.7),linestyle='None',label='LF "true" MCS')
                # plt.plot(v_train_l2,p_train_l2, color='grey', marker='o',linestyle='None', label='LF samples')
                
                ###### Plot MF
                # Plot the posterior mean and variance (U95) from the MF gaussian process
                plt.figure(2000+100*mm+bb,figsize=(fsize_h,fsize_v))
                plt.yscale('log')
                plt.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None',label='HF "true" MCS (validation)')
                plt.plot(v_train_l2,p_train_l2, color='grey', marker='o',linestyle='None', label='LF samples')
                plt.plot(v_train_h2,p_train_h2, color='r', marker='d',linestyle='None', label=f'{convsamp2} HF samples')
                plt.plot(mean_low_mfgp_model,Prange_predLF,'--', color='b',label='MF-GPR predictions')
                plt.fill_betweenx(Prange_predLF.flatten(), (mean_low_mfgp_model - 0.5*U95_low_mfgp_model).flatten(), 
                                  (mean_low_mfgp_model + 0.5*U95_low_mfgp_model).flatten(), facecolor='b', alpha=0.3,label='MF-GPR predictions U95%')
                plt.plot(mean_high_mfgp_model,Prange_predHF, '--', color='b',label='MF-GPR prediction HF (label not used)')
                plt.fill_betweenx(Prange_predHF.flatten(), (mean_high_mfgp_model - 0.5*U95_high_mfgp_model).flatten(), 
                                  (mean_high_mfgp_model + 0.5*U95_high_mfgp_model).flatten(), facecolor='b', alpha=0.3,label='MF-GPR prediction HF U95% (label not printed)')
                plt.plot([0,maxVal],[pexp,pexp],':',color='black')
                plt.text(0.9*maxVal,pexp*1.4,'$P_{exp}$')
                plt.ylabel('P(C\'\'$\geq$c) [-]')
                plt.xlabel('c [m]')
                plt.ylim([minP,1])
                plt.xlim([0,maxVal])
                handles, labels = plt.gca().get_legend_handles_labels() # get handles and labels (in order to change legend order)
                order = [0,1,2,3,4] #specify order of items in legend
                plt.legend([handles[idx] for idx in order],[labels[idx] for idx in order]) #add legend to plot
                plt.savefig(f'{out_figs}/{plotcase}_Prediction_MFGPR_Set-{MFkern1}-{MFkern2}_{sampopt_txt}_{acqname}_{convsamp2}HFsamps_A.pdf',bbox_inches='tight')
                plt.close()
                
                # Alternative side-by-side plot
                plt.figure(3000+100*mm+bb,figsize=(fsize_h*1.7,fsize_v*0.8))
                # -- subplot 2 (HF data)
                plt.subplot(1,2,2)
                plt.yscale('log') 
                plt.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None',label='HF "true" MCS (validation)')
                plt.plot(v_train_h2,p_train_h2, color='r', marker='d',linestyle='None',markersize=4, label=f'{convsamp2} HF samples')
                plt.plot(mean_high_mfgp_model,Prange_predHF, '--', color='b',label='MF-GPR pred. HF')
                plt.fill_betweenx(Prange_predHF.flatten(), (mean_high_mfgp_model - 0.5*U95_high_mfgp_model).flatten(), 
                                 (mean_high_mfgp_model + 0.5*U95_high_mfgp_model).flatten(), facecolor='b', alpha=0.1,label='MF-GPR pred. HF U95%')
                plt.plot([0,maxVal],[pexp,pexp],':',color='black') #,label='$P_{exp}$')
                plt.text(0.9*maxVal,pexp*1.4,'$P_{exp}$')
                plt.ylabel('P(C\'\'$\geq$c) [-]')
                plt.xlabel('c [m]')
                plt.xlim(left=0,right=maxVal)
                plt.ylim([minP,1])
                plt.legend(loc='best')
                # -- subplot1 (LF data)
                plt.subplot(1,2,1)
                plt.yscale('log') 
                plt.plot(v_train_l2,p_train_l2, color='grey', marker='o',linestyle='None',markersize=4, label='LF samples')
                plt.plot(mean_low_mfgp_model,Prange_predLF, '--', color='b',alpha=0.4,label='MF-GPR pred. LF')
                plt.fill_betweenx(Prange_predLF.flatten(), (mean_low_mfgp_model - 0.5*U95_low_mfgp_model).flatten(), 
                                  (mean_low_mfgp_model + 0.5*U95_low_mfgp_model).flatten(), facecolor='b', alpha=0.1,label='MF-GPR pred. LF U95%')
                plt.ylabel('P(C\'$\geq$c) [-]')
                plt.xlabel('c [m]')
                plt.xlim(left=0,right=maxVal_lf)
                # -- y-axis
                plt.ylim([minP,1])
                plt.legend(loc='best')
                plt.savefig(f'{out_figs}/{plotcase}_Prediction_MFGPR_Set-{MFkern1}-{MFkern2}_{sampopt_txt}_{acqname}_{convsamp2}HFsamps_B.pdf',bbox_inches='tight')
                plt.close()
           
    
        ###### Derive inputs for M2 and M3
        
        # Derive target exposure duration probability level, accounting for vessel speed and heading (preparation M2)
        Tpe = calc_T_enc(Tp,Vs,mu)
        ne_exp = dur/Tpe

        # Derive true MPM and predicted MPMs (preparation M2 and M3)
        MPMtrue[bb,mm] = MPM_from_DNR(v_true_h,p_true_h,ne_exp)
        MPMpredGPR[bb,mm] = MPM_from_DNR(mean_high_gp_model,Prange_predHF,ne_exp)
        MPMpredMFGPR[bb,mm] = MPM_from_DNR(mean_high_mfgp_model,Prange_predHF,ne_exp)
        
        # Interpolate true distribution on HF prediction range (preparation M3)
        v_true_h_Prange = np.interp(Prange_predHF,np.flipud(p_true_h),np.flipud(v_true_h))
        
        # M2 for GPR
        M2_gpr[bb,mm] = np.abs(MPMpredGPR[bb,mm]-MPMtrue[bb,mm])/MPMtrue[bb,mm]

        # M3 for GPR
        M3_gpr[bb,mm] = np.max(np.abs(mean_high_gp_model-v_true_h_Prange))/MPMtrue[bb,mm]        

        # M2 and M3 for MF-GPR
        if acqname != 'IVR':
            M2_mfgpr[bb,mm] = np.abs(MPMpredMFGPR[bb,mm]-MPMtrue[bb,mm])/MPMtrue[bb,mm]         
            M3_mfgpr[bb,mm] = np.max(np.abs(mean_high_mfgp_model-v_true_h_Prange))/MPMtrue[bb,mm]
        
    ############### Finalise and save convergence plot
    
    plt.figure(20000+bb*100)
    plt.xlabel('Number of samples $j$ [-]')
    plt.ylabel(ylabs)
    plt.xticks(ticks=x_ticks2)
    plt.title(f'Adaptive Screening with GPR, {indcase}')
    plt.legend(ncol=len(acqnames))
    plt.xlim([0,sampmaxplot2])
    plt.ylim([min_gpr_unc,max_gpr_unc])
    plt.savefig(f'{out_figs}/{plotcase}_ConvergenceAllIts_GPR-{sampopt_txt}.pdf',bbox_inches='tight')
    plt.close()
    
    plt.figure(30000+bb*100)
    plt.xlabel('Number of samples $j$ [-]')
    plt.ylabel(ylabs)
    plt.xticks(ticks=x_ticks2)
    plt.title(f'Adaptive Screening with GPR, {indcase}')
    plt.legend(ncol=len(acqnames))
    plt.xlim([0,sampmaxplot2])
    plt.ylim([min_gpr_unc,max_gpr_unc])
    plt.savefig(f'{out_figs}/{plotcase}_ConvergenceAllIts_MFGPR-{sampopt_txt}.pdf',bbox_inches='tight')
    plt.close()
    
    plt.figure(140000+bb*100)
    # plt.plot([0,sampmaxplot1],0.96*np.array([MPM_true_MCS,MPM_true_MCS]),linestyle=':',color='lightgrey',linewidth=0.3)
    # plt.plot([0,sampmaxplot1],1.04*np.array([MPM_true_MCS,MPM_true_MCS]),linestyle=':',color='lightgrey',linewidth=0.3)
    # plt.plot([0,sampmaxplot1],0.92*np.array([MPM_true_MCS,MPM_true_MCS]),linestyle=':',color='lightgrey',linewidth=0.3)
    # plt.plot([0,sampmaxplot1],1.08*np.array([MPM_true_MCS,MPM_true_MCS]),linestyle=':',color='lightgrey',linewidth=0.3)
    # if nn == 0:
    #     plt.ylabel('$\widehat{V_{hog}\'\'}$ [Nm]')
    #     plt.text(0.93*sampmaxplot1,0.96*MPM_true_MCS+0.03,'-4%',color='lightgrey',fontsize=8)
    #     plt.text(0.93*sampmaxplot1,1.04*MPM_true_MCS+0.03,'+4%',color='lightgrey',fontsize=8)
    #     plt.text(0.93*sampmaxplot1,0.92*MPM_true_MCS+0.03,'-8%',color='lightgrey',fontsize=8)
    #     plt.text(0.93*sampmaxplot1,1.08*MPM_true_MCS+0.03,'+8%',color='lightgrey',fontsize=8)
    plt.ylabel(ylabs)
    plt.xlabel('Number of samples $j$ [-]')
    plt.xticks(ticks=x_ticks1)
    # plt.title(f'Adaptive Screening with GPR, {indcase}')
    plt.legend(ncol=len(acqnames))
    plt.xlim([0,sampmaxplot1])
    plt.ylim([min_gpr_unc,max_gpr_unc])
    if annon == 1:
        plt.savefig(f'{out_figs}/{plotcase}_Convergence_GPR-{sampopt_txt}_ann.pdf',bbox_inches='tight')
    else:
        plt.savefig(f'{out_figs}/{plotcase}_Convergence_GPR-{sampopt_txt}.pdf',bbox_inches='tight')

    plt.figure(150000+bb*100)
    plt.ylabel(ylabs)
    plt.xlabel('Number of samples $j$ [-]')
    plt.xticks(ticks=x_ticks1)
    # plt.title(f'Adaptive Screening with MF-GPR, {indcase}')
    plt.legend(ncol=len(acqnames))
    plt.xlim([0,sampmaxplot1])
    plt.ylim([min_gpr_unc,max_gpr_unc])
    if annon == 1:
        plt.savefig(f'{out_figs}/{plotcase}_Convergence_MFGPR-{sampopt_txt}_ann.pdf',bbox_inches='tight')
    else:
        plt.savefig(f'{out_figs}/{plotcase}_Convergence_MFGPR-{sampopt_txt}.pdf',bbox_inches='tight')

    plt.figure(160000+bb*100)
    plt.subplot(211)
    plt.ylabel(ylabs)
    # plt.xlabel('Number of samples $j$ [-]')
    plt.xticks(ticks=x_ticks1,labels=[])
    plt.title('GPR')
    plt.legend(ncol=len(acqnames))
    plt.xlim([0,sampmaxplot1])
    plt.ylim([min_gpr_unc,max_gpr_unc])
    plt.subplot(212)
    plt.ylabel(ylabs)
    plt.xlabel('Number of samples $j$ [-]')
    plt.xticks(ticks=x_ticks1)
    plt.title('MF-GPR')
    # plt.legend(ncol=len(acqnames))
    plt.xlim([0,sampmaxplot1])
    plt.ylim([min_gpr_unc,max_gpr_unc])
    if annon == 1:
        plt.savefig(f'{out_figs}/{plotcase}_Convergence_{sampopt_txt}_ann.pdf',bbox_inches='tight')
    else:
        plt.savefig(f'{out_figs}/{plotcase}_Convergence_{sampopt_txt}.pdf',bbox_inches='tight')
    
    
    plt.figure(40000+bb*100)
    plt.ylabel(ylabs)
    plt.xlabel('Number of samples $j$ [-]')
    plt.xticks(ticks=x_ticks1)
    # plt.title(f'Adaptive Screening with GPR, {indcase}')
    plt.legend(ncol=len(acqnames))
    plt.xlim([0,sampmaxplot1])
    plt.ylim([min_gpr_unc,max_gpr_unc])
    if annon == 1:
        plt.savefig(f'{out_figs}/{plotcase}_Convergence_GPR-{sampopt_txt}_10samp_ann.pdf',bbox_inches='tight')
    else:
        plt.savefig(f'{out_figs}/{plotcase}_Convergence_GPR-{sampopt_txt}_10samp.pdf',bbox_inches='tight')
    plt.close()
    
    plt.figure(50000+bb*100)
    plt.ylabel(ylabs)
    plt.xlabel('Number of samples $j$ [-]')
    plt.xticks(ticks=x_ticks1)
    # plt.title(f'Adaptive Screening with MF-GPR, {indcase}')
    plt.legend(ncol=len(acqnames))
    plt.xlim([0,sampmaxplot1])
    plt.ylim([min_gpr_unc,max_gpr_unc])
    if annon == 1:
        plt.savefig(f'{out_figs}/{plotcase}_Convergence_MFGPR-{sampopt_txt}_10samp_ann.pdf',bbox_inches='tight')
    else:
        plt.savefig(f'{out_figs}/{plotcase}_Convergence_MFGPR-{sampopt_txt}_10samp.pdf',bbox_inches='tight')
    plt.close()
    
    # plt.figure(60000+bb*100)
    # plt.ylabel(ylabs)
    # plt.xlabel('Number of samples $j$ [-]')
    # plt.xticks(ticks=x_ticks1)
    # # plt.title(f'Adaptive Screening with GPR, {indcase}')
    # plt.legend(ncol=len(acqnames))
    # plt.xlim([N_hf-1,sampmaxplot1])
    # plt.ylim([min_gpr_nounc,max_gpr_nounc])
    # if annon == 1:
    #     plt.savefig(f'{out_figs}/{plotcase}_Convergence_GPR-{sampopt_txt}_bw_ann.pdf',bbox_inches='tight')
    # else:
    #     plt.savefig(f'{out_figs}/{plotcase}_Convergence_GPR-{sampopt_txt}_bw.pdf',bbox_inches='tight')

    # plt.figure(70000+bb*100)
    # plt.ylabel(ylabs)
    # plt.xlabel('Number of samples $j$ [-]')
    # plt.xticks(ticks=x_ticks1)
    # # plt.title(f'Adaptive Screening with MF-GPR, {indcase}')
    # plt.legend(ncol=len(acqnames))
    # plt.xlim([N_hf-1,sampmaxplot1])
    # plt.ylim([min_gpr_nounc,max_gpr_nounc])
    # if annon == 1:
    #     plt.savefig(f'{out_figs}/{plotcase}_Convergence_MFGPR-{sampopt_txt}_bw_ann.pdf',bbox_inches='tight')
    # else:
    #     plt.savefig(f'{out_figs}/{plotcase}_Convergence_MFGPR-{sampopt_txt}_bw.pdf',bbox_inches='tight')
        
    # plt.figure(80000+bb*100)
    # plt.ylabel(ylabs)
    # plt.xlabel('Number of samples $j$ [-]')
    # plt.xticks(ticks=x_ticks1)
    # # plt.title(f'Adaptive Screening with GPR, {indcase}')
    # plt.legend(ncol=len(acqnames))
    # plt.xlim([N_hf-1,sampmaxplot1])
    # plt.ylim([min_gpr_nounc,max_gpr_nounc])
    # if annon == 1:
    #     plt.savefig(f'{out_figs}/{plotcase}_Convergence_GPR-{sampopt_txt}_col_ann.pdf',bbox_inches='tight')
    # else:
    #     plt.savefig(f'{out_figs}/{plotcase}_Convergence_GPR-{sampopt_txt}_col.pdf',bbox_inches='tight')

    # plt.figure(90000+bb*100)
    # plt.ylabel(ylabs)
    # plt.xlabel('Number of samples $j$ [-]')
    # plt.xticks(ticks=x_ticks1)
    # # plt.title(f'Adaptive Screening with MF-GPR, {indcase}')
    # plt.legend(ncol=len(acqnames))
    # plt.xlim([N_hf-1,sampmaxplot1])
    # plt.ylim([min_gpr_nounc,max_gpr_nounc])
    # if annon == 1:
    #     plt.savefig(f'{out_figs}/{plotcase}_Convergence_MFGPR-{sampopt_txt}_col_ann.pdf',bbox_inches='tight')
    # else:
    #     plt.savefig(f'{out_figs}/{plotcase}_Convergence_MFGPR-{sampopt_txt}_col.pdf',bbox_inches='tight')
    
    
    # Close figures
    # plt.close(20000+bb*100)
    # plt.close(30000+bb*100)
    # plt.close(140000+bb*100)
    # plt.close(150000+bb*100)
    # plt.close(40000+bb*100)
    # plt.close(50000+bb*100)
    # plt.close(60000+bb*100)
    # plt.close(70000+bb*100)
    # plt.close(80000+bb*100)
    # plt.close(90000+bb*100)
    
    
        


##### M1: Plot average, min and max number of samples required for convergence

# plt.close('all')

# GPR
M1_mean_gpr = np.squeeze(np.nanmean(M1_gpr,axis=0))
M1_min_gpr = np.squeeze(np.nanmin(M1_gpr,axis=0))
M1_max_gpr = np.squeeze(np.nanmax(M1_gpr,axis=0))

# MF-GPR
M1_mean_mfgpr = np.squeeze(np.nanmean(M1_mfgpr,axis=0))
M1_min_mfgpr = np.squeeze(np.nanmin(M1_mfgpr,axis=0))
M1_max_mfgpr = np.squeeze(np.nanmax(M1_mfgpr,axis=0))

# Plot
plt.figure(figsize=(6,3.5))
plt.plot(allacqnames,M1_mean_gpr,color='b',marker='o',label='Mean GPR')
plt.fill_between(allacqnames, M1_min_gpr.flatten(),M1_max_gpr.flatten(), facecolor='b', alpha=0.1,label='Min-Max GPR')
plt.plot(allacqnames,M1_mean_mfgpr,color='r',marker='d',label='Mean MF-GPR')
plt.fill_between(allacqnames, M1_min_mfgpr.flatten(),M1_max_mfgpr.flatten(), facecolor='r', alpha=0.1,label='Min-Max MF-GPR')
plt.ylabel('$M_1$ [-]')
plt.ylim([0,140.001])
plt.legend(ncol=2)
plt.savefig(f'{out_figs}/M1_{case_string}.pdf',bbox_inches='tight')


##### M2: Plot MPM deviation at convergence
 
# GPR
M2_mean_gpr = np.squeeze(np.nanmean(M2_gpr,axis=0))
M2_min_gpr = np.squeeze(np.nanmin(M2_gpr,axis=0))
M2_max_gpr = np.squeeze(np.nanmax(M2_gpr,axis=0))

# MF-GPR
M2_mean_mfgpr = np.squeeze(np.nanmean(M2_mfgpr,axis=0))
M2_min_mfgpr = np.squeeze(np.nanmin(M2_mfgpr,axis=0))
M2_max_mfgpr = np.squeeze(np.nanmax(M2_mfgpr,axis=0))

# Plot
plt.figure(figsize=(6,3.5))
plt.plot(allacqnames,M2_mean_gpr,color='b',marker='o',label='Mean GPR')
plt.fill_between(allacqnames, M2_min_gpr.flatten(),M2_max_gpr.flatten(), facecolor='b', alpha=0.1,label='Min-Max GPR')
plt.plot(allacqnames,M2_mean_mfgpr,color='r',marker='d',label='Mean MF-GPR')
plt.fill_between(allacqnames, M2_min_mfgpr.flatten(),M2_max_mfgpr.flatten(), facecolor='r', alpha=0.1,label='Min-Max MF-GPR')
plt.ylabel('$M_2$ [-]')
plt.ylim([0,0.1001])
plt.legend(ncol=2)
plt.savefig(f'{out_figs}/M2_{case_string}.pdf',bbox_inches='tight')


##### M3: Plot distribution deviation at convergence

# GPR
M3_mean_gpr = np.squeeze(np.nanmean(M3_gpr,axis=0))
M3_min_gpr = np.squeeze(np.nanmin(M3_gpr,axis=0))
M3_max_gpr = np.squeeze(np.nanmax(M3_gpr,axis=0))

# MF-GPR
M3_mean_mfgpr = np.squeeze(np.nanmean(M3_mfgpr,axis=0))
M3_min_mfgpr = np.squeeze(np.nanmin(M3_mfgpr,axis=0))
M3_max_mfgpr = np.squeeze(np.nanmax(M3_mfgpr,axis=0))

# Plot
plt.figure(figsize=(6,3.5))
plt.plot(allacqnames,M3_mean_gpr,color='b',marker='o',label='Mean GPR')
plt.fill_between(allacqnames, M3_min_gpr.flatten(),M3_max_gpr.flatten(), facecolor='b', alpha=0.1,label='Min-Max GPR')
plt.plot(allacqnames,M3_mean_mfgpr,color='r',marker='d',label='Mean MF-GPR')
plt.fill_between(allacqnames, M3_min_mfgpr.flatten(),M3_max_mfgpr.flatten(), facecolor='r', alpha=0.1,label='Min-Max MF-GPR')
plt.ylabel('$M_3$ [-]')
plt.ylim([0,0.1001])
plt.legend(ncol=2)
plt.savefig(f'{out_figs}/M3_{case_string}.pdf',bbox_inches='tight')


##### Combined plot M2 and M3

# Plot
plt.figure(figsize=(6,3.5))
plt.plot(allacqnames,M2_mean_gpr,color='b',marker='o',label='$M_2$, Mean GPR')
plt.fill_between(allacqnames, M2_min_gpr.flatten(),M2_max_gpr.flatten(), facecolor='b', alpha=0.1,label='$M_2$, Min-Max GPR')
plt.plot(allacqnames,M2_mean_mfgpr,color='r',marker='d',label='$M_2$, Mean MF-GPR')
plt.fill_between(allacqnames, M2_min_mfgpr.flatten(),M2_max_mfgpr.flatten(), facecolor='r', alpha=0.1,label='$M_2$, Min-Max MF-GPR')
plt.plot(allacqnames,M3_mean_gpr,linestyle = '--',color='c',marker='o',label='$M_3$, Mean GPR')
plt.fill_between(allacqnames, M3_min_gpr.flatten(),M3_max_gpr.flatten(), facecolor='c', alpha=0.1,label='$M_3$, Min-Max GPR')
plt.plot(allacqnames,M3_mean_mfgpr,linestyle = '--',color='m',marker='d',label='$M_3$, Mean MF-GPR')
plt.fill_between(allacqnames, M3_min_mfgpr.flatten(),M3_max_mfgpr.flatten(), facecolor='m', alpha=0.1,label='$M_3$, Min-Max MF-GPR')
plt.ylabel('$M_2$ or $M_3$ [-]')
plt.legend(ncol=2)
plt.ylim([0,0.1001])
plt.savefig(f'{out_figs}/M2_M3_{case_string}.pdf',bbox_inches='tight')
