# -*- coding: utf-8 -*-
"""
Created on 19 Oct 2023

@author: smvanessen
"""

# Clear all variables
from IPython import get_ipython
get_ipython().magic('reset -sf')

# Standard toolboxes
import numpy as np
import matplotlib.pyplot as plt
import sys 
import os

# Distributions (Rayleigh and Forristall), first add path to functions
sys.path.append('functions')
from functions.theoretical_distributions import make_cdf_Rayleigh as Rayl
from functions.theoretical_distributions import make_cdf_Forristall as Forr
from functions.distribution_scripts import calc_distribution,MPM_from_DNR
from functions.T_enc import calc_T_enc

# Allow for subsampling of large datasets for plotting & saving as PDF
from functions.subsample_routines import subsample_binned

# Close all figures
plt.close('all')


#############
# START INPUT
#############
           
# Figure sizes (units unclear, see matplotlib doc)
fsize_h = 5.5
fsize_v = 3.5

# Choose application and corresponding indicator cases
application = 'Waves'; hfindcases = ['GoodInd','WorseInd']
# application = 'VBM'; hfindcases = ['Hogging']
# # application = 'VBM'; hfindcases = ['Hogging','Sagging']

# Folder name in which to store figures (rel. to present folder)
figfolder = 'out_figs_paper_PRADS'

# Initialise plots 
fig1 = plt.figure(1,figsize=(fsize_h,fsize_v))
plt.yscale('log')

# Define colors
colsMCS_LF = ['blue','cyan']
colMCS_HF = 'red'

# Loop over indicator cases for each application
for ii,hfindcase in enumerate(hfindcases):
    
    colMCS_LF = colsMCS_LF[ii]
    
    if hfindcase == 'Hogging' or hfindcase == 'Sagging':
    
        # Folder name where input files are stored (files to read)
        inputfolder = 'inputfiles_vbm'
           
        # Target duration [s]
        dur = 1800
        
        # Number of seeds
        Ns = 60
        
        # Total available duration [s]
        Texp = Ns*dur
        
        # Wave properties, speed and heading (for derivation wave encounter period and Pexp)
        Tp = 10.0 # [s]
        Vs = 10.0*0.51444444 # [m/s]
        mu = np.pi
        
    elif hfindcase == 'GoodInd' or hfindcase == 'WorseInd':
        
        # Folder name where input files are stored (files to read)
        inputfolder = 'inputfiles_waves'
           
        # Target duration [s]
        dur = 3600
        
        # Number of seeds
        Ns = 50
        
        # Total available duration [s]
        Texp = Ns*dur
        
        # Wave properties, speed and heading (for derivation wave encounter period and Pexp)
        Tp = 11.0 # [s]
        Vs = 0.0 # [m/s]
        mu = np.pi
        
    
    #############
    # END INPUT
    #############
    
    # Check wether output directories exist, if not make them
    isExist = os.path.exists(figfolder)
    if not isExist:
       os.makedirs(figfolder)
      
    # Derive target exposure duration probability level
    Tpe = calc_T_enc(Tp,Vs,mu)
    ne_exp = dur/Tpe
    pexp = 1/ne_exp
    
    
    if hfindcase == 'Hogging' or hfindcase == 'Sagging':
    
        # Read matched LF and HF peaks from CSV file
        AllTrueDat_scr = np.loadtxt(f'{inputfolder}/MARINferry2_MatchedPeaks{hfindcase}_VBM_JONS_Hs13.2m_Tp10.0s_gam3.0_Vs10.00kn_mu180deg.csv',skiprows=1,delimiter=',')
        # t_peaks_lf_all = AllTrueDat_scr[:,0]
        peaks_lf_all = AllTrueDat_scr[:,1]
        # t_peaks_hf_all_scr = AllTrueDat_scr[:,2]
        peaks_hf_all_scr = AllTrueDat_scr[:,3]
    
        # Read HF true peaks from CSV file
        AllTrueDat_true = np.loadtxt(f'{inputfolder}/MARINferry2_PRETTI_peaks_VBM_JONS_Hs13.2m_Tp10.0s_gam3.0_Vs10.00kn_mu180deg.csv',skiprows=1,delimiter=',')
        if hfindcase == 'Hogging':
            # t_peaks_hf_all_true = AllTrueDat_true[:,2]
            peaks_hf_all_true = AllTrueDat_true[:,3]
        elif hfindcase == 'Sagging':
            # t_peaks_hf_all_true = AllTrueDat_true[:,0]
            peaks_hf_all_true = AllTrueDat_true[:,1]
        
    elif hfindcase == 'GoodInd' or hfindcase == 'WorseInd':
        
        # Read matched LF and HF peaks from CSV file
        AllTrueDat_scr = np.loadtxt(f'{inputfolder}/Lin2nd_trueLFandHFdata_{hfindcase}_{Ns}seeds_scr.csv', delimiter=',')
        # t_peaks_lf_all = AllTrueDat_scr[:,0]
        peaks_lf_all = AllTrueDat_scr[:,1]
        # t_peaks_hf_all_scr = AllTrueDat_scr[:,2]
        peaks_hf_all_scr = AllTrueDat_scr[:,3]
        
        # Read HF true peaks from CSV file
        seedinit = '0' # selected 50hour MCS realisation (0 to 9)
        AllTrueDat_true = np.loadtxt(f'{inputfolder}/Lin2nd_trueHFdata_real{seedinit}_{Ns}seeds.csv', delimiter=',')
        # t_peaks_hf_all_true = AllTrueDat_true[:,0]
        peaks_hf_all_true = AllTrueDat_true[:,1]
    
    print('Reading true HF peak and screening matched LF-HF peak CSV files done')
    
    # Calculate true distributions for 'Monte-Carlo' dataset (all data, validation)
    v_true_l,p_true_l_temp,_ = calc_distribution(peaks_lf_all)
    v_true_h,p_true_h_temp,_ = calc_distribution(peaks_hf_all_true)
    
    #####
    # Correct for using one estimated number of waves as basis instead of the number 
    # of data points. Because the LF and HF signals can be different variables, we 
    # need a common basis for their probability of exceedance (PoE) distributions. 
    # This is provided by  Nwaves; the largest LF and HF value both have PoE 
    # 1/Nwaves in the output.
    
    # Estimated number of wave encounters in full experimental duration (where 
    # the full experimental duration is estimated based on the time of the last 
    # indicator peak). 
    Nwaves = Texp/dur*ne_exp
    
    # Correct for basis Nwaves instead of the number of LF or HF peaks
    N_true_l = len(v_true_l)
    N_true_h = len(v_true_h)
    p_true_l = p_true_l_temp*N_true_l/Nwaves
    p_true_h = p_true_h_temp*N_true_h/Nwaves
    
    #####
    
    # Calculate / derive true MPM from DNR
    MPM_true_MCS = MPM_from_DNR(v_true_h,p_true_h,ne_exp)
    
    # Subsample the MCS true distribution data to avoid large PDF files
    maxvalall = np.nanmax((np.nanmax(v_true_l),np.nanmax(v_true_h)))
    binsize = maxvalall/200
    bins = np.arange(0,maxvalall,binsize)
    v_true_l_sampled,p_true_l_sampled = subsample_binned(v_true_l,p_true_l,bins)
    v_true_h_sampled,p_true_h_sampled = subsample_binned(v_true_h,p_true_h,bins)
    
    # Plot true MCS DNR and Pexp + finish and save plot
    if application == 'Waves':
        plt.plot(v_true_l_sampled,p_true_l_sampled,marker='.',color=colMCS_LF,linestyle='None',label=f'LF MCS, {hfindcase}') 
    else:
        plt.plot(v_true_l_sampled,p_true_l_sampled,marker='.',color=colMCS_LF,linestyle='None',label='LF MCS') 
    if ii == len(hfindcases)-1:
        plt.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color=colMCS_HF,linestyle='None',label='HF validation MCS')
        plt.plot([0,maxvalall*1.1],[pexp,pexp],':',color='black')
        plt.text(maxvalall*0.9,pexp*1.4,'$P_{exp}$')
        if hfindcase == 'Hogging' or hfindcase == 'Sagging':
            plt.ylabel('P(V $\geq$ v) [-]')
            plt.xlabel('v [Nm]')
        elif hfindcase == 'GoodInd' or hfindcase == 'WorseInd':
            plt.ylabel('P(C $\geq$ c) [-]')
            plt.xlabel('c [m]')
        plt.ylim([10**(-5),1])
        plt.xlim([0,maxvalall*1.1])
        
# Finish
plt.legend()
plt.savefig(f'{figfolder}/MCS_distributions_input_{application}.pdf',bbox_inches='tight')

print(f'{dur/3600}h MPM True MCS: {MPM_true_MCS} Nm')
    
