# -*- coding: utf-8 -*-
"""
Created on Thu Nov  3 11:58:45 2023
@author: smvanessen
"""

# Import standard toolboxes
import numpy as np
import matplotlib.pyplot as plt
# import sys
import pandas as pd
import os
import csv

# Import GPy and Emukit functions. Refer to this documentation:
# https://github.com/SheffieldML/GPy
# https://gpy.readthedocs.io/en/deploy/
# https://emukit.github.io/
# We use the linear MF-GPR model, but there is also a (customisable) non-linear model: 
# https://github.com/EmuKit/emukit/blob/main/emukit/multi_fidelity/models/non_linear_multi_fidelity_model.py
import GPy
import emukit.multi_fidelity
from emukit.model_wrappers.gpy_model_wrappers import GPyMultiOutputWrapper, GPyModelWrapper
from emukit.multi_fidelity.models import GPyLinearMultiFidelityModel           
from emukit.multi_fidelity.convert_lists_to_array import convert_x_list_to_array, convert_xy_lists_to_arrays

# Additions for Experimental Design / Adaptive Sampling loop
from emukit.core import ParameterSpace, ContinuousParameter

# Acquisition functions from Emukit library
from emukit.experimental_design.acquisitions import IntegratedVarianceReduction, ModelVariance

# Import custom acquisition functions as new classes
# sys.path.append('functions')
from functions.acq_maxval import MaxV
from functions.acq_parab_pdist import ParabPdist
from functions.acq_maxdist import MaxDist

# Import some other custom functions
from functions.distribution_scripts import calc_distribution,MPM_from_DNR,fit_weibull3_lsq
from functions.subsample_routines import subsample_binned,subsample_binned_2D

# Reference generalised extreme value distributions
from functions.GP_functions import GP_pexc,GP_pexc_log
from scipy.optimize import curve_fit

# # Filters
# from scipy.signal import savgol_filter
# from scipy.ndimage import gaussian_filter1d, uniform_filter1d
# from scipy.signal import medfilt

##########################################
def adaptive_screening(peaks_LF_all,peaks_HF_all_scr,peaks_HF_all_true,Tdata,acqtypeHF,acqtypeMF,maxit_HFGPR,maxit_MFGPR,prange,sampinit,initplevels,dur,ne_exp,filefolder,figfolder,casename,varsymb,varunit,debugplotswitch):
    """

    Parameters
    ----------
    This function performs adaptive screening with both single-fidelity HF-GPR 
    and multi-fidelity GPR, for a case study of linear & second-order waves.
    Four cases are defined for this case study: perfect indicator, no noise, 
    LongTp noise and ShortTp noise. Refer to OMAE2023 and JOMAE2023/24 paper 
    for a description of the cases. Each can be evaluated with different adaptive
    sampling acquisition functions and max number of iterations. Outcome is a 
    plot of the predicted HF DNR compared to the true HF data and predictions 
    of the MPM and its uncertainty.
        
    peaks_LF_all : 0D array
        All LF indicator peaks in the Monte-Carlo simulation
    peaks_HF_all_scr : 0D array
        All HF peaks in the Monte-Carlo simulation that are matched to the LF 
        peaks, in exactly the same order. These are normally not available, as 
        they need to be derived using a CFD calculation or experiment for the 
        new event that is defined in each iteration. However, for the case study 
        in the paper*, these were available at the start of the procedure from 
        2nd order wave theory.
    peaks_HF_all_true : 0D array
        All HF peaks in the validation Monte-Carlo simulation. These are usually 
        not available in a real strongly non-linear response application either.
        They are only used as ground truth for the validation.   
    Tdata : float
        Duration of the present dataset in seconds. Used together with ne_exp 
        to estimate the number of wave encounters in this duration (Nwaves). 
        This in turn is used as a basis for the probability of exceedance (PoE) 
        levels in all LF and HF distributions. Both the largest LF and HF value 
        have a PoE equal to 1/Nwaves.
    acqtypeHF : string
        Acquisition function used for single-fidelity HF GPR
        - 'US'      = Uncertainty sampling
        - 'IVR'     = Integrated variance reduction - do not use in combination 
                      with MF-GPR (only with GPR), this is not implemented yet.
        - 'MV'      = Mean predicted value (not in paper)
        - 'USMV'    = Product US and mean predicted value
        - 'IVRMV'   = Product IVR and mean predicted value (not in paper) - 
                      do not use in combination with MF-GPR (only with GPR), 
                      this is not implemented yet.
        - 'USPD'    = Product US and normalised hill parabola with top at Pexp
        - 'IVRPD'   = Product IVR and normalised hill parabola with top at Pexp 
                      (not in paper) - do not use in combination with MF-GPR 
                      (only with GPR), this is not implemented yet.
        - 'MD'      = Function that simply selects the location within prange 
                      that is furthest from any of the earlier sample points.
        - 'MVPD'    = Product mean predicted value and normalised hill parabola 
                      with top at Pexp.
    acqtypeMF : string
        Acquisition function used for multi-fidelity GPR
        Same choices, but note that none of the options with IVR are implemented 
        for MF-GPR yet.
    maxit_HFGPR : integer
        Max. number of HF-GPR iterations (on top of 3 initial samples)
    maxit_MFGPR : integer
        Max. number of MF-GPR iterations (on top of 3 initial samples)
    prange : 1D array
        Array providing the probability of exceedance prediction range. The GPR 
        or MF-GPR exceedance probability distribution will be predicted over 
        this range. Defined from low to high probability values. Example: 
        prange = np.linspace(plow,phigh,pstep)[:, None]
        The new samples will always be selected from within this range.
    sampinit : string
        Initial sample option for LF and HF data. Different options, that only 
        differ in their definition of the LF samples used by MF-GPR. The options
        are identical for GPR:
        - 'A' = initial HF samples defined at <initplevels> - LF sampling for 
                MF-GPR using resampling with one sample per maxLF value / 100 interval
        - 'B' = initial HF samples defined at <initplevels> - LF sampling for 
                MF-GPR using GEV fit to LF samples over Prange_predLF
        - 'C' = initial HF samples defined at <initplevels> - LF sampling for 
                MF-GPR using GEV fit to available HF samples over Prange_predLF 
                (updated every iteration) - not tested yet.
    initplevels : 0D array
        Array of probability of exceedance levels at which we derive the initial 
        HF sampling. Typically has 2, 3, or 4 elements. These have to be within 
        Prange. Example: initplevels = np.array([p1,p2,p3]) in the case of three 
        initial samples. These probability levels may be outside prange (although 
        it may make sense to use e.g. the lower and upper limits of prange).
    dur : float
        Duration of interest / exposure duration. Typically between 30 min and 
        3 hours. Provided in seconds, so between 1800 and 10800.
    ne_exp : float
        Average number of wave encounters in the duration of interest <dur>. 
        Used to derive the probability of exceedance of interest and basis for 
        the PoE distribution axes.
    filefolder : string
        Folder name where to save the output CSV files (relative to present folder). 
        Location should exist already.
    figfolder : string
        Folder name where to save the output PDF figures (relative to present folder). 
        Location should exist already.
    casename : string
        Customisable name of the present case - will appear in the plot file names.
    varsymb : string
        Symbols for the LF and HF variable on the x-axis of the PoE plots. Have
        to be lower-case. Format: ['LF variable symbol','HF variable symbol']
    varunit : string
        Units for the variable on the x-axis of the probability of exceedance 
        plots. Have to be lower-case. Format: ['LF unit','HF unit']
    debugplotswitch : integer
        1 = plot (extra) debugging plots
        0 = do not plot these (extra) debugging plots
    
    Returns
    -------   
    itrangeHF : numpy array
        Number of samples included in each iteration of the GPR procedure.
    ass_MPM_hfGPR : numpy array
        MPM derived in each iteration of the GPR procedure.
    ass_MPM_hfGPR_u95 : numpy array
        U95% uncertainty of the MPM derived in each iteration of the GPR procedure.
    itrangeMF : numpy array
        Number of samples included in each iteration of the MF-GPR procedure.
    ass_MPM_mfGPR : numpy array
        MPM derived in each iteration of the MF-GPR procedure.
    ass_MPM_mfGPR_u95 : numpy array
        U95% uncertainty of the MPM derived in each iteration of the MF-GPR procedure.
    mean_high_gp_model : numpy array
        Mean HF predicted exceedance distribution at the last iteration, when 
        convergence was reached with the GPR procedure. Defined over <prange>.
    var_high_gp_model : numpy array
        Variance of the HF predicted exceedance distribution at the last iteration, 
        when convergence was reached with the GPR procedure. Defined over <prange>.
    mean_low_mfgp_model : numpy array
        Mean LF predicted exceedance distribution at the last iteration, when 
        convergence was reached with the MF-GPR procedure. Defined over <prange>.
    var_low_mfgp_model : numpy array
        Variance of the LF predicted exceedance distribution at the last iteration, 
        when convergence was reached with the MF-GPR procedure. Defined over <prange>.
    mean_high_mfgp_model : numpy array  
        Mean HF predicted exceedance distribution at the last iteration, when 
        convergence was reached with the MF-GPR procedure. Defined over <prange>.
    var_high_mfgp_model : numpy array
        Variance of the HF predicted exceedance distribution at the last iteration, 
        when convergence was reached with the MF-GPR procedure. Defined over <prange>.

    
    * Van Essen, S.M. and Seyffert, H.C. (2024?). Designing for dangerous waves – 
    a new ‘Adaptive Screening’ method to predict extreme values for non-linear 
    responses of marine structures to waves. ADD JOURNAL DETAILS.
    
    Revision history
    ---------
    v1 Mar'24   Sanne van Essen     Connected to draft journal paper with a case 
                                    study of second-order waves, submitted to RESS
    v2 ??       Sanne van Essen     Connected to the updated journal paper.
    v3 Apr'24   Sanne van Essen     Version used for a new case study (SCREAM 
                                    experiments, predicting MPM values of non-linear
                                    RWE and / or wave impact loads)
    
    
    """
      
    # Figure sizes (see matplotlib doc)
    fsize_h = 7
    fsize_v = 5
    
    # Variable symbols and units
    varsymb_lf = varsymb[0]
    varsymb_hf = varsymb[1]
    varunit_lf = varunit[0]
    varunit_hf = varunit[1]
    
    # Check wether output directories exist, if not make them
    isExist = os.path.exists(filefolder)
    if not isExist:
       os.makedirs(filefolder)
    isExist = os.path.exists(figfolder)
    if not isExist:
       os.makedirs(figfolder)
    
    # Max values (for plotting) - cannot be used for normalisation, as we don't start with many HF datapoints in practical procedure
    plotmax = np.ceil(1.1*np.nanmax(peaks_HF_all_true)) # HF
    plotmax_lf = np.ceil(1.1*np.nanmax(peaks_LF_all)) # LF
      
    # Exposure duration probability level in the DNR, and location in plots
    pexp = 1/ne_exp  
    locpexp = plotmax*0.9
    
    # Estimated number of wave encounters in full experimental duration (where 
    # the full experimental duration is estimated based on the time of the last 
    # indicator peak). Because the LF and HF signals can be different variables, 
    # we need a common basis for their probability of exceedance (PoE) distributions. 
    # This is provided by  Nwaves; the largest LF and HF value both have PoE 
    # 1/Nwaves in the output.
    Nwaves = Tdata/dur*ne_exp
            
    # Hardcoded: Matern32 kernels for the MF-GPR computation    
    kern_MFgp_lf = GPy.kern.Matern32(input_dim=1, variance=1., lengthscale=1.); MFkern1 = 'Mat32'
    kern_MFgp_delta = GPy.kern.Matern32(input_dim=1, variance=1., lengthscale=1.); MFkern2 = 'Mat32'
    kernels_MFgp = [kern_MFgp_lf,kern_MFgp_delta]
       
    # Hardcoded: Matern32 kernel for the GPR computation    
    kernel_HFgp = GPy.kern.Matern32(input_dim=1, variance=1., lengthscale=1.); HFkern = 'Mat32'
    
    # Prediction value range for the LF predictions in MF-GPR
    Prange_predLF = prange.copy()
    Prange_predLF_log = np.log(Prange_predLF)
    
    # Prediction value range for the HF predictions in GPR and MF-GPR
    Prange_predHF = prange.copy()
    Prange_predHF_log = np.log(Prange_predHF)
        
    # Instead, change the data points with zero HF response values from the 
    # matched data back to NaN values (these wave events) led to HF responses  
    # below the POT threshold. Let's see how GPR accounts for this.
    i_zero_hf_scr = np.where(peaks_HF_all_scr==0)[0]
    peaks_HF_all_scr[i_zero_hf_scr] = np.nan
        
    ## ------- Derived distribution of true data
    
    # Calculate true distributions for 'Monte-Carlo' dataset (all data, validation) - also obtain 'screening' order indices
    v_true_l,p_true_l_temp,i_screening = calc_distribution(peaks_LF_all)
    v_true_h,p_true_h_temp,i_HFtrue = calc_distribution(peaks_HF_all_true) # Not available normally

    # Correct for basis Nwaves instead of the number of LF or HF peaks
    N_lf = len(peaks_LF_all)
    N_hf = len(peaks_HF_all_true)
    p_true_l = p_true_l_temp*N_lf/Nwaves
    p_true_h = p_true_h_temp*N_hf/Nwaves
    
    # Min PoE value (for plotting)
    minP = 0.9*np.nanmin(p_true_h) # HF
    
    # Calculate true MPM values
    MPM_true_MCS = MPM_from_DNR(v_true_h,p_true_h,ne_exp)
 
    # Subsample the MCS true distribution data for plotting later
    binsizeHF = plotmax/500
    binsHF = np.arange(0,plotmax,binsizeHF)
    v_true_h_sampled,p_true_h_sampled = subsample_binned(v_true_h,p_true_h,binsHF)
    binsizeLF = plotmax_lf/500
    binsLF = np.arange(0,plotmax_lf,binsizeLF)
    v_true_l_sampled,p_true_l_sampled = subsample_binned(v_true_l,p_true_l,binsLF)
            
    # Max number of available LF datapoints (for DEM = Nseeds) 
    Nmax = len(v_true_l)

    # ------- Show difference true HF distribution and best possible screening result with all events and present indicator
    
    # Screening result in the probability of exceedence distribution domain        
    v_scr_h = peaks_HF_all_scr[i_screening]
    p_scr_h = p_true_l.copy()
    
    # Find probability of exceedence level where half of the points is above and below (based on the LF indicator)
    i_half = np.int(np.ceil(len(v_true_l)/2))
    p_half = p_true_l[i_half]
    
    # Corresponding exposure duration (depends on the present LF MCS duration, 
    # response non-linearity and  maybe indicator quality)
    dur_half = pexp*dur/p_half
        
    # create bins and calculate bin mean
    nbins = 10
    Pbins_log = np.linspace(np.nanmin(np.log(p_scr_h)),np.nanmax(np.log(p_scr_h)),nbins)
    Pbins = np.exp(Pbins_log)
    Pbins_middle,v_scr_h_smooth4 = bin_means_smoothing(p_scr_h,v_scr_h,Pbins)
        
    plt.figure(600,figsize=(fsize_h,fsize_v))
    plt.xlabel(r'$\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (varsymb_hf,varunit_hf), fontsize=12)
    plt.ylabel(r'$\mathrm{P(}\mathrm{%s} \geq \mathrm{%s} \mathrm{)}$' % (varsymb_hf.upper(),varsymb_hf), fontsize=12)
    plt.xlabel(f'{varsymb_hf} [{varunit_hf}]')
    plt.ylim([minP,1])
    plt.xlim([0,plotmax])
    plt.plot(v_scr_h,p_scr_h,marker='.',linestyle='none',color='purple',linewidth=0.8,label='All HF screening samples')
    plt.plot(v_true_h_sampled,p_true_h_sampled,marker='.',linestyle='--',color='pink',linewidth=0.8,label='HF "true" MCS')
    plt.plot(v_scr_h_smooth4,Pbins_middle,'-.',color='r',label=f'Smoothing by bin means ({nbins} bins)')
    plt.plot([0,plotmax],[p_half,p_half],'--k',linewidth=0.5,label=r'$\mathrm{P}_{\mathrm{half}} \mathrm{ (}{%i}\mathrm{s)}$' % (np.int(dur_half)))
    plt.plot([0,plotmax],[pexp,pexp],':',color='black',label=r'$\mathrm{P}_{\mathrm{exp}} \mathrm{ (}{%i}\mathrm{s)}$' % (np.int(dur)))
    plt.savefig(f'{figfolder}/ScreeningSamples_{casename}_ylin.pdf',bbox_inches='tight')  
    plt.yscale('log')
    plt.legend()
    plt.savefig(f'{figfolder}/ScreeningSamples_{casename}_ylog.pdf',bbox_inches='tight')  
      
    # ------- All LF samples (this leads to too many LF peaks for multiple realisations)
    v_true_l_t = v_true_l[~np.isnan(v_true_l)]
    p_true_l_t = p_true_l[~np.isnan(v_true_l)]
    v_scr_h_t = v_scr_h[~np.isnan(v_true_l)]
    p_scr_h_t = p_scr_h[~np.isnan(v_true_l)]


    # ------- Initial HF sampling (same for all three sampling options)
    
    # Initialise 
    i_sel_dist = np.zeros(np.shape(initplevels),dtype=int)

    # Find corresponding indices of these samples (may be outside available pool)
    for cc,p_lev in enumerate(initplevels):
        i_sel_dist[cc],_ = find_nearest(p_true_l_t,p_lev)    
        
    # Selected LF points where we pick the matched HF points (may be outside available pool)
    p_train_l_REF = p_true_l_t[i_sel_dist]
    v_train_l_REF = v_true_l_t[i_sel_dist]
    
    # Now we have to get back to the original time series peak matching to 
    # find the corresponding HF values > these are the initial HF training data values
    v_train_h = np.zeros(np.size(v_train_l_REF))
    for ii,selvalue in enumerate(v_train_l_REF):
        i_sel_time = np.where(peaks_LF_all == selvalue)
        v_train_h[ii] = peaks_HF_all_scr[i_sel_time]
        
    # HF probabilities are set identical to LF probabilities
    p_train_h = p_train_l_REF    
    
    
    #######################
    # ------- NORMALISATION
    #######################
    #
    # Normalise input values (as the performance of GPR proves to be quite 
    # sensitive to scaling of the distributions). 
    # HF: normalised with largest initial sampling value
    # LF: normalised with LF value corresponding to the same initial PoE level
    # Note that sampled true distributions are not scaled (as they are only used for plotting)
    norm_hf = np.max(v_train_h)
    i_norm_hf = np.argmax(v_train_h)
    norm_lf = v_train_l_REF[i_norm_hf]
    
    peaks_LF_all = peaks_LF_all/norm_lf
    peaks_HF_all_scr = peaks_HF_all_scr/norm_hf
    peaks_HF_all_true = peaks_HF_all_true/norm_hf
    
    v_train_l_REF = v_train_l_REF/norm_lf
    v_true_l = v_true_l/norm_lf
    v_true_l_t = v_true_l_t/norm_lf
    
    v_true_h = v_true_h/norm_hf
    v_train_h = v_train_h/norm_hf
    v_scr_h = v_scr_h/norm_hf
    v_scr_h_t = v_scr_h_t/norm_hf
        
    # ------- LF sampling (varies per sampling option)
    if sampinit == 'A':
        # HF initial samples for GPR and MF-GPR defined at <initplevels>
        # LF samples for MF-GPR defined by subsampling all LF samples.
        txtadd = f'A_LFsubsamp_exp{np.int(dur/60)}min'
        
        # Subsample the MCS true distribution data for plotting later
        plotmax_lf_norm = plotmax_lf/norm_lf
        binsize1 = plotmax_lf_norm/100 # normalised
        bins1 = np.arange(0,plotmax_lf_norm,binsize1) # normalised
        v_train_l,p_train_l = subsample_binned(v_true_l_t,p_true_l_t,bins1)
        
    elif sampinit == 'B':
        # HF initial samples for GPR and MF-GPR defined at <initplevels>
        # LF samples for MF-GPR defined by a Weibull fit to the topperc% highest LF datapoints.
        txtadd = f'B_LFfit_exp{np.int(dur/60)}min'

        # Generate 3-param Weibull fit to all LF data
        N_lf_points = 50 # number of LF points over the range indicated by topperc
        topperc = 50 # percentage of peaks to include in fit
        maxit = 100 # maximum number of iterations
        _,_,theta,alpha,beta,corr = fit_weibull3_lsq(p_true_l_t,v_true_l_t,topperc,maxit)
        vmin = v_true_l_t[-int(np.floor(len(v_true_l_t) * (topperc / 100)))]
        vmax = np.max(v_true_l_t) 
        v_topperc = np.linspace(vmin,vmax,N_lf_points)[:, None]
        p_weib_LF = np.exp(-((v_topperc - theta)/alpha)**beta);
        
        # LF samples to use in prediction
        v_train_l = np.squeeze(v_topperc)
        p_train_l = np.squeeze(p_weib_LF)
        
        # Write LF fit to CSV file (required for fig 8)
        dat_wei = np.transpose(np.stack((np.squeeze(v_train_l*norm_lf),np.squeeze(p_train_l))))
        my_df_wei = pd.DataFrame(dat_wei)
        head_wei = ['Vweibull','Pweibull']
        my_df_wei.to_csv(f'{filefolder}/{casename}_LFsamples_3pWeibullfit_to_top{topperc}p.csv',header = head_wei, index= False)
    
    elif sampinit == 'C':
        # HF initial samples for GPR and MF-GPR defined at <initplevels>
        # LF samples for MF-GPR defined by a Weibull fit to the available HF 
        # points. This fit needs to be updated every iteration, so the LF 
        # datapoints also change every iteration. This is handled later in the 
        # code. This option was not yet tested, so may have bugs.
        txtadd = f'C_LFisHFfit_exp{np.int(dur/60)}min'
                       
        # Generate 3-param Weibull fit to all available HF datapoints
        N_lf_points = 50 # number of LF points over the range indicated by topperc
        topperc = 100 # percentage of peaks to include in fit - all of available ones
        maxit = 100 # maximum number of iterations
        _,_,theta,alpha,beta,corr = fit_weibull3_lsq(p_train_h,v_train_h,topperc,maxit)
        vmin = v_train_h[-int(np.floor(len(v_true_l_t) * (topperc / 100)))]
        vmax = np.max(v_train_h) 
        v_topperc = np.linspace(vmin,vmax,N_lf_points)[:, None]
        p_weib_LF = np.exp(-((v_topperc - theta)/alpha)**beta);
       
        # LF samples to use in prediction
        v_train_l = np.squeeze(v_topperc)
        p_train_l = np.squeeze(p_weib_LF)

    # Debugging; check LF sampling
    plt.figure()
    plt.plot(v_true_l_t*norm_lf,p_true_l_t,'x',label='All LF data points')
    plt.plot(v_train_l*norm_lf,p_train_l,'o',label='LF samples')
    plt.title(f'Debugging sample option: {txtadd}')
    plt.yscale('log')
    plt.legend()
    
    # Make a pool of samples, within the prediction range prange
    inds_range = np.where((p_true_l_t>=min(prange)) & (p_true_l_t<=max(prange)))[0]
    p_initsamplepool = p_true_l_t[inds_range]
    v_initsamplepool = v_true_l_t[inds_range]
    v_scrsamplepoolref = v_scr_h_t[inds_range]
    
    # # LP VERSION: Make a pool of samples (make a copy to avoid removing samples from the true LF data later)
    # p_initsamplepool = p_true_l_t.copy()
    # v_initsamplepool = v_true_l_t.copy()
    # v_scrsamplepoolref = v_scr_h_t.copy()
        
    # Check and print how many events in the HF MCS are within prediction range
    noevents_in_range = len(inds_range)
    noevents_in_range_nonnan = len(v_scrsamplepoolref[~np.isnan(v_scrsamplepoolref)])
    print(f'There are {noevents_in_range} events within probability range "prange", of which {noevents_in_range_nonnan} HF events are not equal to NaN')
    if noevents_in_range_nonnan < (max(maxit_HFGPR,maxit_MFGPR)+len(initplevels)):
        raise Exception(f'The number of non-NaN HF events within probability range "prange" ({noevents_in_range_nonnan}) is lower than the number of iterations specified (for either GPR or MF-GPR) plus the number of initial samples. In other words, there are not enough events within the specified range.') 
        # LP VERSION: print(f'WARNING: the number of non-NaN HF events within probability range "prange" ({noevents_in_range_nonnan}) is lower than the number of iterations specified (for either GPR or MF-GPR) plus the number of initial samples. In other words, there are not enough events within the specified range.') 

    # If initial samples are inside the available pool, exclude them to choose from later in adaptive sampling
    # Important: use prange limits here instead of sample pool limits, to avoid selecting the same point later (which in turn can introduce GPR issues)
    for cc,p_lev in enumerate(initplevels):
        if (p_lev >= min(prange)) & (p_lev <= max(prange)):
            i_sel_dist_pool,_ = find_nearest(p_initsamplepool,p_lev)
            p_initsamplepool = np.delete(p_initsamplepool,i_sel_dist_pool)
            v_initsamplepool = np.delete(v_initsamplepool,i_sel_dist_pool)

    print(f'Initial sampling done: {txtadd}')
    
    ## ------- Plot initial true functions & training data
    # Separately for HF-GPR and MF-GPR, as different samples will be added in the
    # adaptive sampling procedures for both. Update these plots later with every 
    # iteration (add the new points).
    
    N_hf = len(v_train_h)
    
    # Subsample the true and screening data specifically for the scatter plot
    plotmax_hf_norm = plotmax/norm_hf
    xbinsize = plotmax_hf_norm/500
    ybinsize = plotmax_hf_norm/500
    xbins = np.arange(0,plotmax_hf_norm,xbinsize)
    ybins = np.arange(0,plotmax_hf_norm,ybinsize)
    peaks_LF_all_sampled,peaks_HF_all_scr_sampled = subsample_binned_2D(peaks_LF_all,peaks_HF_all_scr,xbins,ybins)
    
    #  HF-GPR: Scatter plot indicator vs NL response peaks, including initial samples
    plt.figure(5,figsize=(fsize_h,0.8*fsize_v))
    plt.yscale('linear')
    plt.plot(peaks_LF_all_sampled*norm_lf,peaks_HF_all_scr_sampled*norm_hf,marker='o',color='grey',label='All peaks',linestyle = 'None')
    plt.xlabel(r'$\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (varsymb_lf.upper(),varunit_lf), fontsize=12)
    plt.ylabel(r'Corresponding $\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (varsymb_hf.upper(),varunit_hf), fontsize=12)
    plt.xlim([0,plotmax_lf]); plt.ylim([0,plotmax])
    plt.plot(v_train_l_REF*norm_lf,v_train_h*norm_hf,marker='x',color='red',markersize = 10,label='HF samples',linestyle = 'None')
    for cc in range(N_hf):
        plt.text(v_train_l_REF[cc]*norm_lf+0.05,v_train_h[cc]*norm_hf,'init')
    plt.legend()
    plt.savefig(f'{figfolder}/HFGPR_Scatter_{casename}_{txtadd}_InitSamps.pdf',bbox_inches='tight')
    
    # MF-GPR: Scatter plot indicator vs NL response peaks, including initial samples
    plt.figure(6,figsize=(fsize_h,0.8*fsize_v))
    plt.yscale('linear')
    plt.plot(peaks_LF_all_sampled*norm_lf,peaks_HF_all_scr_sampled*norm_hf,marker='o',color='grey',label='All peaks',linestyle = 'None')
    plt.xlabel(r'$\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (varsymb_lf.upper(),varunit_lf), fontsize=12)
    plt.ylabel(r'Corresponding $\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (varsymb_hf.upper(),varunit_hf), fontsize=12)
    plt.xlim([0,plotmax_lf]); plt.ylim([0,plotmax])
    plt.plot(v_train_l_REF*norm_lf,v_train_h*norm_hf,marker='x',color='red',markersize = 10,label='HF samples',linestyle = 'None')
    for cc in range(N_hf):
        plt.text(v_train_l_REF[cc]*norm_lf+0.05,v_train_h[cc]*norm_hf,'init')
    plt.legend()
    plt.savefig(f'{figfolder}/MFGPR_Scatter_{casename}_{txtadd}_InitSamps.pdf',bbox_inches='tight')
    
    # Scale factor HF to LF
    v_train_h_MOD = v_train_h/plotmax*plotmax_lf
    
    # HF-GPR: True and training data plot, including initial samples (use secondary x-axis for training data LF and HF)
    fig88 = plt.figure(88,figsize=(fsize_h,fsize_v))
    plt.yscale('log')
    plt.ylabel(r'$\mathrm{P(}\mathrm{%s} \geq \mathrm{%s} \mathrm{)}$   and   $\mathrm{P(}\mathrm{%s} \geq \mathrm{%s} \mathrm{)}$' % (varsymb_hf.upper(),varsymb_hf,varsymb_lf.upper(),varsymb_lf), fontsize=12)
    # -- x-axis1 (HF data)
    aX1a = plt.subplot()
    lnd1 = aX1a.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None',label='HF "true" MCS (normally N/A)')
    lnd2 = aX1a.plot(v_train_h*norm_hf,p_train_h,linestyle='None',color='r',marker='d',label='Resulting HF samples')
    lnd3 = aX1a.plot([0,plotmax],[pexp,pexp],':',color='black') #,label='$P_{exp}$')
    aX1a.text(locpexp,pexp*1.4,'$P_{exp}$')
    aX1a.plot(v_train_h*norm_hf,p_train_h,linestyle='None',color='r',marker='d') # plot repeated on top, no label
    aX1a.set_xlabel(r'$\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (varsymb_hf,varunit_hf), fontsize=12, color='red')
    aX1a.set_xlim(left=0,right=plotmax)
    # -- x-axis2 (LF data)
    aX2a = aX1a.twiny()
    lnd4 = aX2a.plot(v_true_l_sampled,p_true_l_sampled,marker='.',color=(0.7,0.7,0.7),linestyle='None',label='LF "true" MCS') 
    ind6 = aX2a.plot(v_train_l_REF*norm_lf,p_train_l_REF,marker='x',markerfacecolor='none',color='grey',linestyle='None',label='LF basis for HF sampling')
    aX2a.set_xlabel(r'$\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (varsymb_lf,varunit_lf), fontsize=12, color=(0.4,0.4,0.4))
    aX2a.set_xlim(left=0,right=plotmax_lf)
    for bb,_ in enumerate(v_train_h): # add connection lines
        lnd7 = aX2a.plot([v_train_l_REF[bb]*norm_lf,v_train_h_MOD[bb]*norm_hf],[p_train_l_REF[bb],p_train_h[bb]],linestyle='--',color='red',label='Screening connections')    
    for cc in range(N_hf):
        aX2a.text(v_train_l_REF[cc]*norm_lf+0.05,p_train_l_REF[cc],'init')  
    # -- y-axis
    plt.ylim([minP,1])
    lnd = lnd4+ind6+lnd7+lnd2+lnd1
    labs = [l.get_label() for l in lnd]
    aX1a.legend(lnd,labs,loc='lower left')
    plt.savefig(f'{figfolder}/MFGPR_TrainData_{casename}_{txtadd}_inclcon_InitSamps.pdf',bbox_inches='tight')    
   
    # MF-GPR: True and training data plot, including initial samples (use secondary x-axis for training data LF and HF)
    fig99 = plt.figure(99,figsize=(fsize_h,fsize_v))
    plt.yscale('log')
    plt.ylabel(r'$\mathrm{P(}\mathrm{%s} \geq \mathrm{%s} \mathrm{)}$   and   $\mathrm{P(}\mathrm{%s} \geq \mathrm{%s} \mathrm{)}$' % (varsymb_hf.upper(),varsymb_hf,varsymb_lf.upper(),varsymb_lf), fontsize=12)
    # -- x-axis1 (HF data)
    aX1b = plt.subplot()
    lnd1 = aX1b.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None',label='HF "true" MCS (normally N/A)')
    lnd2 = aX1b.plot(v_train_h*norm_hf,p_train_h,linestyle='None',color='r',marker='d',label='Resulting HF samples')
    lnd3 = aX1b.plot([0,plotmax],[pexp,pexp],':',color='black') #,label='$P_{exp}$')
    aX1b.text(locpexp,pexp*1.4,'$P_{exp}$')
    aX1b.plot(v_train_h*norm_hf,p_train_h,linestyle='None',color='r',marker='d') # plot repeated on top, no label
    aX1b.set_xlabel(r'$\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (varsymb_hf,varunit_hf), fontsize=12, color='red')
    aX1b.set_xlim(left=0,right=plotmax)
    # -- x-axis2 (LF data)
    aX2b = aX1b.twiny()
    lnd4 = aX2b.plot(v_true_l_sampled,p_true_l_sampled,marker='.',color=(0.7,0.7,0.7),linestyle='None',label='LF "true" MCS') 
    ind6 = aX2b.plot(v_train_l_REF*norm_lf,p_train_l_REF,marker='x',markerfacecolor='none',color='grey',linestyle='None',label='LF basis for HF sampling')
    aX2b.set_xlabel(r'$\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (varsymb_lf,varunit_lf), fontsize=12, color=(0.4,0.4,0.4))
    aX2b.set_xlim(left=0,right=plotmax_lf)
    for bb,_ in enumerate(v_train_h): # add connection lines
        lnd7 = aX2b.plot([v_train_l_REF[bb]*norm_lf,v_train_h_MOD[bb]*norm_hf],[p_train_l_REF[bb],p_train_h[bb]],linestyle='--',color='red',label='Screening connections')    
    for cc in range(N_hf):
        aX2b.text(v_train_l_REF[cc]*norm_lf+0.05,p_train_l_REF[cc],'init')  
    # -- y-axis
    plt.ylim([minP,1])
    lnd = lnd4+ind6+lnd7+lnd2+lnd1
    labs = [l.get_label() for l in lnd]
    aX1b.legend(lnd,labs,loc='lower left')
    plt.savefig(f'{figfolder}/MFGPR_TrainData_{casename}_{txtadd}_inclcon_InitSamps.pdf',bbox_inches='tight')

    
    ## ------- Write true and initial training data to file
    
    # Prepare data
    dat1 = np.transpose(np.stack((np.squeeze(v_true_l*norm_lf),np.squeeze(p_true_l))))
    dat2 = np.transpose(np.stack((np.squeeze(v_true_h*norm_hf),np.squeeze(p_true_h))))
    dat3 = np.transpose(np.stack((np.squeeze(v_train_l*norm_lf),np.squeeze(p_train_l))))
    dat4 = np.transpose(np.stack((np.squeeze(v_train_h*norm_hf),np.squeeze(p_train_h))))
    
    # Write CSV file using pandas
    my_df1 = pd.DataFrame(dat1)
    my_df1.to_csv(f'{filefolder}/TrueLFdata_{casename}_{txtadd}.csv',header = False, index= False)
    my_df2 = pd.DataFrame(dat2)
    my_df2.to_csv(f'{filefolder}/TrueHFdata_{casename}_{txtadd}.csv',header = False, index= False)
    my_df3 = pd.DataFrame(dat3)
    my_df3.to_csv(f'{filefolder}/TrainLFdata_{casename}_{txtadd}.csv',header = False, index= False)
    my_df4 = pd.DataFrame(dat4)
    my_df4.to_csv(f'{filefolder}/TrainHFdata_init_{casename}_{txtadd}.csv',header = False, index= False)
        
    print('Writing data to files done')
    
        
    ## ------- Modify true functions and training data to emphasise tail (operate on log)
    
    p_true_l_log = np.log(p_true_l)
    p_true_h_log = np.log(p_true_h)
    p_train_l_log = np.log(p_train_l)
    p_train_h_log = np.log(p_train_h)
        
    # Adaptive sampling
    p_initsamplepool_log = np.log(p_initsamplepool)
    

    ## ------- 0D to vertical 1D arrays, as required by the GPy / Emukit toolboxes
    v_true_l = np.transpose([v_true_l])
    p_true_l = np.transpose([p_true_l])
    p_true_l_log = np.transpose([p_true_l_log])
    v_train_l = np.transpose([v_train_l])
    p_train_l = np.transpose([p_train_l])
    p_train_l_log = np.transpose([p_train_l_log])
    v_true_h = np.transpose([v_true_h])
    p_true_h = np.transpose([p_true_h])
    p_true_h_log = np.transpose([p_true_h_log])
    v_train_h = np.transpose([v_train_h])
    p_train_h = np.transpose([p_train_h])
    p_train_h_log = np.transpose([p_train_h_log])
    
    print('All preparations for (MF-)GPR done')
    print('From here we start with adaptive sampling, separately for HF-GPR and MF-GPR')
    
    
    #########################################
    ## ------- HF-GPR application, including adaptive sampling 
    
    if maxit_HFGPR != 0:
        
        # Prepare CSV file where MPM results will be stored over iterations
        csv_filename_gpr = f'{filefolder}/Conv_HF-GPR_{casename}_Set-{HFkern}_{txtadd}_{acqtypeHF}_Nhf{N_hf}.csv'
        headHFGPR = np.array(['NoSamp','MPM_HFGPR','U95_MPM_HFGPR','SumVar_HFGPR','MeanVar_HFGPR'])
        np.savetxt(csv_filename_gpr,[headHFGPR],delimiter=',',fmt='%s')
        
        # Save prediction value range, this does not change with the loops (if it does, 
        # add to initialised arrays below). Do not keep these variables coupled, to 
        # avoid coupling of the HF-GPR and MF-GPR results.
        HFGPR_P1_log = Prange_predHF_log.copy() 
        p_samplepool1_log = p_initsamplepool_log.copy()
        v_samplepool1 = v_initsamplepool.copy()
        
        # Debug plot: check that sample pool is within prange
        if debugplotswitch == 1:
            plt.figure(10000000)
            plt.plot(v_samplepool1*norm_hf,p_samplepool1_log,'.-')
            plt.plot([0,plotmax],[np.min(Prange_predHF_log),np.min(Prange_predHF_log)],'--k')
            plt.plot([0,plotmax],[np.max(Prange_predHF_log),np.max(Prange_predHF_log)],'--k')
            plt.title('GPR: selected points on log(p)')
                
        # Initialise lists and arrays to save HF-GPR results as function of number of loops
        p_train_h1_arr_log = []
        v_train_h1_arr = []
        HFGPR_mean_arr_log = []
        HFGPR_var_arr_log = []
        
        ass_MPM_hfGPR = np.array([])
        ass_MPM_hfGPR_u95 = np.array([])
        ass_sumvar_pred_hfGPR = np.array([])
        ass_meanvar_pred_hfGPR = np.array([])
        
        HP_vals_gpr = []
        
        # Define input variables separately for HF-GPR
        p_train_h1_log = p_train_h_log.copy()
        v_train_h1 = v_train_h.copy()
        
        # Initialise
        count1 = 0
        MPM_hfGPR_U95 = 10000 # high value so the while loop starts
        
        # Start while loop
        while count1+N_hf <= Nmax and count1 <= maxit_HFGPR:
            
            # Perform HF-GPR with present samples
            print(f'HF-GPR adaptive sampling loop number {count1} with {len(p_train_h1_log)} HF samples')
                        
            # Create standard GP model using only high-fidelity data
            high_gp_model = GPy.models.GPRegression(p_train_h1_log,v_train_h1,kernel_HFgp)
            
            # Estimate homoscedastic noise variance around the available samples, using least-squares linear interpolation
            if len(v_train_h1) > 2:
                A = np.vstack([np.squeeze(p_train_h1_log), np.ones(len(p_train_h1_log))]).T
                m, c = np.linalg.lstsq(A,v_train_h1,rcond=None)[0]
                v_lin = m*p_train_h1_log+c
                v_resid = v_train_h1-v_lin
                est_var = np.var(v_resid)
            else: 
                est_var = 1
     
            # HF GPR noise definition
            high_gp_model.constrain_positive("") # constrain all parameters to positive values
            high_gp_model.Gaussian_noise.constrain_bounded(0.5*est_var,10000*est_var)
                   
            # Constrain the kernel length scale
            Llow_gpr = max([10,0.1*len(v_train_h1)])
            Lhig_gpr = 1000
            high_gp_model['.*lengthscale'].constrain_bounded(Llow_gpr,Lhig_gpr)
            
            # Wrap with emukit (necessary for adaptive sampling)
            emukit_high_gp_model = GPyModelWrapper(high_gp_model)
           
            # Fit the model using Emukit: Fit the GP model with the initial samples
            emukit_high_gp_model.optimize()
            
            # Access optimized hyperparameter names and values at the present iteration, and save them.
            names_gpr,vals_gpr = present_hyperparams(high_gp_model)
            if count1 == 0:
                HP_names_gpr = names_gpr.copy()
            HP_vals_gpr.append(vals_gpr)
            
            # Using Emukit: Compute mean predictions and associated variance and U95
            mean_high_gp_model, var_high_gp_model = emukit_high_gp_model.predict(HFGPR_P1_log)
                
            # Calculate MPM and its uncertainty U95%
            U95_high_gp_model = 1.96*np.sqrt(var_high_gp_model)
            MPM_hfGPR = np.interp(pexp,np.squeeze(Prange_predHF),np.squeeze(mean_high_gp_model))
            MPM_hfGPR_U95 = np.interp(pexp,np.squeeze(Prange_predHF),np.squeeze(U95_high_gp_model))
            
            # Save full results of present iteration to list
            p_train_h1_arr_log.append(p_train_h1_log)
            v_train_h1_arr.append(v_train_h1)
            HFGPR_mean_arr_log.append(mean_high_gp_model)
            HFGPR_var_arr_log.append(var_high_gp_model)
            
            # Save MPM and its uncertainty to array
            ass_MPM_hfGPR = np.append(ass_MPM_hfGPR,MPM_hfGPR)
            ass_MPM_hfGPR_u95 = np.append(ass_MPM_hfGPR_u95,MPM_hfGPR_U95)
            
            # Also save overall uncertainty metric (integrated variance over prediction range) to array
            sum_var_pred_hfGPR = np.sum(var_high_gp_model)
            ass_sumvar_pred_hfGPR = np.append(ass_sumvar_pred_hfGPR,sum_var_pred_hfGPR)
            mean_var_pred_hfGPR = np.mean(var_high_gp_model)
            ass_meanvar_pred_hfGPR = np.append(ass_meanvar_pred_hfGPR,mean_var_pred_hfGPR)
            
            ## ------- Write present training data and prediction to file
          
            # Prepare data
            p_train_h1 = np.exp(p_train_h1_log)
            dat_trainHF_hfpgr = np.transpose(np.stack((np.squeeze(v_train_h1*norm_hf),np.squeeze(p_train_h1))))
            dat_predHF_hfgpr  = np.transpose(np.stack((np.squeeze(np.exp(Prange_predHF_log)),np.squeeze(mean_high_gp_model*norm_hf),np.squeeze(var_high_gp_model*norm_hf**2))))
        
            # Write CSV file for present iteration using pandas
            df_trainHF_hfpgr = pd.DataFrame(dat_trainHF_hfpgr)
            head_trainHF_hfpgr = ['v_train_h1','p_train_h1']
            df_trainHF_hfpgr.to_csv(f'{filefolder}/HFGPR_TrainData_{casename}_Set-{HFkern}-nf_{txtadd}_HFacq-{acqtypeHF}_Nhf{N_hf}_HFit{count1}.csv',header = head_trainHF_hfpgr, index= False)
            df_predHF_hfgpr = pd.DataFrame(dat_predHF_hfgpr)
            head_predHF_hfgpr = ['Prange_predHF','mean_high_gp_model','var_high_gp_model']
            df_predHF_hfgpr.to_csv(f'{filefolder}/HFGPR_HFpred_{casename}_Set-{HFkern}-nf_{txtadd}_HFacq-{acqtypeHF}_Nhf{N_hf}_HFit{count1}.csv',header = head_predHF_hfgpr, index= False)
                   
            # Append convergence data to file (iteration number, MPM, U95 of MPM, sumvar and meanvar)
            nosamps = N_hf+len(ass_MPM_hfGPR)-1
            newdat_conv_HFGPR = np.array([nosamps,MPM_hfGPR*norm_hf,MPM_hfGPR_U95*norm_hf,sum_var_pred_hfGPR*norm_hf,mean_var_pred_hfGPR*norm_hf])
            with open(csv_filename_gpr,'a') as file:
                np.savetxt(file, [newdat_conv_HFGPR], delimiter=',', fmt='%s')

            # Do this for each loop, except the last one (we don't need a new sample in that case)
            # Defining it this way saves the double notation of the whole GPR procedure (before & inside the loop)
            # (Only works for the maximum # iterations stopping criterion)
            # if MPM_hfGPR_U95 > c_u95_MPM and count1+N_hf <= Nmax and count1 <= maxit_HFGPR-1:
            if count1 < maxit_HFGPR:
                
                # Define emukit parameter space object (required for the IVR function and optimizer)
                EMU_space_log = ParameterSpace([ContinuousParameter('P', np.min(HFGPR_P1_log), np.max(HFGPR_P1_log))]) 
                
                ################ DEBUGGING
                # emukit_high_gp_model.calculate_variance_reduction
                ################ 
                
                # Define acquisition function HF-GPR
                if acqtypeHF == 'US':
                    func_acquisition = ModelVariance(emukit_high_gp_model)
                elif acqtypeHF == 'IVR':
                    func_acquisition = IntegratedVarianceReduction(emukit_high_gp_model,EMU_space_log)
                elif acqtypeHF == 'MV':
                    func_acquisition = MaxV(emukit_high_gp_model)
                elif acqtypeHF == 'USMV':
                    func_acquisition_p1 = ModelVariance(emukit_high_gp_model)
                    func_acquisition_p2 = MaxV(emukit_high_gp_model)
                    func_acquisition = func_acquisition_p1.__rmul__(func_acquisition_p2) # Product of p1 and p2
                elif acqtypeHF == 'IVRMV':
                    func_acquisition_p1 = IntegratedVarianceReduction(emukit_high_gp_model,EMU_space_log)
                    func_acquisition_p2 = MaxV(emukit_high_gp_model)
                    func_acquisition = func_acquisition_p1.__rmul__(func_acquisition_p2) # Product of p1 and p2
                elif acqtypeHF == 'USPD':
                    func_acquisition_p1 = ModelVariance(emukit_high_gp_model)
                    func_acquisition_p2 = ParabPdist(np.log(pexp))
                    func_acquisition = func_acquisition_p1.__rmul__(func_acquisition_p2) # Product of p1 and p2
                elif acqtypeHF == 'IVRPD':
                    func_acquisition_p1 = IntegratedVarianceReduction(emukit_high_gp_model,EMU_space_log)
                    func_acquisition_p2 = ParabPdist(np.log(pexp))
                    func_acquisition = func_acquisition_p1.__rmul__(func_acquisition_p2) # Product of p1 and p2    
                elif acqtypeHF == 'MD':
                    func_acquisition = MaxDist(p_train_h1_log)
                elif acqtypeHF == 'MVPD':
                    func_acquisition_p1 = MaxV(emukit_high_gp_model)
                    func_acquisition_p2 = ParabPdist(np.log(pexp))
                    func_acquisition = func_acquisition_p1.__rmul__(func_acquisition_p2) # Product of p1 and p2
                
                # Now evaluate this acquisition function on the target probability range
                func_acq_plot = func_acquisition.evaluate(HFGPR_P1_log)
                
                # Debugging
                if debugplotswitch == 1:
                    plt.figure()
                    plt.plot(HFGPR_P1_log,func_acq_plot)
                    plt.title(f'GPR, {count1}')
                
                # Based on results, define a new sample
                # To find the next point to evaluate we optimize the acquisition function using 
                # a standard gradient descent optimizer. It seems like this finds the coordinate 
                # of the max. value of the acq function in 1D.
                # optimizer = GradientAcquisitionOptimizer(EMU_space_log) # SOMETIMES GOES WRONG
                # p_new_sample_log_target, _ = optimizer.optimize(func_acquisition) # SOMETIMES GOES WRONG
                p_new_sample_log_target = HFGPR_P1_log[np.argmax(func_acq_plot),0]
               
                # Find new point that is not equal to NaN 
                v_train_h1_new = np.nan
                while np.isnan(v_train_h1_new): 
                    # Find closest sample in the LF screening MCS result
                    i_new_sample_log,p_new_sample_log = find_nearest(p_samplepool1_log,p_new_sample_log_target)
                    # print(i_new_sample_log)
                    
                    # Selected LF points where we pick the matched HF points 
                    p_train_l1_log_REF_new = p_samplepool1_log[i_new_sample_log] # = p_new_sample_log
                    v_train_l1_REF_new = v_samplepool1[i_new_sample_log]
                    
                    # Now we have to get back to the original time series peak matching to find the
                    # corresponding HF value > this is the HF sample
                    i_sel_time_new = np.where(peaks_LF_all == v_train_l1_REF_new)
                    v_train_h1_new = peaks_HF_all_scr[i_sel_time_new]
                        
                    # HF probabilities are set identical to LF probabilities
                    p_train_h1_log_new = p_train_l1_log_REF_new.copy()
                    
                    # Remove newly chosen sample from the pool of samples for the next iteration
                    p_samplepool1_log = np.delete(p_samplepool1_log,i_new_sample_log)
                    v_samplepool1 = np.delete(v_samplepool1,i_new_sample_log) 
                
                # Debug
                if debugplotswitch == 1:
                    print(f'Iteration {count1}: new point close to log(p) = {p_new_sample_log_target}')
                    plt.figure(10000000)
                    plt.plot(v_train_l1_REF_new*norm_lf,p_train_l1_log_REF_new,'xr')
                    plt.text(v_train_l1_REF_new*norm_lf,p_train_l1_log_REF_new,f'{count1}') #,color='r')
                            
                print(f'HF-GPR adaptive: new sample number {count1} is defined')
            
                # Add the new samples to the scatter plot
                plt.figure(5)
                plt.plot(v_train_l1_REF_new*norm_lf,v_train_h1_new*norm_hf,marker='x',color='red',markersize = 10,linestyle = 'None')
                plt.text(v_train_l1_REF_new*norm_lf+0.05,v_train_h1_new*norm_hf,f'New sample {count1+1}')
                            
                # Scale factor HF to LF
                v_train_h1_new_MOD = v_train_h1_new/norm_hf*norm_lf
                
                # Add the new samples to the training data plot
                plt.figure(88)
                aX2a.plot(v_train_l1_REF_new*norm_lf,np.exp(p_train_l1_log_REF_new),marker='x',markerfacecolor='none',color='grey',linestyle='None') # ,label='LF basis for HF sampling')
                aX1a.text(v_train_h1_new*norm_hf+0.3,np.exp(p_train_h1_log_new),f'it{count1+1}',color='r')
                aX2a.plot([v_train_l1_REF_new*norm_lf,v_train_h1_new_MOD*norm_hf],np.exp([p_train_l1_log_REF_new,p_train_h1_log_new]),linestyle='--',color='red')
                aX1a.plot(v_train_h1_new*norm_hf,np.exp(p_train_h1_log_new),linestyle='None',color='r',marker='d') # ,label='Resulting HF samples')
                
                # Append new sample to existing HF samples
                p_train_h1_log = np.append(p_train_h1_log,p_train_h1_log_new)
                v_train_h1 = np.append(v_train_h1,v_train_h1_new)
                
                # Add dimension such that we have a column array
                p_train_h1_log = p_train_h1_log[:,None]
                v_train_h1 = v_train_h1[:,None]
            
            # --------- Add to counter to continue the while loop, unless we reached the max. # seeds
            count1 = count1+1
                         
        # Number of last iteration
        lastcount1 = count1-1
        
        # Inverse the log of the final samples and results
        p_train_h1 = np.exp(p_train_h1_log)
                
        # Save final scatter plot and sample distribution plot
        if lastcount1 > 5:
            plt.figure(5)
            plt.savefig(f'{figfolder}/HFGPR_Scatter_{casename}_{txtadd}_{acqtypeHF}_Nhf{N_hf}_it{lastcount1}.pdf',bbox_inches='tight')
            
            plt.figure(88)
            plt.savefig(f'{figfolder}/HFGPR_TrainData_{casename}_{txtadd}_inclcon_{acqtypeHF}_Nhf{N_hf}_it{lastcount1}.pdf',bbox_inches='tight')


        # Plot the posterior mean and variance (U95) from the reference HF gaussian process
        fig200 = plt.figure(200,figsize=(fsize_h,fsize_v))
        plt.yscale('log')
        plt.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None') #,label='HF "true" MCS')
        plt.plot(v_train_h1*norm_hf,p_train_h1, color='r', marker='d',linestyle='None',markersize=4, label=f'{len(v_train_h1)} HF samples')
        plt.plot(mean_high_gp_model*norm_hf,Prange_predHF,'-.',color='g',label='GPR prediction')
        plt.fill_betweenx(Prange_predHF.flatten(), (mean_high_gp_model*norm_hf - 0.5*U95_high_gp_model*norm_hf).flatten(), 
                         (mean_high_gp_model*norm_hf + 0.5*U95_high_gp_model*norm_hf).flatten(), facecolor='g', alpha=0.1,label='GPR prediction U95%')
        plt.plot([0,plotmax],[pexp,pexp],':',color='black') #,label='$P_{exp}$')
        plt.plot(v_true_h_sampled[0],p_true_h_sampled[0],marker='.',color='pink',linestyle='None',label='HF "true" MCS (validation)') # To put this one at the end of the legend
        plt.text(locpexp,pexp*1.4,'$P_{exp}$')
        plt.xlabel(r'$\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (varsymb_hf,varunit_hf), fontsize=12)
        plt.ylabel(r'$\mathrm{P(}\mathrm{%s} \geq \mathrm{%s} \mathrm{)}$' % (varsymb_hf.upper(),varsymb_hf), fontsize=12)
        plt.ylim([minP,1])
        plt.xlim([0,plotmax])
        plt.legend(loc='best')
        if lastcount1 > 5:
            plt.savefig(f'{figfolder}/Prediction_HFGPR_{casename}_Set-{HFkern}-nf_{txtadd}_{acqtypeHF}_Nhf{N_hf}_it{lastcount1}.pdf',bbox_inches='tight')
             
        itrangeHF = np.arange(N_hf,N_hf+len(ass_MPM_hfGPR),1)
        print('HF single-fidelity GPR done')
    
    
    
    #########################################
    ## ------- MF-GPR application, including adaptive sampling 
    
    if maxit_MFGPR != 0:
        
        # Prepare CSV file where MPM results will be stored over iterations
        csv_filename_mfgpr = f'{filefolder}/Conv_MF-GPR_{casename}_Set-{MFkern1}-{MFkern2}_{txtadd}_{acqtypeMF}_Nhf{N_hf}.csv'
        headMFGPR = np.array(['NoSamp','MPM_MFGPR','U95_MPM_MFGPR','SumVar_MFGPR','MeanVar_MFGPR'])
        np.savetxt(csv_filename_mfgpr,[headMFGPR],delimiter=',',fmt='%s')
        
        # Save prediction value range, this does not change with the loops (if it does, 
        # add to initialised arrays below). Do not keep these variables coupled, to 
        # avoid coupling of the HF-GPR and MF-GPR results.
        MFGPR_P2_LF_log = Prange_predLF_log.copy() 
        MFGPR_P2_HF_log = Prange_predHF_log.copy() 
        p_samplepool2_log = p_initsamplepool_log.copy()
        v_samplepool2 = v_initsamplepool.copy()
        
        # Debug plot: check that sample pool is within prange
        if debugplotswitch == 1:
            plt.figure(20000000)
            plt.plot(v_samplepool2*norm_hf,p_samplepool2_log,'.-')
            plt.plot([0,14],[np.min(Prange_predHF_log),np.min(Prange_predHF_log)],'--k')
            plt.plot([0,14],[np.max(Prange_predHF_log),np.max(Prange_predHF_log)],'--k')
            plt.title('MF-GPR: selected points on log(p)')
                
        # Initialise lists and arrays to save HF-GPR results as function of number of loops
        p_train_h2_arr_log = []
        v_train_h2_arr = []
        MFGPR_mean_arr_LF_log = []
        MFGPR_var_arr_LF_log = []
        MFGPR_mean_arr_HF_log = []
        MFGPR_var_arr_HF_log = []
        
        ass_MPM_mfGPR = np.array([])
        ass_MPM_mfGPR_u95 = np.array([])
        ass_sumvar_pred_mfGPR = np.array([])
        ass_meanvar_pred_mfGPR = np.array([])
        
        HP_vals_mfgpr = []
        
        # Define input variables separately for MF-GPR
        p_train_l2_log = p_train_l_log.copy()
        v_train_l2 = v_train_l.copy()
        p_train_h2_log = p_train_h_log.copy()
        v_train_h2 = v_train_h.copy()
        
        # Create multi-fidelity GP model using LF and HF data
        lin_mf_kernel = emukit.multi_fidelity.kernels.LinearMultiFidelityKernel(kernels_MFgp)
        
        # Initialise
        count2 = 0
        MPM_mfGPR_U95 = 10000 # high value so the while loop starts
        
        # Start while loop
        # while MPM_mfGPR_U95 > c_u95_MPM and count2+N_hf <= Nmax and count2 <= maxit_MFGPR:
        while count2+N_hf <= Nmax and count2 <= maxit_MFGPR:
            
            print(f'MF-GPR adaptive sampling loop number {count2} with {len(p_train_h2_log)} HF samples')
            
            # Perform MF-GPR with present samples
              
            # Linear multi-fidelity model (proposed by Kennedy and O'Hagen (2000), similar to Forrester et al (2007)):
            # F_HF = F_err + rho * f_LF
            # f_LF is from GP of LF model(s)
            # rho is correlation between HF and LF (0 = no correlation)
            # f_err is GP for bias term for HiFi data
        
            # Convert lists of arrays to ndarrays augmented with fidelity indicators
            P_train2_log,V_train2  = convert_xy_lists_to_arrays([p_train_l2_log,p_train_h2_log],[v_train_l2,v_train_h2])
            
            # Make a linear MF GPR (modified version based on Emukit, that also stores the models for the fidelity levels)
            mf_gp_model = GPyLinearMultiFidelityModel(P_train2_log,V_train2,lin_mf_kernel,n_fidelities=2)
            # mf_gp_model = GPyLinearMultiFidelityModel_mod(P_train2_log,V_train2,lin_mf_kernel,n_fidelities=2) # DOES NOT WORK
            
            # Estimate homoscedastic noise variance around the available samples, using least-squares linear interpolation
            if len(v_train_l2) > 2:
                A_lf = np.vstack([np.squeeze(p_train_l2_log), np.ones(len(p_train_l2_log))]).T
                m_lf,c_lf = np.linalg.lstsq(A_lf,v_train_l2,rcond=None)[0]
                v_LF_lf = m_lf*p_train_l2_log+c_lf
                v_resid_lf = v_train_l2-v_LF_lf
                est_var_lf = np.var(v_resid_lf)
            else: 
                est_var_lf = 1
            
            if len(v_train_h2) > 2:
                A_hf = np.vstack([np.squeeze(p_train_h2_log), np.ones(len(p_train_h2_log))]).T
                m_hf,c_hf = np.linalg.lstsq(A_hf,v_train_h2,rcond=None)[0]
                v_LF_hf = m_hf*p_train_h2_log+c_hf
                v_resid_hf = v_train_h2-v_LF_hf
                est_var_hf = np.var(v_resid_hf)
            else:
                est_var_hf = 1
            
            # MF GPR noise definition 
            mf_gp_model.constrain_positive("") # constrain all parameters to positive values
            mf_gp_model.mixed_noise.Gaussian_noise.constrain_bounded(0.5*est_var_lf,10000*est_var_lf)
            mf_gp_model.mixed_noise.Gaussian_noise_1.constrain_bounded(0.5*est_var_hf,10000*est_var_hf)
                  
            # Constrain the kernel length scales
            Llow_mfgpr = max([10,0.1*len(v_train_h2)])
            Lhig_mfgpr = 1000
            mf_gp_model['.*lengthscale'].constrain_bounded(Llow_mfgpr,Lhig_mfgpr)
            
            # Wrap model using modified class 'GPyMultiOutputWrapper_mod'
            emukit_LF_mf_model = GPyMultiOutputWrapper(mf_gp_model,2,n_optimization_restarts=5) # DOES NOT WORK
           
            # Fit the model
            emukit_LF_mf_model.optimize()
        
            # Access optimized hyperparameter names and values at the present iteration, and save them.
            names_mfgpr,vals_mfgpr = present_hyperparams(mf_gp_model)
            if count2 == 0:
                HP_names_mfgpr = names_mfgpr.copy()
            HP_vals_mfgpr.append(vals_mfgpr)
            
            # Convert x_plot to its ndarray representation
            P_plot2_log = convert_x_list_to_array([MFGPR_P2_LF_log, MFGPR_P2_HF_log])
            P_plot_l2_log = P_plot2_log[:len(MFGPR_P2_LF_log)]
            P_plot_h2_log = P_plot2_log[len(MFGPR_P2_HF_log):]
        
            # Compute mean predictions and associated variance for LF and HF 
            mean_low_mfgp_model, var_low_mfgp_model = emukit_LF_mf_model.predict(P_plot_l2_log)
            std_low_mfgp_model = np.sqrt(var_low_mfgp_model)
            U95_low_mfgp_model = 1.96*std_low_mfgp_model
        
            mean_high_mfgp_model, var_high_mfgp_model = emukit_LF_mf_model.predict(P_plot_h2_log)
            std_high_mfgp_model = np.sqrt(var_high_mfgp_model)
            U95_high_mfgp_model = 1.96*std_high_mfgp_model
            
            # Calculate MPM and its uncertainty U95
            MPM_mfGPR = np.interp(pexp,np.squeeze(Prange_predHF),np.squeeze(mean_high_mfgp_model))
            MPM_mfGPR_U95 = np.interp(pexp,np.squeeze(Prange_predHF),np.squeeze(U95_high_mfgp_model))
                  
            # Save full results of present iteration to list
            p_train_h2_arr_log.append(p_train_h2_log)
            v_train_h2_arr.append(v_train_h2)
            MFGPR_mean_arr_LF_log.append(mean_low_mfgp_model)
            MFGPR_var_arr_LF_log.append(var_low_mfgp_model)
            MFGPR_mean_arr_HF_log.append(mean_high_mfgp_model)
            MFGPR_var_arr_HF_log.append(var_high_mfgp_model)
            
            # Save HF MPM and its uncertainty to array
            ass_MPM_mfGPR = np.append(ass_MPM_mfGPR,MPM_mfGPR)
            ass_MPM_mfGPR_u95 = np.append(ass_MPM_mfGPR_u95,MPM_mfGPR_U95)
            
            # Also save overall uncertainty metric (integrated variance over prediction range) to array
            sum_var_pred_mfGPR = np.sum(var_high_mfgp_model)
            ass_sumvar_pred_mfGPR = np.append(ass_sumvar_pred_mfGPR,sum_var_pred_mfGPR)
            mean_var_pred_mfGPR = np.sum(var_high_mfgp_model)
            ass_meanvar_pred_mfGPR = np.append(ass_meanvar_pred_mfGPR,mean_var_pred_mfGPR)
            
            ## ------- Write present training data and prediction to file
        
            # Prepare data
            dat_trainLF_mfpgr = np.transpose(np.stack((np.squeeze(v_train_l2*norm_lf),np.squeeze(np.exp(p_train_l2_log)))))
            dat_trainHF_mfpgr = np.transpose(np.stack((np.squeeze(v_train_h2*norm_hf),np.squeeze(np.exp(p_train_h2_log)))))
            dat_predLF_mfgpr  = np.transpose(np.stack((np.squeeze(np.exp(Prange_predLF_log)),np.squeeze(mean_low_mfgp_model*norm_lf),np.squeeze(var_low_mfgp_model*norm_lf**2))))
            dat_predHF_mfgpr  = np.transpose(np.stack((np.squeeze(np.exp(Prange_predHF_log)),np.squeeze(mean_high_mfgp_model*norm_hf),np.squeeze(var_high_mfgp_model*norm_hf**2))))
        
            # Write CSV file for present iteration using pandas
            df_trainLF_mfpgr = pd.DataFrame(dat_trainLF_mfpgr)
            head_trainLF_mfpgr = ['v_train_l2','p_train_l2']
            df_trainLF_mfpgr.to_csv(f'{filefolder}/MFGPR_LFTrainData_{casename}_Set-{MFkern1}-{MFkern2}-nf_{txtadd}_MFacq-{acqtypeMF}_Nhf{N_hf}_MFit{count2}.csv',header = head_trainLF_mfpgr, index= False)  
            
            df_trainHF_mfpgr = pd.DataFrame(dat_trainHF_mfpgr)
            head_trainHF_mfpgr = ['v_train_h2','p_train_h2']
            df_trainHF_mfpgr.to_csv(f'{filefolder}/MFGPR_HFTrainData_{casename}_Set-{MFkern1}-{MFkern2}-nf_{txtadd}_MFacq-{acqtypeMF}_Nhf{N_hf}_MFit{count2}.csv',header = head_trainHF_mfpgr, index= False)
             
            df_predLF_mfgpr = pd.DataFrame(dat_predLF_mfgpr)
            head_predLF_mfgpr = ['Prange_predLF','mean_low_mfgp_model','var_low_mfgp_model']
            df_predLF_mfgpr.to_csv(f'{filefolder}/MFGPR_LFpred_{casename}_Set-{MFkern1}-{MFkern2}-nf_{txtadd}_MFacq-{acqtypeMF}_Nhf{N_hf}_MFit{count2}.csv',header = head_predLF_mfgpr, index= False)
            
            df_predHF_mfgpr = pd.DataFrame(dat_predHF_mfgpr)
            head_predHF_mfgpr = ['Prange_predHF','mean_high_mfgp_model','var_high_mfgp_model']
            df_predHF_mfgpr.to_csv(f'{filefolder}/MFGPR_HFpred_{casename}_Set-{MFkern1}-{MFkern2}-nf_{txtadd}_MFacq-{acqtypeMF}_Nhf{N_hf}_MFit{count2}.csv',header = head_predHF_mfgpr, index= False)
            
            # Append convergence data to file (iteration number, MPM, U95 of MPM, sumvar and meanvar)
            nosamps = N_hf+len(ass_MPM_mfGPR)-1
            newdat_conv_MFGPR = np.array([nosamps,MPM_mfGPR*norm_hf,MPM_mfGPR_U95*norm_hf,sum_var_pred_mfGPR*norm_hf,mean_var_pred_mfGPR*norm_hf])
            with open(csv_filename_mfgpr,'a') as file:
                np.savetxt(file, [newdat_conv_MFGPR], delimiter=',', fmt='%s')
            
            
            # # Do this for each loop, except the last one (we don't need a new sample in that case)
            # # Defining it this way saves the double notation of the whole GPR procedure (before & inside the loop)
            # (Only works for the maximum # iterations stopping criterion)
            if count2 < maxit_MFGPR:  
           
                ######### WORKAROUND: use mock single-fidility GPRs as surrogate for the MF-GPR high-fidelity result
                # There seems to be no dedicated Emukit multi-fidelity GradientAcquisitionOptimizer, 
                # so translate the results for the HF level from the multi-fidelity 
                # 'emukit.model_wrappers.gpy_model_wrappers.GPyMultiOutputWrapper' object  to a 
                # single-fidelity 'emukit.model_wrappers.gpy_model_wrappers.GPyModelWrapper' object
                # There seems to be no dedicated function to do this, so first extract data manually.
                # Steps:
                # 1. Extract Data: Extract the mean & variance data for the desired 
                #    fidelity level from the original dataset.
                # 2. Train two single-Fidelity GPR models, separately on the mean and 
                #    variance data. Use a regular Gaussian Process (GPyModelWrapper) 
                # 3. Use these single-fidelity surrogates in the adaptive sampling
           
                # Kernel of the mock single-fidelity GPRs
                kern_mock = kernel_HFgp.copy()
                
                # MEAN Mock SF-GPR model for the HF output of the MF-GPR model
                mock_mean_sfgp_model = GPy.models.GPRegression(MFGPR_P2_HF_log,mean_high_mfgp_model,kern_mock)   
                mock_mean_sfgp_model.constrain_positive("") # constrain all parameters to positive values
                # mock_mean_sfgp_model.Gaussian_noise.constrain_bounded(0.5*np.var(mean_high_mfgp_model),10000*np.var(mean_high_mfgp_model))
                mock_mean_sfgp_model.Gaussian_noise.constrain_bounded(0.8*np.var(mean_high_mfgp_model),10000*np.var(mean_high_mfgp_model))
                # Wrap and optimise model
                emukit_mock_mean_model = GPyModelWrapper(mock_mean_sfgp_model)
                emukit_mock_mean_model.optimize()
                # Predict the surrogate process values
                mean_mock,_ = emukit_mock_mean_model.predict(MFGPR_P2_HF_log)
                
                # VARIANCE Mock SF-GPR model for the HF output of the MF-GPR model
                mock_var_sfgp_model = GPy.models.GPRegression(MFGPR_P2_HF_log,var_high_mfgp_model,kern_mock)   
                mock_var_sfgp_model.constrain_positive("") # constrain all parameters to positive values
                # mock_var_sfgp_model.Gaussian_noise.constrain_bounded(0.5*np.var(var_high_mfgp_model),10000*np.var(var_high_mfgp_model))
                mock_var_sfgp_model.Gaussian_noise.constrain_bounded(0.8*np.var(var_high_mfgp_model),10000*np.var(var_high_mfgp_model))
                # Wrap and optimise model
                emukit_mock_var_model = GPyModelWrapper(mock_var_sfgp_model)
                emukit_mock_var_model.optimize()
                # Predict the surrogate process values
                var_mock,_ = emukit_mock_var_model.predict(MFGPR_P2_HF_log) # this is right; we model only the variance, so the mean of this model is a surrogate for the variance of the MF-GPR model
                           
        
                ######### 
                        
                # Define acquisition function
                if acqtypeMF == 'US':
                    # Do not use ModelVariance, but the mean prediction of the variance mock surrogate model
                    func_acquisition2 = MaxV(emukit_mock_var_model)
                elif acqtypeMF == 'IVR':
                    # func_acquisition2 = IntegratedVarianceReduction(MODEL,EMU_space_log2)
                    raise Exception('Error: IVR acquisition function not yet implemented for MF-GPR')
                elif acqtypeMF == 'MV':
                    # Here we also use MaxV, but now applied to the mean prediction of the mean mock surrogate model
                    func_acquisition2 = MaxV(emukit_mock_mean_model)
                elif acqtypeMF == 'USMV':
                    func_acquisition_p1 = MaxV(emukit_mock_var_model)
                    func_acquisition_p2 = MaxV(emukit_mock_mean_model)
                    func_acquisition2 = func_acquisition_p1.__rmul__(func_acquisition_p2) # Product of p1 and p2
                elif acqtypeMF == 'IVRMV':
                    func_acquisition_p2 = MaxV(emukit_mock_mean_model)
                    func_acquisition2 = func_acquisition_p1.__rmul__(func_acquisition_p2) # Product of p1 and p2
                    raise Exception('Error: IVR acquisition function not yet implemented for MF-GPR')
                elif acqtypeMF == 'USPD':
                    func_acquisition_p1 = MaxV(emukit_mock_var_model)
                    func_acquisition_p2 = ParabPdist(np.log(pexp))
                    func_acquisition2 = func_acquisition_p1.__rmul__(func_acquisition_p2) # Product of p1 and p2
                elif acqtypeMF == 'IVRPD':
                    func_acquisition_p2 = ParabPdist(np.log(pexp))
                    func_acquisition2 = func_acquisition_p1.__rmul__(func_acquisition_p2) # Product of p1 and p2    
                    raise Exception('Error: IVR acquisition function not yet implemented for MF-GPR')
                elif acqtypeMF == 'MD':
                    func_acquisition2 = MaxDist(p_train_h2_log)
                elif acqtypeHF == 'MVPD':
                    func_acquisition_p1 = MaxV(emukit_mock_mean_model)
                    func_acquisition_p2 = ParabPdist(np.log(pexp))
                    func_acquisition2 = func_acquisition_p1.__rmul__(func_acquisition_p2) # Product of p1 and p2
                 
                # Now evaluate this acquisition function on the target probability range
                # (Do not use MFGPR_P2_log as this is not in MF-format)
                func_acq_plot2 = func_acquisition2.evaluate(P_plot_h2_log)
                
                # Debugging
                if debugplotswitch == 1:
                    plt.figure()
                    plt.plot(P_plot_h2_log,func_acq_plot2)
                    plt.title(f'MF-GPR, {count2}')
                
                # Based on results, define a new sample
                # To find the next point to evaluate we optimize the acquisition function using 
                # a standard gradient descent optimizer. It seems like this finds the coordinate 
                # of the max. value of the acq function in 1D.
                # optimizer = GradientAcquisitionOptimizer(EMU_space_log2) # SOMETIMES GOES WRONG
                # p_new_sample_log_target2, _ = optimizer.optimize(func_acquisition2) # SOMETIMES GOES WRONG
                p_new_sample_log_target2 = P_plot_h2_log[np.argmax(func_acq_plot2,axis=0)[0],0]   ### MODIFIED FOR ACQUISITION FUNCTION WORK AS VALUES BELOW 0 LED TO SELECTION OF A SAMPLE ON THE OTHER AXIS [1]
               
                # Find new point that is not equal to NaN 
                v_train_h2_new = np.nan
                while np.isnan(v_train_h2_new): 
                    
                    # Find closest sample in the LF screening MCS result
                    i_new_sample_log2,p_new_sample_log2 = find_nearest(p_samplepool2_log,p_new_sample_log_target2)
                    
                    # Selected LF points where we pick the matched HF points 
                    p_train_l2_log_REF_new = p_samplepool2_log[i_new_sample_log2] # = p_new_sample_log2
                    v_train_l2_REF_new = v_samplepool2[i_new_sample_log2]
                    
                    # Now we have to get back to the original time series peak matching to find the
                    # corresponding HF value > this is the HF sample
                    i_sel_time_new2 = np.where(peaks_LF_all == v_train_l2_REF_new)
                    v_train_h2_new = peaks_HF_all_scr[i_sel_time_new2]
                    
                    # HF probabilities are set identical to LF probabilities
                    p_train_h2_log_new = p_train_l2_log_REF_new
                    
                    # Remove newly chosen sample from the pool of samples for the next iteration
                    p_samplepool2_log = np.delete(p_samplepool2_log,i_new_sample_log2)
                    v_samplepool2 = np.delete(v_samplepool2,i_new_sample_log2) 
                                
                # Debug
                if debugplotswitch == 1:
                    print(f'Iteration {count2}: new point close to log(p) = {p_new_sample_log_target2}')
                    # print(f'Available pool: {np.expand_dims(p_samplepool2_log,1)}')
                    plt.figure(20000000)
                    plt.plot(v_train_l2_REF_new*norm_lf,p_train_l2_log_REF_new,'xr')
                    plt.text(v_train_l2_REF_new*norm_lf,p_train_l2_log_REF_new,f'{count2}') #,color='r')
                
                print(f'MF-GPR adaptive: new sample number {count2} is defined')
            
                # Add the new samples to the scatter plot
                plt.figure(6)
                plt.plot(v_train_l2_REF_new*norm_lf,v_train_h2_new*norm_hf,marker='x',color='red',markersize = 10,linestyle = 'None')
                plt.text(v_train_l2_REF_new*norm_lf+0.05,v_train_h2_new*norm_hf,f'New sample {count2+1}')
                
                # Scale factor HF to LF
                v_train_h2_new_MOD = v_train_h2_new/norm_hf*norm_lf
                
                # Add the new samples to the training data plot
                plt.figure(99)
                aX2b.plot(v_train_l2_REF_new*norm_lf,np.exp(p_train_l2_log_REF_new),marker='x',markerfacecolor='none',color='red',linestyle='None') # ,label='LF basis for HF sampling')
                aX1b.text(v_train_h2_new*norm_hf+0.3,np.exp(p_train_h2_log_new),f'it{count2+1}',color='r')
                aX2b.plot([v_train_l2_REF_new*norm_lf,v_train_h2_new_MOD*norm_hf],np.exp([p_train_l2_log_REF_new,p_train_h2_log_new]),linestyle='--',color='red')
                aX1b.plot(v_train_h2_new*norm_hf,np.exp(p_train_h2_log_new),linestyle='None',color='r',marker='d') # ,label='Resulting HF samples')               
                
                # Append new sample to existing HF samples
                p_train_h2_log = np.append(p_train_h2_log,p_train_h2_log_new)
                v_train_h2 = np.append(v_train_h2,v_train_h2_new)
                
                # Add dimension such that we have a column array
                p_train_h2_log = p_train_h2_log[:,None]
                v_train_h2 = v_train_h2[:,None]
                
                # For sample option 7, the LF data consist of a fit to the present 
                # HF data points. This means that we also need to update the LF 
                # training data for this option.                
                if sampinit == 'C':
                    
                    p_train_h2 = np.exp(p_train_h2_log)
                    
                    # Generate 3-param Weibull fit to all available HF datapoints
                    N_lf_points = 50 # number of LF points over the range indicated by topperc
                    topperc = 100 # percentage of peaks to include in fit - all of available ones
                    maxit = 100 # maximum number of iterations
                    _,_,theta,alpha,beta,corr = fit_weibull3_lsq(p_train_h2,v_train_h2,topperc,maxit)
                    vmin = v_train_h[-int(np.floor(len(v_true_l_t) * (topperc / 100)))]
                    vmax = np.max(v_train_h) 
                    v_topperc = np.linspace(vmin,vmax,N_lf_points)[:, None]
                    p_weib_LF = np.exp(-((v_topperc - theta)/alpha)**beta);
                           
                    # LF samples to use in prediction
                    v_train_l = np.squeeze(v_topperc)
                    p_train_l = np.squeeze(p_weib_LF)
                    
            # --------- Add to counter to continue the while loop, unless we reached the max. # seeds
            count2 = count2+1

        # Number of last iteration 
        lastcount2 = count2-1
        
        # Inverse the log of the final samples and results
        p_train_l2 = np.exp(p_train_l2_log)
        p_train_h2 = np.exp(p_train_h2_log)
                
        # Save final scatter plot and sample distribution plot
        if lastcount2 > 5:
            plt.figure(6)
            plt.savefig(f'{figfolder}/MFGPR_Scatter_{casename}_{txtadd}_{acqtypeMF}_Nhf{N_hf}_it{lastcount2}.pdf',bbox_inches='tight')
            
            plt.figure(99)
            plt.savefig(f'{figfolder}/MFGPR_TrainData_{casename}_{txtadd}_inclcon_{acqtypeMF}_Nhf{N_hf}_it{lastcount2}.pdf',bbox_inches='tight')

        # Plot the posterior mean and variance (U95) from the MF gaussian process  (use secondary x-axis for training data LF and HF)
        fig300 = plt.figure(300,figsize=(fsize_h,fsize_v))
        plt.yscale('log')
        plt.ylabel(r'$\mathrm{P(}\mathrm{%s} \geq \mathrm{%s} \mathrm{)}$   and   $\mathrm{P(}\mathrm{%s} \geq \mathrm{%s} \mathrm{)}$' % (varsymb_hf.upper(),varsymb_hf,varsymb_lf.upper(),varsymb_lf), fontsize=12)
        # -- x-axis1 (HF data)
        aX1c = plt.subplot()
        leg1 = aX1c.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None',label='HF "true" MCS (validation)')
        leg2 = aX1c.plot(v_train_h2*norm_hf,p_train_h2, color='r', marker='d',linestyle='None',markersize=4, label=f'{len(v_train_h2)} HF samples')
        leg3 = aX1c.plot(mean_high_mfgp_model*norm_hf,Prange_predHF, '--', color='g',label='MF-GPR pred. HF (+U95%)')
        leg4 = aX1c.fill_betweenx(Prange_predHF.flatten(), (mean_high_mfgp_model*norm_hf - 0.5*U95_high_mfgp_model*norm_hf).flatten(), 
                         (mean_high_mfgp_model*norm_hf + 0.5*U95_high_mfgp_model*norm_hf).flatten(), facecolor='g', alpha=0.1,label='MF-GPR pred. HF U95% (label not printed)')
        leg5 = aX1c.plot([0,plotmax],[pexp,pexp],':',color='black') #,label='$P_{exp}$')
        aX1c.text(locpexp,pexp*1.4,'$P_{exp}$')
        aX1c.set_xlabel(r'$\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (varsymb_hf,varunit_hf), fontsize=12,color='red')
        aX1c.set_xlim(left=0,right=plotmax)
        # -- x-axis2 (LF data)
        aX2c = aX1c.twiny()
        leg7 = aX2c.plot(v_train_l2*norm_lf,p_train_l2, color='grey', marker='o',linestyle='None',markersize=4, label='LF samples')
        leg8 = aX2c.plot(mean_low_mfgp_model*norm_lf,Prange_predLF, '--', color='b',alpha=0.4,label='MF-GPR pred. LF (+U95%)')
        leg9 = aX2c.fill_betweenx(Prange_predLF.flatten(), (mean_low_mfgp_model*norm_lf - 0.5*U95_low_mfgp_model*norm_lf).flatten(), 
                          (mean_low_mfgp_model*norm_lf + 0.5*U95_low_mfgp_model*norm_lf).flatten(), facecolor='b', alpha=0.1,label='MF-GPR pred. U95% (label not printed)')
        aX2c.set_xlabel(r'$\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (varsymb_lf,varunit_lf), fontsize=12,color=(0.4,0.4,0.4))
        aX2c.set_xlim(left=0,right=plotmax_lf)
        # -- y-axis
        plt.ylim([minP,1])
        leg = leg7+leg2+leg8+leg3+leg1
        labs = [l.get_label() for l in leg]
        aX1c.legend(leg,labs,loc='best')
        if lastcount2 > 5:
            plt.savefig(f'{figfolder}/Prediction_MFGPR_{casename}_Set-{MFkern1}-{MFkern2}-nf_{txtadd}_{acqtypeMF}_Nhf{N_hf}_it{lastcount2}.pdf',bbox_inches='tight')
            
        
        # Alternative side-by-side plot
        fig3000 = plt.figure(3001,figsize=(fsize_h*1.4,fsize_v*0.7))
        # -- subplot 2 (HF data)
        plt.subplot(1,2,2)
        plt.yscale('log') 
        plt.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None',label='HF "true" MCS (validation)')
        plt.plot(v_train_h2*norm_hf,p_train_h2, color='r', marker='d',linestyle='None',markersize=4, label=f'{len(v_train_h2)} HF samples')
        plt.plot(mean_high_mfgp_model*norm_hf,Prange_predHF, '--', color='g',label='MF-GPR pred. HF')
        plt.fill_betweenx(Prange_predHF.flatten(), (mean_high_mfgp_model*norm_hf - 0.5*U95_high_mfgp_model*norm_hf).flatten(), 
                         (mean_high_mfgp_model*norm_hf + 0.5*U95_high_mfgp_model*norm_hf).flatten(), facecolor='g', alpha=0.1,label='MF-GPR pred. HF U95%')
        plt.plot([0,plotmax],[pexp,pexp],':',color='black') #,label='$P_{exp}$')
        plt.text(locpexp,pexp*1.4,'$P_{exp}$')
        plt.xlabel(r'$\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (varsymb_hf,varunit_hf), fontsize=12)
        plt.ylabel(r'$\mathrm{P(}\mathrm{%s} \geq \mathrm{%s} \mathrm{)}$' % (varsymb_hf.upper(),varsymb_hf), fontsize=12)
        plt.xlim(left=0,right=plotmax)
        plt.ylim([minP,1])
        plt.legend(loc='best')
        # -- subplot1 (LF data)
        plt.subplot(1,2,1)
        plt.yscale('log') 
        plt.plot(v_train_l2*norm_lf,p_train_l2, color='grey', marker='o',linestyle='None',markersize=4, label='LF samples')
        plt.plot(mean_low_mfgp_model*norm_lf,Prange_predLF, '--', color='b',alpha=0.4,label='MF-GPR pred. LF')
        plt.fill_betweenx(Prange_predLF.flatten(), (mean_low_mfgp_model*norm_lf - 0.5*U95_low_mfgp_model*norm_lf).flatten(), 
                          (mean_low_mfgp_model*norm_lf + 0.5*U95_low_mfgp_model*norm_lf).flatten(), facecolor='b', alpha=0.1,label='MF-GPR pred. LF U95%')
        plt.xlabel(r'$\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (varsymb_lf,varunit_lf), fontsize=12)
        plt.ylabel(r'$\mathrm{P(}\mathrm{%s} \geq \mathrm{%s} \mathrm{)}$' % (varsymb_lf.upper(),varsymb_lf), fontsize=12)
        plt.xlim(left=0,right=plotmax_lf)
        # -- y-axis
        plt.ylim([minP,1])
        plt.legend(loc='best')
        if lastcount2 > 5:
            plt.savefig(f'{figfolder}/Prediction_MFGPR_{casename}_Set-{MFkern1}-{MFkern2}-nf_{txtadd}_{acqtypeMF}_Nhf{N_hf}_it{lastcount2}_v2.pdf',bbox_inches='tight')
            
        itrangeMF = np.arange(N_hf,N_hf+len(ass_MPM_mfGPR),1)
        print('Multi-fidelity GPR done')

    
    
    #######################
    # ------- DE-NORMALISATION
    #######################
    
    # First de-normalise, to obtain dimensional plots and output   
    mean_low_mfgp_model = mean_low_mfgp_model*norm_lf
    var_low_mfgp_model  = var_low_mfgp_model*norm_lf**2
    U95_low_mfgp_model  = U95_low_mfgp_model*norm_lf
    
    mean_high_mfgp_model = mean_high_mfgp_model*norm_hf
    var_high_mfgp_model  = var_high_mfgp_model*norm_hf**2
    U95_high_mfgp_model  = U95_high_mfgp_model*norm_hf
    
    mean_high_gp_model = mean_high_gp_model*norm_hf 
    var_high_gp_model  = var_high_gp_model*norm_hf**2
    U95_high_gp_model  = U95_high_gp_model*norm_hf
    
    ass_MPM_hfGPR     = ass_MPM_hfGPR*norm_hf 
    ass_MPM_hfGPR_u95 = ass_MPM_hfGPR_u95*norm_hf 
    ass_MPM_mfGPR     = ass_MPM_mfGPR*norm_hf 
    ass_MPM_mfGPR_u95 = ass_MPM_mfGPR_u95*norm_hf 
    
    MPM_hfGPR = MPM_hfGPR*norm_hf 
    MPM_hfGPR_U95 = MPM_hfGPR_U95*norm_hf 
    MPM_mfGPR = MPM_mfGPR*norm_hf 
    MPM_mfGPR_U95 = MPM_mfGPR_U95*norm_hf 
    
    
    #########################################
    ## ------- Overall plots
    
    if maxit_HFGPR*maxit_MFGPR != 0:
        
        
        ## -------  Visualise iterations and convergence
        
        # Visualise iterations and convergence
        seedmaxplot = maxit_HFGPR+N_hf
        
        plt.figure(figsize=(6,4))
        ax9 = plt.subplot(2,1,1)
        plt.plot(itrangeHF,ass_MPM_hfGPR,'-',color='b',label=f'Adap. HF-GPR (acq. {acqtypeHF})')
        plt.plot(itrangeMF,ass_MPM_mfGPR,'--',color='g',label=f'Adap. MF-GPR (acq. {acqtypeMF})')
        plt.plot([0,seedmaxplot],[MPM_true_MCS,MPM_true_MCS],':r',label='True HF MPM (from MCS)')
        plt.ylabel('HF MPM [m]')
        plt.xlim([N_hf-1,seedmaxplot])
        ax9.xaxis.set_tick_params(labelbottom=False)
        plt.subplot(2,1,2)
        plt.plot(itrangeHF,ass_MPM_hfGPR_u95,'-',color='b',label=f'Adap. HF-GPR (acq. {acqtypeHF})')
        plt.plot(itrangeMF,ass_MPM_mfGPR_u95,'--',color='g',label=f'Adap. MF-GPR (acq. {acqtypeMF})')
        plt.xlabel('Number of HF samples')
        plt.ylabel('U95% of HF MPM [m]')
        plt.xlim([N_hf-1,seedmaxplot])
        plt.tight_layout()
        plt.legend()
        if lastcount2 > 5:
            plt.savefig(f'{figfolder}/Convergence_HFandMFGPR_{casename}_Set-{HFkern}-nf_{txtadd}_{acqtypeMF}_Nhf{N_hf}_it{lastcount2}.pdf',bbox_inches='tight')
        
        
        
        ## ------- Plot results of HF- and MF-GPR in one plot (without samples)
        
        # ALL IN ONE Plot the posterior mean and variance (U95) from the MF gaussian process (without training data, as these are different for HF-GPR and MF-GPR)
        fig13 = plt.figure(13,figsize=(fsize_h,fsize_v))
        plt.yscale('log')
        plt.plot(v_true_h_sampled,p_true_h_sampled,marker='.',color='pink',linestyle='None',label='HF "true" MCS') 
        plt.plot(mean_high_mfgp_model,Prange_predHF, '--', color='b',label='MF-GPR prediction HF')
        plt.fill_betweenx(Prange_predHF.flatten(), (mean_high_mfgp_model - 0.5*U95_high_mfgp_model).flatten(), 
                         (mean_high_mfgp_model + 0.5*U95_high_mfgp_model).flatten(), facecolor='b', alpha=0.1,label='MF-GPR prediction HF U95%')
        plt.plot(mean_high_gp_model,Prange_predHF,'-.',color='g',label='GPR prediction HF')
        plt.fill_betweenx(Prange_predHF.flatten(), (mean_high_gp_model - 0.5*U95_high_gp_model).flatten(), 
                         (mean_high_gp_model + 0.5*U95_high_gp_model).flatten(), facecolor='g', alpha=0.1,label='GPR prediction HF U95%')
        plt.plot([0,plotmax],[pexp,pexp],':',color='black') #,label='$P_{exp}$')
        plt.text(locpexp,pexp*1.4,'$P_{exp}$')
        plt.xlabel(r'$\mathrm{%s}$ $\mathrm{[} \mathrm{%s} \mathrm{]}$' % (varsymb_hf,varunit_hf), fontsize=12)
        plt.ylabel(r'$\mathrm{P(}\mathrm{%s} \geq \mathrm{%s} \mathrm{)}$' % (varsymb_hf.upper(),varsymb_hf), fontsize=12)
        plt.ylim([minP,1])
        plt.xlim([0,plotmax])
        plt.legend(loc='best')
        plt.savefig(f'{figfolder}/Pred_HFandMFGPR_{casename}_{txtadd}_HFacq-{acqtypeHF}_MFacq-{acqtypeMF}_Nhf{N_hf}_HFit{lastcount1}_MFit{lastcount2}.pdf',bbox_inches='tight')
        
                 
        # --------------- Calculate MPM values from DNR
        print(f'Final results after {lastcount1} HF-GPR iterations and {lastcount2} MF-GPR iterations:')
        print(f'MPM TrueMCS {dur/3600}h: {MPM_true_MCS} m')
        print(f'MPM adaptive MF-GPR {dur/3600}h (it.{lastcount2}): {MPM_mfGPR} m with U95 {MPM_mfGPR_U95} m')
        print(f'MPM adaptive HF-GPR {dur/3600}h (it.{lastcount1}): {MPM_hfGPR} m with U95 {MPM_hfGPR_U95} m')
        
        # Write all MPM values to a CSV file
        CSV_fn_MPM = f'{filefolder}/MPMandU95_{casename}_{txtadd}_adapMFGPRset{MFkern1},{MFkern2},{acqtypeMF},it{lastcount2}_adapHFGPRset{HFkern},{acqtypeHF},it{lastcount1}.csv'
        head_MPM = ['True MCS DNR','GPR mean','GPR U95','MF-GPR mean','MF-GPR U95']
        CSV_data_MPM = np.array([[MPM_true_MCS,MPM_hfGPR,MPM_hfGPR_U95,MPM_mfGPR,MPM_mfGPR_U95]])
        CSV_df_MPM = pd.DataFrame(CSV_data_MPM)
        CSV_df_MPM.to_csv(CSV_fn_MPM,header = head_MPM,index= False)
        
        print('MPM calc done')
        
    
        print('Everything done')
    
    
    return itrangeHF,ass_MPM_hfGPR,ass_MPM_hfGPR_u95,itrangeMF,ass_MPM_mfGPR,ass_MPM_mfGPR_u95,mean_high_gp_model,var_high_gp_model,mean_low_mfgp_model,var_low_mfgp_model,mean_high_mfgp_model, var_high_mfgp_model



##########################################   
def present_hyperparams(gpr_model):
    """
    Access the names and values of the present hyperparameters of a GPR object
    gpr_model (made with toolbox GPy) before or after optimization, and print 
    their names and values.
    """
    # Assuming high_gp_model is your GPR model
    values = gpr_model.param_array
    
    # Get the names of the hyperparameters
    names = gpr_model.parameter_names()
    
    # # Print names and values of the optimized hyperparameters
    # for name1, value1 in zip(names, values):
    #     print(f"{name1}: {value1}")
        
    return names,values   
    

##########################################
def find_nearest(array, value):
       
    array = np.asarray(array)
    # idx = (np.abs(array - value)).argmin()
    temp = (np.abs(array - value))
    i_nearest = np.nanargmin(temp)
    
    nearest = array[i_nearest]
        
    return i_nearest,nearest


##########################################
def bin_means_smoothing(x,y,Xbins):
    # Initialize an array to store smoothed values
    smoothed_y = np.zeros(len(Xbins)-1)
    Xbins_middle = np.zeros(len(Xbins)-1)
    
    # Iterate over bins and calculate the mean value of y within each bin
    for i in range(len(Xbins) - 1):
        mask = np.logical_and(x >= Xbins[i], x < Xbins[i+1])
        mean_y = np.nanmean(y[mask])
        smoothed_y[i] = mean_y
        Xbins_middle[i] = (Xbins[i] + Xbins[i+1])/2
    
    return Xbins_middle,smoothed_y
