# -*- coding: utf-8 -*-
"""
Created on Fri Dec  2 15:07:13 2022

@author: smvanessen

This library contains some scripts that are dedicated to the screening procedure
as described in the Ocean Eng. screening paper based on CRS GW.

"""

# Import toolboxes
import numpy as np
import matplotlib.pyplot as plt
import scipy

# # Import zero up- and down-crossing script. This is not the proper way to refer.. update later.
# from helpfunctions.distribution_scripts import calc_zup
# from helpfunctions.wavefunctions import calc_wavenumber



# =============================================================================


def calc_risetime_steepness_impulse_height_ind_cfs_IAP_SCREAM(time1,signal1,thr_cr,plotswitch,discontcheck,wd,Vs):
    """
    ----------
    Input:
    ----------
    - time1      = time axis [s]
    - signal1    = time trace to calculate rise times & steepnesses for peaks [<unit>]
    - thr_cr     = threshold for peaks over threshold analysis - crests
                   (default = no POT analysis, thr_cr = 0) [<unit>]
    - plotswitch = plot on (1) or not (0)
    - discontcheck = discontinuity check (1) or not (0). In case of 0, there
                   is still a check based on NaN values.
    ----------
    Output:
    ----------
    - peaks_a           = values crests
    - i_peaks_tot_a     = indices crests (w.r.t. total input time)
    - troughs_a         = values troughs
    - i_troughs_tot_a   = indices troughs (w.r.t. total input time)
    - rt_a              = rise time of crest from 10# of maximum to maximum for
                          each identified peak [s]
    - rst_a             = rise steepness of crest height diff / rt from 10# of max
                          to max for each identified peak, so actually velocity [<unit>/s]
    - impulse_a         = area under crest between 10#max before to 10#max after
                          for each identified peak [<unit>^2*s]
    - heights_a         = heights of crest + depth of preceding trough
    - cfs_a             = crest front steepness (Myrhaug and Kjeldsen, 1986)
    - Tz_up_a           = zero up-crossing periods
    - Tz_down_a         = zero down-crossing periods
    - timez_up_a        = times of zero up-crossings [s]
    - IAP               = impact alert parameter of Stansberg (2008) [(m/s)**2]
    ----------
    
    Note 1: with discontcheck = 1, the result are given per 'section' of the 
    data, where the section changes are automatically detected at 
    discontinuities in the input signal. The idea is that the input signal is 
    concatenated from multiple runs, and you don't want to calculate rise 
    time etc. around a discontinuity.
    
    Note 2: rt, rst and impulse can have NaN values if a crest is close to a
    discontinuity in the signal (due to a run change for instance).
    
    Note 3: rise time, steepness and impulse are only defined for the crest
    in this function, from 10# of the crest value. The function needs to be
    rewritten if these variables are required for the height.

    ----------
    v1.0  SvE (Oct. 2018)
    v1.1  SvE (Nov. 2019) - added crest front steepness
    v2    SvE (Nov. 2022) - matlab > python
    """

    
    # Figure sizes (units unclear, see matplotlib doc)
    fsize_h = 7
    fsize_v = 5
    
    # Check whether no. of time indices corresponds to no. of signal entries
    if len(time1) != len(signal1):
        # error('Error: number of entries in <time> is not equal to number of entries in <signal>') 
        raise ValueError('Error: number of entries in <time1> is not equal to number of entries in <signal1>')

    # For concatenated runs: identify jumps between runs. This is necessary to
    # avoid that this influence the rise time and steepness results
    ind_disc2 = np.where(np.isnan(signal1)) 
    if discontcheck == 1:
        diffsig = np.diff(signal1) 
        diffsig2 = np.diff(diffsig) 
        ind_disc1 = np.where(abs(diffsig2) > 20*np.nanstd(diffsig2))
        ind_disc = np.union1d(ind_disc1,ind_disc2) 
    else:
        ind_disc = ind_disc2.copy() # Avoid binding between copy and original 
    
    # Run if there are discontinuities to be checked    
    if len(ind_disc) != 0:
        
        # Group discontinuities indices (one group if there are no indices in between)
        difference = np.diff(ind_disc)
        ii_diff = np.argwhere(difference != 1) 
        ind_disc_st = np.squeeze(np.concatenate(([[ind_disc[0]]],ind_disc[ii_diff+1]))+1)
        ind_disc_en = np.squeeze(np.concatenate((ind_disc[ii_diff],[[ind_disc[-1]]]))+1)
        
        if ind_disc_en[0] < min(np.argwhere(~np.isnan(signal1))):
            # If first non-nan value is later than the end of the first group, remove indices first group
            ind_disc_st = np.delete(ind_disc_st,0)
            ind_disc_en = np.delete(ind_disc_en,0)
        if ind_disc_en[-1] > max(np.argwhere(~np.isnan(signal1))):
            # If last non-nan value is earlier than the end of the last group, remove indices last group
            ind_disc_st = np.delete(ind_disc_st,-1)
            ind_disc_en = np.delete(ind_disc_en,-1)
        
        # Turn this into the start and end indices of the time trace groups
        if ind_disc_st[0] != 0:
            ind_ttpart_st = np.concatenate(([0],ind_disc_en))
        else:
            ind_ttpart_st = ind_disc_en 
            
        if ind_disc_en[-1] != len(signal1):
            ind_ttpart_en = np.concatenate((ind_disc_st,[len(signal1)-1]))
        else:
            ind_ttpart_en = ind_disc_st 
        
        # Plot to check
        if plotswitch == 1:
            figX = plt.figure(figsize=(fsize_h,fsize_v))
            axX = figX.add_subplot(2,1,1)
            plt.plot(time1,signal1,'--', color='b')
            plt.plot(time1[ind_disc_st],np.zeros(np.shape(ind_disc_st)),'or',label='discontinuity start',markerfacecolor='None') 
            plt.plot(time1[ind_disc_en],np.zeros(np.shape(ind_disc_en)),'ok',label='discontinuity end',markerfacecolor='None') 
            plt.plot(time1[ind_ttpart_st],np.zeros(np.shape(ind_ttpart_st)),'ob',label='time trace group start',markerfacecolor='None') 
            plt.plot(time1[ind_ttpart_en],np.zeros(np.shape(ind_ttpart_en)),'og',label='time trace group end',markerfacecolor='None') 
            axX.legend()
            plt.ylabel('signal')  
            plt.tight_layout()
            
            if discontcheck == 1:
                ax2 = figX.add_subplot(2,1,2)
                plt.plot(time1[0:-1],abs(diffsig),'-', color='b',label='diff.')
                plt.plot(time1[1:-1],abs(diffsig2),'-',color='r',label='2nd diff.')
                plt.xlabel('time [s]')  
                plt.ylabel('diff. signal')  
                plt.legend() 
                plt.tight_layout()
            
    else:
        # Start and end indices
        ind_ttpart_st = 0 
        ind_ttpart_en = len(signal1) 
     
    # Initialise the lists that assemble the data over all time series sections
    # List used here as they are easy to append to?
    peaks_a = []  
    i_peaks_tot_a = [] 
    troughs_a = []  
    i_troughs_tot_a = [] 
    rt_a = []  
    rst_a = [] 
    impulse_a = [] 
    heights_a = [] 
    cfs_a = [] 
    Tz_up_a = [] 
    Tz_down_a = [] 
    timez_up_a = [] 
    timez_down_a = [] 
    IAP_a = []
    
    # Number of sections
    Nsect = len(ind_ttpart_st) 
    
    #  Initialise the arrays that assemble the total indices per section
    isave = np.zeros(Nsect,)
    
    # -------------- Cut time traces in different sections between the discontinuities
    for qq in np.arange(Nsect):
    
        print('Section {} of {}'.format(qq+1,Nsect))
        
        # Start and end index of this group        
        ii_start = ind_ttpart_st[qq] 
        ii_end = ind_ttpart_en[qq] 
        
        # Group time traces
        time_t = time1[ii_start:ii_end]
        signal_t = signal1[ii_start:ii_end]
        
        # Calculate zero up-crossing (peaks) and zero down-crossings
        t_peaks,peaks,t_troughs,troughs,timez_up,timez_down = calc_zup(time_t,signal_t,0)

        # Find indices corresponding to troughs and peaks in original time series
        i_peaks = np.nan*np.ones(np.shape(t_peaks),dtype=int)
        i_troughs = np.nan*np.ones(np.shape(t_troughs),dtype=int)
        for bb,_ in enumerate(t_peaks):
            if ~np.isnan(t_peaks[bb]):
                i_peaks[bb] = np.argmax(time_t >= t_peaks[bb])
        del bb
        for bb,_ in enumerate(t_troughs):
            if ~np.isnan(t_troughs[bb]):
                i_troughs[bb] = np.argmax(time_t >= t_troughs[bb])
        del bb
        
        # Calculate impact alert parameter for this section
        plotswitch = 0
        IAP_all = impact_alert_param(time_t,signal_t,wd,Vs,plotswitch)
        IAP_all = np.append(IAP_all,np.nan)      
        
        # Set last 10 and first 10 indices of impact alert parameter to NaN (as 
        # it gives unwanted peaks near the end or start of the sections)
        IAP_all[0:10] = np.nan
        IAP_all[-10:-1] = np.nan
        
        # plt.figure()
        # plt.subplot(2,1,1)
        # plt.plot(time_t,signal_t,'b',label='W23 [m]')
        # plt.legend()
        # plt.subplot(2,1,2)
        # plt.plot(time_t,IAP_all,'r',label='IAP [(m/s)**2]')
                
        # If there is no zero up-crossing, or no trough or peak in this group, discard the group
        if len(timez_up) == 0 or len(t_troughs) == 0 or len(t_peaks) == 0:
            print('This section (qq={}) of the time trace does not contain zero up-crossings, continue with next section.'.format(qq)) 
        else:           

            # Order: zero downcr, trough, zero up-cr, peak for all peaks
            # This is ensured by the calc_zup script for the peaks & troughs, but not for the down-crossings
            # Check first zero down-crossing (should be after the first up-crossing)
            if timez_down[0] < timez_up[0]:
                timez_down = np.delete(timez_down,0)
            # Check last zero down-crossing (should be after the last up-crossing)
            if timez_down[-1] < timez_up[-1]:
                timez_down = np.append(timez_down,np.nan)
            # Check equal lengths
            if len(timez_down) != len(timez_up):
                raise ValueError('The numbers of zero up- and down-crossings after processing are not equal in section (qq={})'.format(qq))
                
            # Display number of peaks
            n_peaks = len(peaks) 
            n_troughs = len(troughs) 
            print('# crests zero up-crossing = {}'.format(n_peaks)) 
            print('# troughs zero up-crossing = {}'.format(n_troughs)) 
            
                        
            ## ---- Optionally: use peak over threshold analysis for crests
            
            # Threshold has to exist now otherwise there is no check of no. of troughs
            if thr_cr != 0:
            
                # Find number and indices of peaks over threshold
                peaks_thr          = peaks[peaks > thr_cr]
                i_peaks_thr        = i_peaks[peaks > thr_cr]
                 
                # Similar for troughs, but check first whether first trough is before
                # first crest (4 conditions)
                if i_peaks[0] > i_troughs[0] and i_peaks[-1] > i_troughs[-1]:
                    
                    cond = 1 
                    
                    troughs_thr    = troughs[peaks > thr_cr]
                    i_troughs_thr  = i_troughs[peaks > thr_cr] 
                                       
                elif i_peaks[0] < i_troughs[0] and i_peaks[-1] > i_troughs[-1]:
                    
                    cond = 2 
                    
                    peaks_t = peaks[1:-1] 
                    
                    troughs_thr    = troughs[peaks_t > thr_cr]
                    i_troughs_thr  = i_troughs[peaks_t > thr_cr] 
                    
                    if peaks_thr[0] == peaks[0]:
                        troughs_thr    = [np.NaN,troughs_thr] 
                        i_troughs_thr  = [0,i_troughs_thr] 
                                           
                elif i_peaks[0] > i_troughs[0] and i_peaks[-1] < i_troughs[-1]:
                    
                    cond = 3 
                     
                    peaks_t     = peaks 
                    troughs_t   = troughs[0:-2] 
                    i_troughs_t = i_troughs[0:-2] 
                    
                    troughs_thr    = troughs_t[peaks_t > thr_cr]
                    i_troughs_thr  = i_troughs_t[peaks_t > thr_cr] 
                   
                elif i_peaks[0] < i_troughs[0] and i_peaks[-1] < i_troughs[-1]:
                    
                    cond = 4 
                    
                    peaks_t = peaks[1:-1] 
                    troughs_t   = troughs[0:-2] 
                    i_troughs_t = i_troughs[0:-2] 
                    
                    troughs_thr    = troughs_t[peaks_t > thr_cr]
                    i_troughs_thr  = i_troughs_t[peaks_t > thr_cr] 
                    
                    if peaks_thr[0] == peaks[0]:
                        troughs_thr    = [np.NaN,troughs_thr] 
                        i_troughs_thr  = [0,i_troughs_thr] 
                       
                
                # Display number of peaks above threshold
                n_pot = len(peaks_thr) 
                print('# peaks above threshold = {}'.format(n_pot)) 
                
                # Overwrite peaks and indices of peaks with peaks above threshold values
                peaks = peaks_thr 
                troughs = troughs_thr 
                i_peaks = i_peaks_thr 
                i_troughs = i_troughs_thr            
            
            # Calculate zero up- and down-crossing times (preparation for crest front steepness CFS) 
            Tz_up = timez_up[1:] - timez_up[0:-1] 
            Tz_down = timez_down[1:] - timez_down[0:-1]
            Tz_up = np.append(Tz_up,np.NaN)
            Tz_down = np.append(Tz_down,np.NaN)
            
            
            ## ----- Crests analysis
            # Calc 10# of maximum for each maximum
            peaks_10p = 0.1*peaks 
            
            # Initialise
            i_start1    = np.nan*np.ones(np.shape(peaks)) 
            i_end1      = np.nan*np.ones(np.shape(peaks)) 
            impulse1    = np.nan*np.ones(np.shape(peaks)) 
            heights1    = np.nan*np.ones(np.shape(peaks)) 
            usable      = np.ones(np.shape(peaks)) 
            rt1         = np.ones(np.shape(peaks)) 
            rst1        = np.ones(np.shape(peaks)) 
            wast1       = np.ones(np.shape(peaks)) 
            IAP1        = np.ones(np.shape(peaks)) 
            
            
            # Loop over number of crests
            for ii,_ in enumerate(peaks):
                
                if np.isnan(peaks[ii]):
                    
                    usable[ii] = np.nan
                    impulse1[ii] = np.nan
                    heights1[ii] = np.nan 
                    IAP1[ii] = np.nan
                    
                else:
                    # Find indices (i_start1 = last time 10# value occurs before each peak, i_end1 = first time 10# value occurs after each peak)
                    _,i_start1[ii] = np.nonzero(signal_t[0:np.int(i_peaks[ii])] <= peaks_10p[ii])[0][[0,-1]]
                    i_end1_t,_ = np.nonzero(signal_t[np.int(i_peaks[ii]):-1] <= peaks_10p[ii])[0][[0,-1]]
                    i_end1[ii] = i_end1_t + i_peaks[ii]-1 
                
                    # Check whether there is a discontinuity between start & peak index or between peak & end index
                    temp1 = np.ones(np.shape(ind_disc)) 
                    temp2 = np.ones(np.shape(ind_disc))
                    # for vv = 1:len(ind_disc):
                    for vv,_ in enumerate(ind_disc):    
                        if i_start1[ii] < ind_disc[vv] and i_peaks[ii] > ind_disc[vv]:
                            temp1[vv] = np.nan
                        if i_peaks[ii] < ind_disc[vv] and i_end1[ii] > ind_disc[vv]:
                            temp2[vv] = np.nan
                    if np.any(np.isnan(temp1)) or np.any(np.isnan(temp2)):
                        usable[ii] = np.nan
                 
                    # Integrate impulse over this time interval
                    i_st2 = np.int(i_start1[ii])
                    i_en2 = np.int(i_end1[ii])
                    tdiff = time_t[i_st2+1:i_en2] - time_t[i_st2:i_en2-1] 
                    impulse1[ii] = sum(signal_t[i_st2+1:i_en2]*tdiff) 
                
                    # Identify preceding trough for each crest, calculate heights
                    if i_peaks[0] > i_troughs[0]:
                        heights1[ii] = peaks[ii] - troughs[ii] 
                    else:
                        if ii == 1:
                            heights1[ii] = np.nan 
                        else:
                            heights1[ii] = peaks[ii] - troughs[ii] 
                            
                    # Identify maximum impact alert parameter value between zero up-crossings
                    if ii < len(peaks):
                        iz_up_1 = np.argmax(time_t>=timez_up[ii])
                        iz_up_2 = np.argmax(time_t>=timez_up[ii+1])
                        IAP1[ii] = np.nanmax(IAP_all[iz_up_1:iz_up_2])
                    else:
                        IAP1[ii] = np.nan
            
            # nonan = false for peaks equal to nan and true for peaks not equal to nan
            nonan = ~np.isnan(peaks)
            
            # Rise times
            rt1[nonan] = time_t[i_peaks[nonan].astype(int)] - time_t[i_start1[nonan].astype(int)]
            
            # Rise steepnesses
            diffV = signal_t[i_peaks[nonan].astype(int)] - signal_t[i_start1[nonan].astype(int)]
            rst1[nonan] = diffV/rt1[nonan]
            wast1[nonan] = heights1[nonan]/(Tz_up[nonan]**2) 
             
            # Crest front steepness (Myrhaug & Kjeldsen)
            cfs1 = peaks/((9.81/(2*np.pi))*Tz_down*rt1) 
             
            # Set values to NaN at run changes (or other sudden jumps)
            rt = rt1*usable 
            rst = rst1*usable 
            wast = wast1*usable 
            # wast = 0  # temp fix as long as no of heights entries is not the same as no of peaks.
            cfs = cfs1*usable 
            impulse = impulse1*usable 
            heights = heights1*usable 
            i_start = i_start1*usable 
            i_end = i_end1*usable 
            IAP = IAP1*usable 
            
            # REMOVE? Not sure what this does.
            # i_start_plot = i_start.copy
            # i_start_plot(np.isnan(i_start_plot)) = [] 
            # i_end_plot = i_end.copy
            # i_end_plot(np.isnan(i_end_plot)) = [] 
             
            # Number of crests and number of usable crests
            nind = len(i_peaks) 
            nind_se = nind - sum(np.isnan(usable)) 
             
            # Plot with check of numbers for the first nind indices (only for the first time series section)
            if plotswitch == 1 and qq == 0:
    
                # Labels
                indt_p = ['{}'.format(i) for i,_ in enumerate(peaks)]
                indt_t = ['{}'.format(i) for i,_ in enumerate(troughs)]
                indt_zup = ['{}'.format(i) for i,_ in enumerate(timez_up)]
                indt_zdown = ['{}'.format(i) for i,_ in enumerate(timez_down)]
                                    
                # Make plot
                figY = plt.figure(figsize=(fsize_h,fsize_v))
                axY = figY.add_subplot(2,1,1)
                plt.plot(time_t,signal_t,'-', color='b',label='signal')
                plt.plot(time_t[i_peaks.astype(int)],signal_t[i_peaks.astype(int)],'or',label='peaks') 
                plt.plot(time_t[i_troughs.astype(int)],signal_t[i_troughs.astype(int)],'*g',label='troughs') 
                plt.plot(timez_up,np.zeros(np.shape(timez_up)),'+m',label='zero up-crossings') 
                plt.plot(timez_down,np.zeros(np.shape(timez_down)),'dc',label='zero down-crossings')  
                plt.tight_layout()
                for label, x, y in zip(indt_p, time_t[i_peaks.astype(int)],signal_t[i_peaks.astype(int)]):
                    plt.annotate('{}'.format(label),xy=(x, y), xytext=(-5, 10),color='r',
                             textcoords='offset points', ha='center', va='bottom',
                             arrowprops=dict(arrowstyle='->', connectionstyle='arc3,rad=0'))
                del label,x,y
                for label, x, y in zip(indt_t,time_t[i_troughs.astype(int)],signal_t[i_troughs.astype(int)]):
                    plt.annotate('{}'.format(label),xy=(x, y), xytext=(-5, 10),
                             textcoords='offset points', ha='center', va='bottom',color='g',
                             arrowprops=dict(arrowstyle='->', connectionstyle='arc3,rad=0'))
                del label,x,y
                for label, x, y in zip(indt_zup,timez_up,np.zeros(np.shape(timez_up))):
                    plt.annotate('{}'.format(label),xy=(x, y), xytext=(-5, 10),
                             textcoords='offset points', ha='center', va='bottom',color='m',
                             arrowprops=dict(arrowstyle='->', connectionstyle='arc3,rad=0'))
                del label,x,y
                for label, x, y in zip(indt_zdown,timez_down,np.zeros(np.shape(timez_down))):
                    plt.annotate('{}'.format(label),xy=(x, y), xytext=(-5, 10),
                             textcoords='offset points', ha='center', va='bottom',color='c',
                             arrowprops=dict(arrowstyle='->', connectionstyle='arc3,rad=0'))
                del label,x,y
                axY.legend()
                plt.ylabel('signal')  
                plt.xlabel('time [s]')  
                plt.title('Time series section {} of {}'.format(qq+1,Nsect))
            
            # Also define indices of peaks and troughs with respect to the total 
            # time trace (here only per section qq)
            if qq == 0:
                isave[qq] = len(time_t)
                
                i_peaks_tot = i_peaks
                i_troughs_tot = i_troughs

            elif qq > 0:
                isave[qq] = len(time_t) + isave[qq-1] + 1
                
                i_peaks_tot = i_peaks + isave[qq-1]
                i_troughs_tot = i_troughs + isave[qq-1]


            # Assemble data over all time series sections (qq)
            peaks_a = np.append(peaks_a, peaks)
            i_peaks_tot_a = np.append(i_peaks_tot_a, i_peaks_tot)
            troughs_a = np.append(troughs_a, troughs)
            i_troughs_tot_a = np.append(i_troughs_tot_a, i_troughs_tot)
            rt_a = np.append(rt_a, rt)
            rst_a = np.append(rst_a, rst)
            impulse_a = np.append(impulse_a, impulse)
            heights_a = np.append(heights_a, heights) 
            cfs_a = np.append(cfs_a, cfs)
            Tz_up_a = np.append(Tz_up_a, Tz_up) 
            Tz_down_a = np.append(Tz_down_a, Tz_down) 
            timez_up_a = np.append(timez_up_a, timez_up) 
            timez_down_a = np.append(timez_down_a, timez_down)
            IAP_a = np.append(IAP_a, IAP)
            
            
            ## DEBUG IAP
            t_peaks_tot_a = np.zeros(np.shape(peaks_a))
            for bb,_ in enumerate(peaks_a):
                if np.isnan(i_peaks_tot_a[bb]):
                    t_peaks_tot_a[bb] = np.nan    
                else:
                    t_peaks_tot_a[bb] = time1[int(i_peaks_tot_a[bb])]
            
            plt.figure()
            plt.plot(time1,signal1,'b',label='W23')
            plt.plot(t_peaks_tot_a,IAP_a,'r',label='impact alert parameter')
   
    
    return peaks_a,i_peaks_tot_a,troughs_a,i_troughs_tot_a,rt_a,rst_a,impulse_a,heights_a,cfs_a,Tz_up_a,Tz_down_a,timez_up_a,timez_down_a,IAP_a


# =============================================================================


def match_indicator_to_load(t_peaks_I,peaks_I,t_zc_I,time_L,sig_L,tbefore):
    """  
   	This function correlates maxima in a Londary signal L to the peaks in
   	a lead signal LEAD. The zero up-crossings in the lead signal are used to
   	correlate the two  the maximum peak in L within the preceding and next
   	zero up-crossing in LEAD is identified as the matching peak. This does
   	not have to be a peak in the L signal.
   	
     SCREAM version: the original function was made for CRS green water. In
   	this project, the full time traces of the failure mode forces (Londary 
   	signal) was available. In CRS SCREAM however, only the peaks are 
   	available. Hence the present modified script. Note that the indicator
   	signal (primary signal) is still assumed to be continuous.
   	
   	Note that negative peaks in the L signal require a minus sign in the
   	input!
   	
   	----------
    Input:
    ----------
   	t_peaks_I       = times peaks in indicator signal [s] <numpy array>
   	peaks_I         = value zero up-crossing peaks in indicator signal [<x>]
                      the number of peaks needs to be identical to the number of 
                      zero up-crossings (definition: first zero up-crossing 1, 
                      then zero up-crossing peak 1, then trough 1 and zero 
                      down-crossing 1, after that zero up-crossing 2 etc. )
                      <numpy array>
   	t_zc_I          = times zero up-crossings in indicator signal [s]  <numpy array>
   	time_L          = time axis load signal [s]  <numpy array>
   	sig_L           = load signal [<y>]  <numpy array>
   	tbefore         = Shift of zero up-crossing interval option. In principle 
   	                     this shift is zero. Shift interval by <tbefore> seconds 
   	                     forward to also capture hammerfist events for instance 
   	                     in green water problem. Doubt? Use 0. [s]  <single float>
    	----------
    Output:
    ----------
    	peaks_I         = Peaks lead signal
    	t_peaks_I       = Time of peaks lead signal
    	max_L_corr      = Load maxima corresponding to zero up-crossing peaks in indicator [<y>]
    	t_max_L_corr    = Times of load maxima corresponding to zero up-crossing peaks in indicator [s]
    	i_max_L_corr    = Indices of load maxima
    	t_diff          = Time difference between resulting indicator peaks and load maxima [s]
        
    	----------
    Revision history:
    ----------
   	- Sep 2020 SvE original script
   	- Jan 2023 SvE Matlab > Python
 
   	Uses scripts from analysis toolbox and matlab library SvE
    """
    
    # QA: Check that the number of indicator up-crossings matches the number of indicator peaks
    if len(peaks_I) != len(t_zc_I):
        raise ValueError('The number of indicator up-crossings does not match the number of indicator peaks.')

    # QA: Check that the first peak is after the first zero up-crossing
    if t_zc_I[0] > t_peaks_I[0]:
        raise ValueError('The first indicator peak is not after the first indicator zero up-crossing.')

    # Number of peaks
    Np_I = len(peaks_I)
    
    # Initialise the load peaks corresponding to identified indicator zero 
    # up-crossing peaks, and their time difference
    max_L_corr = np.zeros((Np_I))
    i_max_L_corr = np.zeros((Np_I))
    t_max_L_corr = np.zeros((Np_I))
    t_diff = np.zeros((Np_I))
    
    # Last entry is NaN as there is no next indicator peak (or zero up-crossing)
    if t_peaks_I[-1] > t_zc_I[-1]:
        max_L_corr[-1] = np.NaN 
        i_max_L_corr[-1] = np.NaN 
        t_max_L_corr[-1] = np.NaN 
        t_diff[-1] = np.NaN 

    # Loop over indicator peaks
    for ii,_ in enumerate(np.arange(Np_I-1)):
        
        # Find largest load value between each consecutive set of indicator peaks
        # Interval was shifted <tbefore> forward.
        interval_ii = np.nonzero((time_L >= (t_zc_I[ii]-tbefore)) & (time_L <= (t_zc_I[ii+1]-tbefore)))
        
        if np.size(interval_ii) != 0:
            # Load maximum in the interval and its index within the interval
            # Assume that load peaks are positive, as well as indicator peaks.
            i_temp = np.argmax(sig_L[interval_ii])
            max_L_corr[ii] = sig_L[interval_ii]
            
            # Find total index (from first of interval plus index within interval)
            i_max_L_corr[ii] = i_temp + interval_ii[0]
            
            # Find corresponding time in load signal
            t_max_L_corr[ii] = time_L[i_max_L_corr[ii].astype(int)]
            
            # Check the time difference between the LEAD peak and the corresponding L maximum
            t_diff[ii] =  t_max_L_corr[ii] - t_peaks_I[ii] 
        else:
            max_L_corr[ii] = np.NaN 
            i_max_L_corr[ii] = np.NaN 
            t_max_L_corr[ii] = np.NaN 
            t_diff[ii] = np.NaN 


    # # Debug figure
    # figZ = plt.figure() 
    # axZa = figZ.add_subplot(2,1,1)
    # plt.tight_layout()
    # plt.plot(t_peaks_I,peaks_I ,'xr')
    # axZb = figZ.add_subplot(2,1,2)
    # plt.tight_layout()
    # plt.plot(time_L,sig_L,'b') 
    # plt.plot(t_max_L_corr,max_L_corr,'xr') 

    return max_L_corr,t_max_L_corr,i_max_L_corr,t_diff


# =============================================================================

  
def match_peakedindicator_to_load(time_I,sig_I,time_L,sig_L,thr_I,tint,psw):
    """  
    This function correlates maxima in a indicator signal to the maxima in
    a load signal. This is a two-way coupling model for peaked signal
    (not useful for 'continuous' signals). Peaks in both the indicator and
    the load signal are identified. They are grouped using the time
    interval 'tint' in [s] (only one peak may be present in the interval).
    The maximum value around each of these grouped peaks in the indicator and
    the load are recorded. 
    
   	----------
    Input:
    ----------
    time_I    	= times signal indicator signal [s] (may be continuous or 
                  peaked, see sig_I)
    sig_I       	= indicator signal [<x>] (may be continuous or peaked, but when 
                  it is peaked you need to supply a suitable thr_I value such 
                  that zero up-crossings can be found for sig_I-thr_I)
    time_L    	= times of load signal [s] (only peaks assumed)
    sig_L     	= load signal [<y>] (only peaks assumed)
    thr_I       	= threshold for peak over threshold analysis (shift of
	              indicator signal if there are no zero crossings)
    tint        = time interval in which there may be only one peak [s]
    psw        	= plot switch (1 = plots, 0 = no plots)
   	
    ----------
    Output:
    ----------
    sig_pI      	= peaks indicator signal
    time_pI   	= time of peaks indicator signal
    ind_pI    	= indices of peaks indicator signal
    sig_pL     	= peaks failure mode signal
    time_pL    	= time of peaks failure mode signal
    ind_pL     	= indices of peaks failure mode signal
    t_diff      	= their time difference
   
    ----------
    Revisions:
    ----------
    - Sep 2020 SvE original script
    - Dec 2020 SvE update including two-way peak identification
    - Jan 2023 SvE matlab > python

    """

    # Find peaks & times of peaks in indicator
    time_pI_temp,sig_pI_temp,_,_,_,_ = calc_zup(time_I,sig_I-thr_I,0)
    sig_pI_temp = sig_pI_temp + thr_I
    
    # Peaks & times of peaks in load are already in the input
    # time_pL_temp,sig_pL_temp,_,_,_,_ = calc_zup(time_L,sig_L-thr_L,0) # Alternative version for continuous signals, requires input thr_L
    # sig_pL_temp = sig_pL_temp + thr_L  # Alternative version for continuous signals, requires input thr_L
    time_pL_temp = time_L.copy()
    
    # Join and sort these times
    AllPeakTimes = np.sort(np.concatenate((time_pI_temp,time_pL_temp)))
    
    # Reduce the array, taking into account the time interval in which there
    # may be only one peak. The first index of each 'group' is maintained (so
    # the interval in which the maxima are found later has to be identified
    # upward from there.
    diffarray = AllPeakTimes[1:-1] - AllPeakTimes[0:-2]
    i_delete = np.squeeze(np.array(np.where(diffarray < tint)),axis=0) + 1
    UniquePeakTimes = AllPeakTimes.copy()
    UniquePeakTimes = np.delete(UniquePeakTimes,i_delete)
    
    plt.figure(10000) 
    plt.plot(time_pI_temp,np.ones(np.shape(time_pI_temp)),'dc',label='indicator peak times')
    plt.plot(time_pL_temp,np.ones(np.shape(time_pL_temp)),'om',label='load peak times')
    plt.plot(UniquePeakTimes,np.ones(np.shape(UniquePeakTimes)),'+k',label='joint peak times with {} s interval'.format(tint))
    plt.legend()
    
    # Initialise
    sig_pI = np.zeros(np.shape(UniquePeakTimes))
    ind_pI = np.zeros(np.shape(UniquePeakTimes))
    time_pI = np.zeros(np.shape(UniquePeakTimes))
    sig_pL = np.zeros(np.shape(UniquePeakTimes))
    ind_pL = np.zeros(np.shape(UniquePeakTimes))
    time_pL = np.zeros(np.shape(UniquePeakTimes))
    t_diff = np.zeros(np.shape(UniquePeakTimes))
    
    # Find max value indicator and L in interval around each 'peak time'
    for nn,_ in enumerate(UniquePeakTimes):
        
        # Time interval for indicator and load (lowest index in UniquePeakTimes, time interval tint [s])
        t_interval_I = np.squeeze(np.array(np.where((time_I >= UniquePeakTimes[nn]) & (time_I <= (UniquePeakTimes[nn]+tint)))),axis=0)
        t_interval_L = np.squeeze(np.array(np.where((time_L >= UniquePeakTimes[nn]) & (time_L <= (UniquePeakTimes[nn]+tint)))),axis=0)
        
        # Indicator peaks
        if len(t_interval_I) != 0:
            # Indicator maximum in the interval and its index within the interval
            sig_pI[nn] = max(sig_I[t_interval_I])
            i_temp1 = np.argmax(sig_I[t_interval_I])
            
            # Find total index (from first of interval plus index within interval)
            ind_pI[nn] = i_temp1 + t_interval_I[0]
            
            # Find corresponding time in indicator signal
            time_pI[nn] = time_I[ind_pI[nn].astype(int)]
        else:
            sig_pI[nn] = np.NaN
            ind_pI[nn] = np.NaN
            time_pI[nn] = np.NaN
            
        # Load peaks
        if len(t_interval_L) != 0:
            
            # Load maximum in the interval and its index within the interval
            sig_pL[nn] = max(sig_L[t_interval_L])
            i_temp2 = np.argmax(sig_L[t_interval_L])
            
            # Find total index (from first of interval plus index within interval)
            ind_pL[nn] = i_temp2 + t_interval_L[0]
            
            # Find corresponding time in indicator signal
            time_pL[nn] = time_L[ind_pL[nn].astype(int)]
        else:
            sig_pL[nn] = np.NaN
            ind_pL[nn] = np.NaN
            time_pL[nn] = np.NaN
        
        # Check the time difference between the indicator peak and the corresponding load maximum
        t_diff[nn] =  time_pL[nn] - time_pI[nn] 

    # Check correct matching
    if psw == 1:
        	
        # Labels
        tlab_I = ['{}'.format(i) for i,_ in enumerate(sig_pI)]
        tlab_L = ['{}'.format(i) for i,_ in enumerate(sig_pL)]
        
        # Figure
        plt.figure()
        plt.subplot(2,1,1)
        plt.plot(time_I,sig_I,'b',label='signal')
        plt.plot(time_pI,sig_pI,'xr',label='peaks')
        plt.legend()
        plt.ylabel('Indicator')
        plt.xlim([0,time_I[-1]])
        plt.xlim([90000,100000])
        for label, x, y in zip(tlab_I,time_pI,sig_pI):
            plt.annotate('{}'.format(label),xy=(x, y), xytext=(-5, 10),color='r',
                     textcoords='offset points', ha='center', va='bottom',
                     arrowprops=dict(arrowstyle='->', connectionstyle='arc3,rad=0'))
            
        plt.subplot(2,1,2)
        plt.plot(time_L,sig_L,'ob',label='signal')
        plt.plot(time_pL,sig_pL,'xr',label='matched peaks')
        plt.legend()
        plt.xlabel('time [s]')
        plt.ylabel('Load')
        plt.xlim([0,time_I[-1]])
        plt.xlim([90000,100000])
        for label, x, y in zip(tlab_L,time_pL,sig_pL):
            plt.annotate('{}'.format(label),xy=(x, y), xytext=(-5, 10),color='r',
                     textcoords='offset points', ha='center', va='bottom',
                     arrowprops=dict(arrowstyle='->', connectionstyle='arc3,rad=0'))
    
    return sig_pI,time_pI,ind_pI,sig_pL,time_pL,ind_pL,t_diff


# =============================================================================    


def calc_SQI_orig(p_ind,p_load,thr_range):
    """
    Original screening quality index, as used in van Essen et al. (2021). It 
    provides a quantification of the quality of a screening indicator, based on 
    a validation dataset. This original version does not allow for a single 
    false negative (no load peak above the threshold predicted, but it is 
    there). This is quite sensitive for outliers, which was the reason to 
    define an additional SQI_new (see below).
    ----------
    Input:
    ----------
    - p_ind     = Indicator peaks (or another property of these peaks, e.g., 
                  rise time or crest front steepness)
    - p_load    = Corresponding load peaks following the screening procedure 
                  (matched based on indicator zero up-crossings) 
    - thr_range = Threshold range for the loads. the SQI_orig is calculated as 
                  function of these thresholds.
    ----------
    Output:
    ----------
    - SQI_orig  = Screening quality index as a function of the input load
                  threshold range. Definition, see van Essen et al. (2021).
    ----------
    Revisions:
    ----------
    v1.0  SvE (2020) - original version
    v1.1  SvE (Jan 2023) - matlab > python
    ----------
    References:
    ----------
    van Essen, S.M., Monroy, C., Shen, Z., Helder, J., Kim, D-.H., Seng, S. and
    Ge, Z. (2021). Screening wave conditions for the occurrence of green water 
    events on sailing ships. Ocean Eng. 234 (2021) 109218.
    
    """    
    # Initialise
    Nfp = np.zeros(np.shape(thr_range))
    Nipr = np.zeros(np.shape(thr_range))
    
    # Remove nan values
    p_ind_enan = p_ind[~np.isnan(p_ind)]
    p_load_enan = p_load[~np.isnan(p_ind)]
    
    # Loop over threshold range
    for zz,thr_n in enumerate(thr_range):
                
        # Find indices of all forces above the threshold
        i_load_abthr = np.where(p_load_enan > thr_n)
        
        # Calculate number of force peaks Nfp above threshold (axis setting is 
        # important, to avoid collapsing to scalar if only one peak is selected)
        Nfp[zz] = len(np.squeeze(np.array(i_load_abthr),axis=0))
        
        # Find the lowest corresponding indicator value and find the number of 
        # indicator peaks above or equal to this lowest indicator value
        if Nfp[zz] == 0:
            Nipr[zz] = 0
        else:
            v_ind_min = np.min(p_ind_enan[i_load_abthr])
            Nipr[zz] = len(np.squeeze(np.array(np.where(p_ind_enan >= v_ind_min)),axis=0))

    # Nipr = number of indicator peaks to analyse
    # Nfp = number of force peaks above threshold
    # SQI = screening quality index
    SQI_orig = Nipr/Nfp
    
    
    return SQI_orig


def calc_block_max(Tblock,optblock,peaks_I,peakvals_I,t_peaks_I,peaks_F,t_peaks_F,peaks_Fc,t_peaks_Fc):
# def calc_block_max(Tblock,optblock,sig_I,t_sig_I,peaks_I,t_peaks_I,peaks_F,t_peaks_F,peaks_Fc,t_peaks_Fc):
    '''
    Block maximum procedure, with two options. Based on signals for which a
    peak matching procedure was already performed.
    
    INPUT:
    Tblock     = Length of blocks [s]
    optblock   = Block matching option
                - 'blockonly' = matching only based on the
                   blocks, so max. indicator and max. force within block.
                - 'keeppeakmatch' = keep the peak matching between indicator
                   and force peaks; find max. indicator peak in block and use
                   corresponding force peak.)
    peaks_I    = Indicator peaks [own unit]
    peakvals_I = Indicator values connected to these peaks. Can be peaks 
                 themselves (then peakvals_I = peaks_I), but can also e.g. be 
                 crest front steepness or risetime of these peaks
    t_peaks_I  = Times of indicator peaks [s]
    peaks_F    = All force peaks [own unit] 
    t_peaks_F  = Times of all force peaks [s]
    peaks_Fc   = Matched force peaks corresponding to the indicator peaks
                 (screening zero up-crossing matching procedure) - req. for 
                 option 'keeppeakmatch'
    t_peaks_Fc = Times of matched force peaks corresponding to the indicator peaks
                 (screening zero up-crossing matching procedure) - req. for 
                 option 'keeppeakmatch'
    
    If a signal is not required for the selected option, it is possible to
    pass a dummy or nan value to this variable.
    
    OUTPUT:
    peaks_I_block = block maximum indicator peaks for option 'optblock'
    peaks_F_block = block maximum force peaks for option 'optblock'
    
    REFERENCES:
    > Report CRS SCREAM task 1.1/1.2 by C. Monroy (Bureau Veritas) - block max. procedure
    > van Essen et al. (2021). Screening wave conditions for the occurrence of 
      green water events on sailing ships. Ocean Eng. 234/109218. 
      https://doi.org/10.1016/j.oceaneng.2021.109218 - peak matching procedure
    '''
    
    # Number of blocks
    Nblock = np.floor(max(t_peaks_I)/Tblock).astype(int)
    
    # Initialise
    t_bl_sta = np.zeros(Nblock)
    t_bl_end = np.zeros(Nblock)
    i_peaks_block = np.zeros(Nblock)
    peaks_I_block = np.zeros(Nblock)
    peakvals_I_block = np.zeros(Nblock)
    t_peaks_I_block = np.zeros(Nblock)
    peaks_F_block = np.zeros(Nblock)
    t_peaks_F_block = np.zeros(Nblock)
    peaks_I_block_check = np.zeros(Nblock)
    
    # Loop over blocks
    for bb in np.arange(Nblock):
        
        
        # Start and end time of the considered block
        t_bl_sta[bb] = bb*Tblock
        t_bl_end[bb] = bb*Tblock + Tblock - 0.0001
        
        # print('bb={}: {} to {} s'.format(bb,t_bl_sta[bb],t_bl_end[bb])) # debugging
        
        
       
        
        if optblock == 'blockonly':
                    
            # Find indicator peak for each block (TIME SERIES DATA)
            # Note*
            # indI = np.where((t_sig_I >= t_bl_sta[bb]) & (t_sig_I <= t_bl_end[bb]))
            indI_l = min(np.squeeze(np.where(t_peaks_I >= t_bl_sta[bb])))
            indI_h = max(np.squeeze(np.where(t_peaks_I <= t_bl_end[bb])))
            indI = np.arange(indI_l,indI_h+1,1,dtype=int)
                        
            # Corresponding peaks (or other indicator property) and their timing
            if np.size(np.squeeze(indI)) != 0:
                peaks_I_block[bb] = np.nanmax(peaks_I[indI])
                i_peaks_I_block = np.nanargmax(peaks_I[indI])
                t_peaks_I_block[bb] = (t_peaks_I[indI])[i_peaks_I_block]
                
                # Substitute peak values with selected peak property
                peakvals_I_block[bb] = (peakvals_I[indI])[i_peaks_I_block]
            else:
                peaks_I_block[bb] = np.NaN
                t_peaks_I_block[bb] = np.NaN
                peakvals_I_block[bb] = np.NaN


            # Find force (failure mode load) for each block: max. force
            # in block, independent of time of max. indicator. Plus its timing
            indF = np.where((t_peaks_F >= t_bl_sta[bb]) & (t_peaks_F <= t_bl_end[bb]))
            if np.size(np.squeeze(indF)) != 0:
                peaks_F_block[bb] = np.nanmax(peaks_F[indF])
                i_peaks_F_block = np.nanargmax(peaks_F[indF])
                t_peaks_F_block[bb] = (t_peaks_F[indF])[i_peaks_F_block]
            else:
                peaks_F_block[bb] = np.NaN
                t_peaks_F_block[bb] = np.NaN
                
            print(f'bb = {bb}: indicator value {peakvals_I_block[bb]}, load value {peaks_F_block[bb]}') # debugging
                
            
        elif optblock == 'keeppeakmatch':
            
            # Find indicator peak for each block (PEAK DATA)
            # Note*
            # indI2 = np.squeeze(np.where((t_peaks_I >= t_bl_sta[bb]) & (t_peaks_I <= t_bl_end[bb]))) 
            indI2_l = min(np.squeeze(np.where(t_peaks_I >= t_bl_sta[bb])))
            indI2_h = max(np.squeeze(np.where(t_peaks_I <= t_bl_end[bb])))
            indI2 = np.arange(indI2_l,indI2_h+1,1,dtype=int)
                        
            # Corresponding peaks
            peaks_I_block[bb] = np.nanmax(peaks_I[indI2])
            i_indI2 = np.nanargmax(peaks_I[indI2]) # ignore nans, as np.argmax gives index of first nan instead of maximum
            
            # Find corresponding overall peak index
            i_peaks_block =  min(indI2) + i_indI2
            
            # Timing of these peaks
            t_peaks_I_block[bb] = t_peaks_I[i_peaks_block]
            
            # Substitute peak values with selected peak property
            peakvals_I_block[bb] = peakvals_I[i_peaks_block]
                        
            # plt.figure()  # debugging
            # plt.subplot(2,1,1)
            # plt.plot(t_peaks_I[indI2],peaks_I[indI2],'xb')
            # plt.xlim([t_bl_sta[bb],t_bl_end[bb]])
            # plt.subplot(2,1,2)
            # plt.plot(t_peaks_F,peaks_F,'xr')
            # plt.xlim([t_bl_sta[bb],t_bl_end[bb]])
        
            # Find force (failure mode load) for each block: keep matching with indicator peaks.
            peaks_I_block_check[bb] = peaks_I[i_peaks_block]
            if peaks_I_block_check[bb] != peaks_I_block[bb]:
                raise Exception("There is an error in the block maximum code for option keeppeakmatch, for block = {}".format(bb))
            peaks_F_block[bb] = peaks_Fc[i_peaks_block]
            t_peaks_F_block[bb] = t_peaks_Fc[i_peaks_block]
            
            # print(f'bb = {bb}: indicator value {peakvals_I_block[bb]}, load value {peaks_F_block[bb]}') # debugging
                        
    
    # *Note: The top line in both versions omits NaN values in the time axis.
    # That goes wrong. This could also be fixed by avoiding setting the time  
    # instances to NaN/zero in scripts 'calc_RSetc' or 'match_peakedindicator_to_load',
    # but for now we chose to fix it here.
    
    
    return t_peaks_I_block,peakvals_I_block,t_peaks_F_block,peaks_F_block


def impact_alert_param(time,wave,d,U,plotswitch):
    """
    Calculates 'impact alert parameter' as defined in:
    - Stansberg, C.T. (2008). A wave impact parameter, OMAE2008-57801. OMAE Conf 
      June 15-20 2008, Estoril, Portugal
    - Stansberg, C.T. (2003). Second- and higher-order effects in steep random 
      waves. ISOPE Conf. May 25-30 2003, Honolulu, Hawaii, USA.
    
    INPUT:
        time        = time axis [s]
        wave        = linear(ised) wave elevation time trace [m]
        d           = water depth [m]
        U           = current speed [m/s]
        plotswitch  = switch, 1 = make plots, 0 = do not make plots
    
    OUTPUT:
        IAP         = Impact Alert Parameter as function of time [(m/s)**2]
        
        
        
    Created on Thu May  4 10:17:56 2023
    @author: smvanessen
    """

    # Hilbert transform of linear elevation and its phase angle
    wave_H = scipy.signal.hilbert(wave)

    wave_H_env = abs(wave_H)
    wave_H_phase = np.angle(wave_H)
    wave_H_phase_uw = np.unwrap(wave_H_phase)

    # Energy envelope
    enenv = 0.5*wave_H_env**2

    # Local frequency
    om_Hz = np.diff(wave_H_phase_uw)
    
    # This frequency is only defined when the energy envelope is not close to 
    # zero (see Stansberg, 2003). This is necessary in order to avoid large 
    # peak values of the impact parameter for zero energy. We therefore set an 
    # energy limit of 10% of the maximum energy in the envelope.
    enlimperc = 0.05 
    enlim = enlimperc*max(enenv)
    enenv1 = enenv[0:-1]
    om_Hz[enenv1<enlim] = np.nan
    
    # Wave number, local steepness and local phase speed
    om = om_Hz*2*np.pi
    k,c_ph,_ = calc_wavenumber(om,d,U)
    # k = om**2/9.81        # deep water 
    # c_ph = om/k           # deep water
    
    kA_sq = k**2*wave_H_env[0:-1]**2
    kA = np.sqrt(kA_sq)

    # impact alert parameter
    IAP = c_ph**2*kA**4
 
    # Plot
    if plotswitch == 1:
        plt.figure()
        plt.subplot(6,1,1)
        plt.plot(time,wave,'b',label='Wave elevation [m]')
        # plt.xlim([20,140])
        plt.ylim([-15,20])
        plt.legend()
        
        plt.subplot(6,1,2)
        plt.plot(time,enenv,'b',label='Energy envelope [m**2]')
        # plt.xlim([20,140])
        plt.ylim([0,160])
        plt.legend()

        plt.subplot(6,1,3)
        plt.plot(time[0:-1],om_Hz,'b',label='Local frequency [Hz]')
        # plt.xlim([20,140])
        plt.ylim([0,0.1])
        plt.legend()

        plt.subplot(6,1,4)
        plt.plot(time[0:-1],kA,'b',label='Local steepness kA [-]')
        # plt.xlim([20,140])
        plt.ylim([0,0.5])
        plt.legend()

        plt.subplot(6,1,5)
        plt.plot(time[0:-1],c_ph,'b',label='Phase speed [m/s]')
        # plt.xlim([20,140])
        plt.ylim([0,150])
        plt.legend()

        plt.subplot(6,1,6)
        plt.plot(time[0:-1],IAP,'b',label='Impact alert parameter [(m/s)**2]')
        # plt.xlim([20,140])
        # plt.ylim([0,14])
        plt.legend()

        # plt.figure()
        # plt.plot(time,wave_H_phase,'b',label='Hilbert phase')
        # plt.plot(time,wave_H_phase_uw,'r',label='Hilbert phase, unwrapped')
        # plt.legend()
        
    return IAP






